/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.util.concurrent.Executor;

import com.aquenos.epics.jackie.common.exception.ConcurrentNotificationException;
import com.aquenos.epics.jackie.common.util.ListenerLockPolicy;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;

/**
 * <p>
 * Monitor that represents a subscription to a {@link ChannelAccessChannel}.
 * Subscriptions are the primary way of getting value updates for a Channel
 * Access channel because they work asynchronously and thus do not require
 * polling.
 * </p>
 * 
 * <p>
 * A monitor can be created in any connection state (except for a destroyed
 * channel) and survives disconnects. However, it obviously will not deliver any
 * values when the channel is not connected.
 * </p>
 * 
 * <p>
 * A monitor keeps strong references to its listeners and to the corresponding
 * {@link ChannelAccessChannel}. However, the client implementation might or
 * might not keep strong references to the monitor. Therefore, user-code should
 * retain a strong reference to a monitor, in order to keep it alive, but should
 * also explicitly {@link #destroy() destroy} a monitor once it is not needed
 * any longer. Referencing the monitor from its listener might not be enough, if
 * the listener is only referenced by the monitor and there are no external
 * strong references.
 * </p>
 * 
 * <p>
 * It is safe to create a separate monitor for every place where one is needed.
 * Monitors for the same channel with the same attributes (data type, element
 * count, event mask) will actually share the same subscription, so that there
 * is no significant overhead for having many monitors. Having a separate
 * monitor for each purpose makes it easier to decide when a monitor is not
 * needed any longer and can be destroyed.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <ValueType>
 *            The exact type of the Channel Access values that are provided by
 *            the monitor events.
 */
public interface ChannelAccessMonitor<ValueType extends ChannelAccessGettableValue<?>> {

    /**
     * <p>
     * Destroys this monitor. The monitor will stop receiving events. This
     * method should be called when a monitor is not needed any longer.
     * </p>
     * 
     * <p>
     * If the {@link ListenerLockPolicy#BLOCK BLOCK} policy is in effect, the
     * monitor listeners are not going to receive another notification after
     * this method returns. If the {@link ListenerLockPolicy#IGNORE IGNORE}
     * policy is in effect, the listeners might be notified once more after this
     * method returns. If the {@link ListenerLockPolicy#REPORT REPORT} policy is
     * in effect, this method throws a {@link ConcurrentNotificationException}
     * if the listeners might be notified once more. The monitor is still
     * destroyed, despite the exception being thrown.
     * </p>
     * 
     * @throws ConcurrentNotificationException
     *             if the {@link ListenerLockPolicy} is set to {@code REPORT}
     *             and the connection-state listeners might receive one more
     *             notification after this method has returned. The channel is
     *             still destroyed, even if this exception is thrown.
     * @see #getListenerLockPolicy()
     */
    void destroy();

    /**
     * Returns the Channel Access channel that is associated with this monitor.
     * 
     * @return Channel Access channel for this monitor.
     */
    ChannelAccessChannel getChannel();

    /**
     * Returns the listener lock-policy used by this monitor. The listener
     * lock-policy defines the behavior of the {@link #destroy()} and
     * {@link #removeMonitorListener(ChannelAccessMonitorListener)} methods.
     * 
     * @return listener lock-policy used by this channel.
     * @since EPICS Jackie 2.0.0
     */
    ListenerLockPolicy getListenerLockPolicy();

    /**
     * Tells whether this monitor has been destroyed. A monitor is destroyed
     * when its {@link #destroy()} method is called or when its
     * {@link #getChannel() channel} is destroyed. It might also be destroyed if
     * it is not used any longer (there are no strong references to it and it is
     * garbage collected).
     * 
     * @return <code>true</code> if this monitor has been destroyed,
     *         <code>false</code> if it is still alive.
     */
    boolean isDestroyed();

    /**
     * <p>
     * Registers a listener with this monitor. The listener will be notified
     * when a monitor event occurs or when there is an error affecting the
     * monitor. If the monitor has already received events, the listener is
     * notified about the most recent event when being added.
     * </p>
     * 
     * <p>
     * If the specified listener is already registered with this monitor, no
     * action is taken and this method returns <code>false</code>.
     * </p>
     * 
     * <p>
     * <strong>Important note</strong>: Typically, the listener is executed by
     * the thread that performs network I/O. For this reason, it is important
     * that the listener does not block or perform any long-running operations.
     * Instead, the listener should delegate any such operation to a different
     * thread (e.g. using an {@link Executor}). Any violation of this rule might
     * result in unexpected or unpredictable behavior. In particular, a blocking
     * listener might cause a dead-lock, which results in the whole Channel
     * Access client being stalled indefinitely.
     * </p>
     * 
     * @param listener
     *            listener to be registered with this monitor.
     * @return <code>true</code> if the listener has been added,
     *         <code>false</code> if the listener has already been added
     *         earlier.
     */
    boolean addMonitorListener(
            ChannelAccessMonitorListener<? super ValueType> listener);

    /**
     * <p>
     * Unregisters a listener from this monitor. The exact behavior of this
     * method depends on the {@linkplain #getListenerLockPolicy() listener
     * lock-policy} used by this channel.
     * </p>
     * 
     * <p>
     * If the {@link ListenerLockPolicy#BLOCK BLOCK} policy is in effect, the
     * listener is not going to receive another notification after this method
     * returns. If the {@link ListenerLockPolicy#IGNORE IGNORE} policy is in
     * effect, the listener might be notified once more after this method
     * returns. If the {@link ListenerLockPolicy#REPORT REPORT} policy is in
     * effect, this method throws a {@link ConcurrentNotificationException} if
     * the listener might be notified once more. The listener is still removed,
     * despite the exception being thrown.
     * </p>
     * 
     * <p>
     * If the specified listener has already been removed from this monitor or
     * has never been added, no action is taken and this method returns
     * <code>false</code>.
     * </p>
     * 
     * @param listener
     *            listener to be unregistered from this monitor.
     * @return <code>true</code> if the listener has been removed,
     *         <code>false</code> if the listener has already been removed
     *         earlier or if it has never been added.
     * @throws ConcurrentNotificationException
     *             if the {@link ListenerLockPolicy} is set to {@code REPORT}
     *             and the removed listener might receive one more notification
     *             after this method has returned. The listener is still
     *             removed, even if this exception is thrown.
     */
    boolean removeMonitorListener(
            ChannelAccessMonitorListener<? super ValueType> listener);

}
