/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.util.concurrent.locks.ReentrantLock;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.tuple.Triple;

import com.aquenos.epics.jackie.client.AbstractChannelAccessMonitor;
import com.aquenos.epics.jackie.client.ChannelAccessChannel;
import com.aquenos.epics.jackie.client.ChannelAccessMonitor;
import com.aquenos.epics.jackie.client.ChannelAccessMonitorListener;
import com.aquenos.epics.jackie.client.ClientThreadingStrategy;
import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.common.exception.ConcurrentNotificationException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.util.ListenerLockPolicy;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Implementation of a {@link ChannelAccessMonitor} that is used by the
 * {@link DefaultChannelAccessClient}. This class acts as a facade for a
 * {@link Subscription} instance. This class is only intended for use by the
 * classes inside this package and by the {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * This class keeps a strong reference to its {@link ChannelAccessChannel} so
 * that the channel will not be destroyed as long as this monitor exists.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <ValueType>
 *            The exact type of the Channel Access values that are provided by
 *            the monitor events.
 */
public class ChannelAccessMonitorImpl<ValueType extends ChannelAccessGettableValue<?>>
        extends AbstractChannelAccessMonitor<ValueType> {

    private ChannelAccessChannelImpl channel;

    // Making destroyed volatile allows us to query it without need to acquire a
    // lock. As this flag is queried quite frequently, this is of great
    // advantage.
    private volatile boolean destroyed;
    private final ReentrantLock notificationLock = new ReentrantLock();
    private Subscription subscription;
    private ClientThreadingStrategy threadingStrategy;

    /**
     * Creates a monitor facade.
     * 
     * @param channel
     *            channel for which this monitor facade is created.
     * @param dataType
     *            the requested data-type for events. If <code>null</code>, the
     *            channel's native data-type is used.
     * @param count
     *            number of elements that shall be sent by the server. A value
     *            of zero indicates that the server shall determine the number
     *            of elements or the native element count shall be used.
     * @param listenerLockPolicy
     *            listener lock-policy used by this monitor. The lock-policy is
     *            passed on to the {@link AbstractChannelAccessMonitor}
     *            constructor and defines the behavior of the
     *            {@link #removeMonitorListener(ChannelAccessMonitorListener)}
     *            method.
     * @param mask
     *            event mask to be used for the monitor. This defines which
     *            events are going to be sent by the server. Must not be
     *            <code>DBE_NONE</code>.
     * @param threadingStrategy
     *            threading strategy used by the client. This monitor instance
     *            uses the threading strategy when notifying monitor listeners.
     * @throws IllegalArgumentException
     *             if <code>mask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE DBE_NONE}.
     */
    public ChannelAccessMonitorImpl(ChannelAccessChannelImpl channel,
            ChannelAccessValueType dataType, int count,
            ChannelAccessEventMask mask, ListenerLockPolicy listenerLockPolicy,
            ClientThreadingStrategy threadingStrategy) {
        super(listenerLockPolicy);
        if (mask == null) {
            mask = channel.getClient().getConfiguration().getDefaultEventMask();
        }
        if (!mask.isAny()) {
            throw new IllegalArgumentException(
                    "At least one bit of the event mask has to be set.");
        }
        // We keep a strong reference to the channel. This ensures that a
        // channel is not destroyed as long as a monitor for this channel
        // exists. A channel can still be destroyed explicitly, but this will
        // also destroy all monitors registered with it.
        this.channel = channel;
        this.threadingStrategy = threadingStrategy;
        while (this.subscription == null) {
            this.subscription = this.channel.getChannel()
                    .findOrCreateSubscription(count, dataType, mask);
            // We have to hold the mutex while trying to register with the
            // subscription. This ensures that we can save the returned value
            // before it might get overwritten by newer values. Otherwise, it
            // could happen that we overwrite newer values with the returned
            // value.
            synchronized (listenerLock) {
                Triple<ChannelAccessGettableValue<?>, ChannelAccessStatus, String> currentValueAndError;
                try {
                    currentValueAndError = subscription
                            .registerMonitorFacade(this);
                } catch (IllegalStateException e) {
                    // The subscription got destroyed before we could register
                    // with it. We remove the destroyed subscription and try to
                    // get a new one.
                    this.channel.getChannel()
                            .removeSubscription(this.subscription);
                    this.subscription = null;
                    continue;
                }
                // This unchecked cast is safe because we know that the values
                // associated with events have the right type. After all, we
                // requested this type.
                @SuppressWarnings("unchecked")
                ValueType value = (ValueType) currentValueAndError.getLeft();
                ChannelAccessStatus errorStatus = currentValueAndError
                        .getMiddle();
                String errorMessage = currentValueAndError.getRight();
                // It is okay to call notifyListeners(...) while holding the
                // mutex because that method will acquire the same mutex for its
                // whole execution.
                if (value != null) {
                    notifyListeners(value);
                }
                // We notify about the value first and about the error second.
                // If the latest value is newer than the latest error, we get
                // null for the status code, and thus only notify about the
                // value.
                if (errorStatus != null) {
                    notifyListeners(errorStatus, errorMessage);
                }
            }
        }
        // We want to register with our channel facade so that this monitor is
        // destroyed when the channel facade is destroyed.
        this.channel.addMonitor(this);
    }

    @Override
    public void destroy() {
        // In order to set the destroyed state, we acquire both the
        // notificationLock and the (unfortunately named) listenerLock. The
        // first lock ensures that no monitor listener is being notified after
        // we have entered the destroyed state. The second lock ensures that we
        // do not interfere with any other code that relies on the destroyed
        // state (particularly code from the parent class).
        // The order in which we take the locks is important: A listener might
        // call a method of this monitor which acquires the listenerLock. As the
        // listener already holds the notificationLock, there would be the
        // chance for a deadlock if we acquired the locks in the opposite order.
        // If the listener lock-policy is non-blocking, we only acquire the
        // notification lock if it is available. If it is not, this only means
        // that up to one monitor listener might be called after the channel has
        // been destroyed.
        boolean gotNotificationLock;
        if (listenerLockPolicy == ListenerLockPolicy.BLOCK) {
            notificationLock.lock();
            gotNotificationLock = true;
        } else {
            gotNotificationLock = notificationLock.tryLock();
        }
        try {
            synchronized (listenerLock) {
                if (destroyed) {
                    return;
                }
                destroyed = true;
            }
        } finally {
            if (gotNotificationLock) {
                notificationLock.unlock();
            }
        }
        // We have to unregister from the subscription. We can do this after
        // releasing the mutex because this monitor has already been marked as
        // destroyed and thus will ignore any further notifications.
        subscription.unregisterMonitorFacade(this);
        // We keep the reference to the channel so that the getChannel() method
        // still works. This has the disadvantage that the channel will not be
        // destroyed automatically as long as there are still references to this
        // monitor, even if the monitor has been destroyed. However, we tell the
        // channel facade that this monitor has been destroyed, so that it can
        // remove it from its list of monitors.
        channel.removeMonitor(this);
        if (!gotNotificationLock) {
            throw new ConcurrentNotificationException(
                    "The monitor has been destroyed, but a notification is already in progress, so one of the listeners is still going to be notified one more time.");
        }
    }

    @Override
    public ChannelAccessChannel getChannel() {
        return channel;
    }

    @Override
    public boolean isDestroyed() {
        return destroyed;
    }

    @Override
    protected void notifyListener(
            ChannelAccessMonitorListener<? super ValueType> listener,
            ChannelAccessStatus status, String message) {
        notificationLock.lock();
        try {
            // If this monitor has been destroyed, the listeners should not be
            // called any longer.
            if (destroyed) {
                return;
            }
            try {
                threadingStrategy.notifyMonitorListenerError(this, listener,
                        status, message);
            } catch (Throwable e) {
                channel.getClient().getConfiguration().getErrorHandler()
                        .handleError(this.getClass(), e,
                                "Monitor listener for channel \""
                                        + StringEscapeUtils
                                                .escapeJava(channel.getName())
                                        + "\" threw an exception: "
                                        + e.getMessage());
            }
        } finally {
            notificationLock.unlock();
        }
    }

    @Override
    protected void notifyListener(
            ChannelAccessMonitorListener<? super ValueType> listener,
            ValueType value) {
        notificationLock.lock();
        try {
            // If this monitor has been destroyed, the listeners should not be
            // called any longer.
            if (destroyed) {
                return;
            }
            try {
                threadingStrategy.notifyMonitorListenerEvent(this, listener,
                        value);
            } catch (Throwable e) {
                channel.getClient().getConfiguration().getErrorHandler()
                        .handleError(this.getClass(), e,
                                "Monitor listener for channel \""
                                        + StringEscapeUtils
                                                .escapeJava(channel.getName())
                                        + "\" threw an exception: "
                                        + e.getMessage());
            }
        } finally {
            notificationLock.unlock();
        }
    }

    /**
     * Notifies the monitor facade that an error event occurred. Such an event
     * occurs when the server sends an event with an error status code or when
     * there is a problem with registering the subscription with the server. A
     * simple disconnect of the channel, however, does not constitute an error.
     * This methods may only be called by the {@link Subscription} instance
     * associated with this monitor facade.
     * 
     * @param status
     *            status code indicating the kind of error.
     * @param errorMessage
     *            extended error message received from the server or
     *            <code>null</code> if no extended message has been received
     *            with the status code.
     */
    void errorReceived(ChannelAccessStatus status, String errorMessage) {
        // If this monitor has been destroyed, we are not interested in any more
        // notifications. The listeners would not be called anyway.
        if (destroyed) {
            return;
        }
        notifyListeners(status, errorMessage);
    }

    /**
     * Notifies the monitor facade that a value event occurred. Such an event
     * occurs when the server send an with a status code indicating normal
     * operation ({@link ChannelAccessStatus.StatusMessage#ECA_NORMAL
     * ECA_NORMAL}). This methods may only be called by the {@link Subscription}
     * instance associated with this monitor facade.
     * 
     * @param value
     *            value that was received from the server. This value can be
     *            expected to be of the requested type.
     */
    void eventReceived(ChannelAccessGettableValue<?> value) {
        // If this channel has been destroyed, we are not interested in any more
        // notifications. The listeners would not be called anyway.
        if (destroyed) {
            return;
        }
        // This unchecked cast is safe because we know that the values
        // associated with events have the right type. After all, we requested
        // this type.
        @SuppressWarnings("unchecked")
        ValueType typedValue = (ValueType) value;
        notifyListeners(typedValue);
    }

    @Override
    protected void finalize() throws Throwable {
        try {
            // We have to call destroy in order to ensure that the underlying
            // subscription's use counter is decremented (if this instance has
            // not been destroyed earlier).
            destroy();
        } finally {
            super.finalize();
        }
    }

}
