/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.util.concurrent.ConcurrentHashMap;

import com.aquenos.epics.jackie.client.ChannelAccessClientConfiguration;
import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolver;
import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.util.IntegerIdPool;

/**
 * <p>
 * Pool of {@link Channel Channels} for a {@link DefaultChannelAccessClient}.
 * This class is only intended for use by the classes inside this package and by
 * the {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * This class implements a lock-free pool that manages channels. It is used to
 * ensure that there is only one active {@link Channel} instance for each
 * channel name.
 * </p>
 * 
 * <p>
 * Internally, this pool keeps its channels in a {@link ConcurrentHashMap}.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ChannelPool {

    private ConcurrentHashMap<String, Channel> channels = new ConcurrentHashMap<String, Channel>();
    private IntegerIdPool cidPool;
    private ChannelNameResolver resolver;
    private ServerPool serverPool;
    // Due to subscription cancellation confirmations and certain subscription
    // event notifications using the same on-wire representation, we cannot be
    // absolutely certain that a subscription has been successfully canceled, so
    // we wait some time before reusing a subscription ID. This way,
    // subscription events that arrive late will not be associated with the
    // wrong subscription.
    private IntegerIdPool subscriptionIdPool = new IntegerIdPool(1, -1, 900000L,
            false);

    /**
     * Creates a channel pool.
     * 
     * @param configuration
     *            configuration to be used by all channels created within this
     *            pool.
     * @param resolver
     *            channel-name resolver to be used to resolve channel names when
     *            creating channels.
     * @param serverPool
     *            server pool to use when connecting channels.
     */
    public ChannelPool(ChannelAccessClientConfiguration configuration,
            ChannelNameResolver resolver, ServerPool serverPool) {
        // The CID -1 may not be used by the pool because it is reserved for
        // signaling certain error conditions (see
        // ChannelAccessErrorMessage#UNKNOWN_CHANNEL_CID).
        this.cidPool = new IntegerIdPool(0, -2,
                configuration.getCIDBlockReuseTime(), true);
        this.resolver = resolver;
        this.serverPool = serverPool;
    }

    /**
     * Returns a channel instance for the specified channel name. If an instance
     * for this channel name already exists within the pool, that instance is
     * returned. Otherwise, a new instance is created. This is handled in a way
     * that guarantees, that no active two instances will be created in
     * parallel.
     * 
     * @param channelName
     *            channel name identifying the channel in the network.
     * @return channel instance for the specified channel name.
     * @throws IllegalStateException
     *             if the channel-name resolver has been destroyed and thus no
     *             search request can be scheduled for a newly created channel.
     * @throws SerializedChannelNameTooLargeException
     *             if the serialized form of the channel name is too large to be
     *             sent in a search request.
     */
    public Channel findOrCreateChannel(String channelName) {
        Channel channel;
        channel = channels.get(channelName);
        if (channel == null) {
            channel = new Channel(this, cidPool, channelName, resolver,
                    serverPool, subscriptionIdPool);
            Channel existingChannel = channels
                    .putIfAbsent(channelName, channel);
            if (existingChannel != null) {
                // Release the CID.
                channel.release();
                channel = existingChannel;
            } else {
                // Calling start may result in an IllegalStateException or in a
                // SerializedChannelNameTooLargeException. We let these
                // exceptions bubble up.
                channel.start();
            }
        }
        // The channel that is in the map might have been destroyed but not
        // removed yet. In this case we remove the channel. However we still
        // return the destroyed channel because the calling code has to check
        // the channel's state anyway.
        if (channel.getState() == Channel.State.DESTROYED) {
            removeChannel(channel);
        }
        return channel;
    }

    /**
     * Removes a channel from the pool. This method is only used internally and
     * by a {@link Channel} when it is destroyed.
     * 
     * @param channel
     *            channel to be removed from the pool.
     */
    void removeChannel(Channel channel) {
        // We only remove exactly the specified channel object because a new
        // channel with the same name might already have been created and we do
        // not want to remove that one.
        channels.remove(channel.getName(), channel);
    }

}
