/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver;

import java.net.Inet4Address;
import java.net.InetSocketAddress;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.ExceptionSafeErrorHandler;
import com.aquenos.epics.jackie.common.exception.JavaUtilLoggingErrorHandler;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.util.EnvironmentUtil;
import com.aquenos.epics.jackie.common.util.HostNameUtil;
import com.aquenos.epics.jackie.common.util.UserNameUtil;

/**
 * Configuration for a {@link ChannelNameResolver}.
 * 
 * @author Sebastian Marsching
 */
public final class ChannelNameResolverConfiguration {

    private final Charset charset;
    private final double echoInterval;
    private final ErrorHandler errorHandler;
    private final int defaultServerPort;
    private final Set<InetSocketAddress> tcpAddresses;
    private final Set<InetSocketAddress> udpAddresses;
    private final boolean autoDetermineAddresses;
    private final double maximumSearchPeriod;
    private final int multicastTimeToLive;
    private final String hostName;
    private final String userName;

    /**
     * Creates a configuration for a channel-name resolver using environment
     * variables. The error handler is initialized with a
     * {@link JavaUtilLoggingErrorHandler}. The <code>EPICS_CA_ADDR_LIST</code>,
     * <code>EPICS_CA_AUTO_ADDR_LIST</code>, <code>EPICS_CA_NAME_SERVERS</code>,
     * <code>EPICS_CA_SERVER_PORT</code>,
     * <code>EPICS_CA_MAX_SEARCH_PERIOD</code>, <code>EPICS_CA_CONN_TMO</code>
     * and <code>EPICS_CA_MCAST_TTL</code> environment variables are used for
     * initializing the other configuration options. They have exactly the same
     * meanings and default values as specified by the Channel Access Reference
     * Manual.
     */
    public ChannelNameResolverConfiguration() {
        this(null, null, null, null, null, null, null, null, null, null, null);
    }

    /**
     * Creates a channel-name resolver configuration using the specified
     * options. Each option, that is <code>null</code> is initialized from its
     * corresponding environment variable, as the constructor
     * {@link #ChannelNameResolverConfiguration()} does.
     * 
     * @param charset
     *            character set used to encode and decode channel names. If this
     *            parameter is <code>null</code> the UTF-8 character set is
     *            used.
     * @param errorHandler
     *            error handler which is used to log errors that are encountered
     *            by the channel-name resolver or <code>null</code> if a
     *            {@link JavaUtilLoggingErrorHandler} shall be used.
     * @param hostName
     *            host name identifying the client computer when connecting to a
     *            server. If <code>null</code> the host name is determined
     *            automatically using the {@link HostNameUtil#getHostName()}
     *            method. If the host name cannot be determined automatically,
     *            the special string "&lt;unknown host&gt;" is used.
     * @param userName
     *            user name identifying the user when connecting to a server. If
     *            <code>null</code> the user name is determined automatically
     *            using the {@link UserNameUtil#getUserName()} method. If the
     *            user name cannot be determined automatically, the empty string
     *            is used.
     * @param defaultServerPort
     *            default port to assume for servers if no port is specified
     *            explicitly or <code>null</code> to use
     *            <code>EPICS_CA_SERVER_PORT</code>.
     * @param tcpAddresses
     *            set of IP-address port pairs of channel name servers that
     *            should be queried via TCP or <code>null</code> to use
     *            <code>EPICS_CA_NAME_SERVERS</code>.
     * @param udpAddresses
     *            set of IP-address port pairs of channel name servers or
     *            broadcast addresses that should be queried via UDP or
     *            <code>null</code> to use <code>EPICS_CA_ADDR_LIST</code>.
     * @param autoDetermineAddresses
     *            <code>true</code> if the broadcast addresses of all local
     *            interfaces should be added to the set of addresses that are
     *            queried via UDP, <code>false</code> if those addresses should
     *            not be added, and <code>null</code> if
     *            <code>EPICS_CA_AUTO_ADDR_LIST</code> should be used.
     * @param maximumSearchPeriod
     *            interval of the longest search period (in seconds) or
     *            <code>null</code> if <code>EPICS_CA_MAX_SEARCH_PERIOD</code>
     *            should be used. The minimum value for this setting is 60. If
     *            it is below 60, it is silently set to 60.
     * @param echoInterval
     *            interval between sending echo requests to a TCP-based
     *            name-server in seconds. If <code>null</code>, the
     *            <code>EPICS_CA_CONN_TMO</code> environment variable is used.
     *            If the environment variable is not set or its value is
     *            invalid, a default value of 30 seconds is used.
     * @param multicastTimeToLive
     *            TTL to be used for UDP search packets that are sent to
     *            multicast addresses. If <code>null</code> the
     *            <code>EPICS_CA_MCAST_TTL</code> environment variable is used.
     *            If the environment variable is not set or its value is
     *            invalid, a default value of 1 is used. The minimum value for
     *            this setting is 1 and the maximum value is 255.
     * @throws IllegalArgumentException
     *             if <code>defaultServerPort</code> is specified and zero or an
     *             invalid port number, <code>tcpAddresses</code> or
     *             <code>udpAddresses</code> contains an entry with a zero or
     *             invalid port number, or <code>echoInterval</code> or
     *             <code>multicastTimeToLive</code> are out of range.
     * @since EPICS Jackie 2.0.0
     */
    public ChannelNameResolverConfiguration(Charset charset,
            ErrorHandler errorHandler, String hostName, String userName,
            Integer defaultServerPort,
            Set<Pair<Inet4Address, Integer>> tcpAddresses,
            Set<Pair<Inet4Address, Integer>> udpAddresses,
            Boolean autoDetermineAddresses, Double maximumSearchPeriod,
            Double echoInterval, Integer multicastTimeToLive) {
        if (charset != null) {
            this.charset = charset;
        } else {
            this.charset = StandardCharsets.UTF_8;
        }
        if (errorHandler == null) {
            errorHandler = new JavaUtilLoggingErrorHandler();
        }
        if (!(errorHandler instanceof ExceptionSafeErrorHandler)) {
            errorHandler = new ExceptionSafeErrorHandler(errorHandler);
        }
        this.errorHandler = errorHandler;
        if (defaultServerPort != null) {
            this.defaultServerPort = defaultServerPort;
            if (defaultServerPort <= 0 || defaultServerPort > 65535) {
                throw new IllegalArgumentException(
                        "Invalid default server port: "
                                + this.defaultServerPort);
            }
        } else {
            this.defaultServerPort = getDefaultServerPortFromEnvironment(
                    this.errorHandler);
        }
        if (tcpAddresses != null) {
            this.tcpAddresses = validateAddresses(tcpAddresses, "TCP");
        } else {
            this.tcpAddresses = getTcpAddressesFromEnvironment(
                    this.defaultServerPort, this.errorHandler);
        }
        if (udpAddresses != null) {
            this.udpAddresses = validateAddresses(udpAddresses, "UDP");
        } else {
            this.udpAddresses = getUdpAddressesFromEnvironment(
                    this.defaultServerPort, this.errorHandler);
        }
        if (autoDetermineAddresses != null) {
            this.autoDetermineAddresses = autoDetermineAddresses;
        } else {
            this.autoDetermineAddresses = getAutoDetermineAddressesFromEnvironment();
        }
        if (maximumSearchPeriod != null) {
            if (maximumSearchPeriod < 60.0 || !Double.isFinite(
                    maximumSearchPeriod)) {
                maximumSearchPeriod = 60.0;
                this.errorHandler.handleError(
                        ChannelNameResolverConfiguration.class, null,
                        "Maxium search period is invalid, using a value of 60 seconds instead.");
            }
            this.maximumSearchPeriod = maximumSearchPeriod;
        } else {
            this.maximumSearchPeriod = getMaximumSearchPeriodFromEnvironment(
                    this.errorHandler);
        }
        if (echoInterval != null) {
            if (echoInterval < 0.1 || !Double.isFinite(echoInterval)) {
                throw new IllegalArgumentException(
                        "The echo interval must be at least 0.1 seconds.");
            }
            this.echoInterval = echoInterval;
        } else {
            this.echoInterval = getEchoIntervalFromEnvironment(
                    this.errorHandler);
        }
        if (multicastTimeToLive != null) {
            if (multicastTimeToLive < 1 || multicastTimeToLive > 255) {
                throw new IllegalArgumentException(
                        "The multicast TTL must be in the range 1 <= TTL <= 255.");
            }
            this.multicastTimeToLive = multicastTimeToLive;
        } else {
            this.multicastTimeToLive = getMulticastTimeToLiveFromEnvironment(
                    this.errorHandler);
        }
        if (hostName != null) {
            this.hostName = hostName;
        } else {
            this.hostName = getHostNameFromEnvironment();
        }
        if (userName != null) {
            this.userName = userName;
        } else {
            this.userName = getUserNameFromEnvironment();
        }
    }

    private static double getEchoIntervalFromEnvironment(
            ErrorHandler errorHandler) {
        double echoInterval = EnvironmentUtil
                .doubleFromEnvironment("EPICS_CA_CONN_TMO", 30.0, errorHandler);
        if (echoInterval < 0.1) {
            errorHandler.handleError(ChannelNameResolverConfiguration.class,
                    null,
                    "Using value 0.1 for EPICS_CA_CONN_TMO because specified value is too small: "
                            + echoInterval);
            echoInterval = 0.1;
        }
        return echoInterval;
    }

    private static int getDefaultServerPortFromEnvironment(
            ErrorHandler errorHandler) {
        return EnvironmentUtil.portNumberFromEnvironment("EPICS_CA_SERVER_PORT",
                ChannelAccessConstants.DEFAULT_SERVER_PORT, false,
                errorHandler);
    }

    private static Set<InetSocketAddress> getTcpAddressesFromEnvironment(
            int defaultServerPort, ErrorHandler errorHandler) {
        List<InetSocketAddress> addressList = EnvironmentUtil.inet4SocketAddressListFromEnvironment(
                "EPICS_CA_NAME_SERVERS", defaultServerPort, false,
                errorHandler);
        return Collections.unmodifiableSet(new LinkedHashSet<>(addressList));
    }

    private static Set<InetSocketAddress> getUdpAddressesFromEnvironment(
            int defaultServerPort, ErrorHandler errorHandler) {
        List<InetSocketAddress> addressList = EnvironmentUtil.inet4SocketAddressListFromEnvironment(
                "EPICS_CA_ADDR_LIST", defaultServerPort, false,
                errorHandler);
        return Collections.unmodifiableSet(new LinkedHashSet<>(addressList));
    }

    private static boolean getAutoDetermineAddressesFromEnvironment() {
        return EnvironmentUtil
                .booleanFromEnvironment("EPICS_CA_AUTO_ADDR_LIST", true);
    }

    private static double getMaximumSearchPeriodFromEnvironment(
            ErrorHandler errorHandler) {
        double maximumSearchPeriod = EnvironmentUtil
                .doubleFromEnvironment("EPICS_CA_MAX_SEARCH_PERIOD", 300.0,
                        errorHandler);
        if (maximumSearchPeriod < 60.0) {
            errorHandler.handleError(ChannelNameResolverConfiguration.class,
                    null,
                    "Using value 60.0 for EPICS_CA_MAX_SEARCH_PERIOD because specified value is too small: "
                            + maximumSearchPeriod);
            maximumSearchPeriod = 60.0;
        }
        return maximumSearchPeriod;
    }

    private static int getMulticastTimeToLiveFromEnvironment(
            ErrorHandler errorHandler) {
        return EnvironmentUtil
                .integerFromEnvironment("EPICS_CA_MCAST_TTL", 1, 1, 255,
                        errorHandler);
    }

    private static String getHostNameFromEnvironment() {
        String hostName = HostNameUtil.getHostName();
        if (hostName == null) {
            hostName = "<unknown host>";
        }
        return hostName;
    }

    private static String getUserNameFromEnvironment() {
        String userName = UserNameUtil.getUserName();
        if (userName == null) {
            userName = "";
        }
        return userName;
    }

    private static Set<InetSocketAddress> validateAddresses(
            Set<Pair<Inet4Address, Integer>> addresses, String type) {
        // We want to validate the addresses and ensure that the resulting set
        // is read-only.
        LinkedHashSet<InetSocketAddress> result = new LinkedHashSet<>();
        for (Pair<Inet4Address, Integer> entry : addresses) {
            Inet4Address host = entry.getLeft();
            if (host == null) {
                throw new NullPointerException();
            }
            int port = entry.getRight();
            if (port <= 0 || port > 65535) {
                throw new IllegalArgumentException(
                        "Invalid entry in " + type + " addresses: "
                                + entry.getLeft().getHostAddress()
                                + ":" + port);
            }
            result.add(new InetSocketAddress(host, port));
        }
        return Collections.unmodifiableSet(result);
    }

    /**
     * Returns the character set that should be used for encoding and decoding
     * channel names.
     * 
     * @return character set to be used for channel names.
     */
    public Charset getCharset() {
        return charset;
    }

    /**
     * <p>
     * Returns the interval between two echo requests.
     * </p>
     * 
     * <p>
     * Echo requests are sent to a TCP-based name-server in order to ensure that
     * the connection is still working. If a beacon packet from the same server
     * is received, this is considered proof that the network connection is okay
     * and thus no echo request is sent. For this reason, this setting should be
     * chosen to be at least two times as large as the max. interval between
     * sending two beacons.
     * </p>
     * 
     * <p>
     * Unlike for channel-servers, the echo mechanism does not affect the
     * connected state of any name-server. However, sending echo requests when
     * no beacons have been received still makes sense because actually sending
     * some data might help the operating system in detecting that the TCP
     * connection has been interrupted.
     * </p>
     * 
     * <p>
     * The minimum for this setting is 0.1 seconds.
     * </p>
     * 
     * @return echo interval (in seconds).
     */
    public double getEchoInterval() {
        return echoInterval;
    }

    /**
     * Returns the error handler that should be used by the channel-name
     * resolver. This configuration object ensures that the error handler
     * returned will never throw an exception by wrapping it in a
     * {@link ExceptionSafeErrorHandler}.
     * 
     * @return error handler that is exception safe.
     */
    public ErrorHandler getErrorHandler() {
        return errorHandler;
    }

    /**
     * Returns the TCP and UDP port that should be used when connecting to a
     * server for name resolution and no port has been specified for this server
     * explicitly.
     * 
     * @return default server port for name resolution.
     */
    public int getDefaultServerPort() {
        return defaultServerPort;
    }

    /**
     * Returns socket addresses of server addresses that should be contacted via
     * TCP for name resolution.
     * 
     * @return addresses of TCP name servers.
     */
    public Set<InetSocketAddress> getTcpAddresses() {
        return tcpAddresses;
    }

    /**
     * Returns socket addresses of server addresses and broadcast addresses that
     * should be contacted via UDP for name resolution.
     * 
     * @return addresses of UDP name servers and broadcast addresses.
     */
    public Set<InetSocketAddress> getUdpAddresses() {
        return udpAddresses;
    }

    /**
     * Returns whether the broadcast addresses for the local interfaces should
     * also be contacted via UDP, even if they are not explicitly listed in the
     * set returned by {@link #getUdpAddresses()}.
     * 
     * @return <code>true</code> if the broadcast addresses of all local
     *         interfaces should be used for name resolution, <code>false</code>
     *         if only the explicitly configured address should be used.
     */
    public boolean isAutoDetermineAddresses() {
        return autoDetermineAddresses;
    }

    /**
     * Returns the maximum search period (in seconds). The maximum search period
     * defines the longest time interval between two attempts to resolve a
     * channel name. However, internally a slightly different time might be used
     * so this value rather defines an order of magnitude. The minimum for this
     * setting is 60. The effective maximum is implementation dependent. For a
     * typical implementation, it is approximately 2000.
     * 
     * @return maximum search period (at least 60).
     */
    public double getMaximumSearchPeriod() {
        return maximumSearchPeriod;
    }

    /**
     * Returns the multicast TTL. This TTL is used when sending UDP packets to a
     * multicast address. On some platforms, the TTL cannot be set. In this
     * case, the multicast TTL is not actually set and some platform-specific
     * default is used instead.
     * 
     * @return multicast TTL for name-search UDP packets.
     */
    public int getMulticastTimeToLive() {
        return multicastTimeToLive;
    }

    /**
     * Returns the host name that shall be used as the identification of the
     * local machine when connecting to a Channel Access server.
     * 
     * @return host name identifying the computer the client is running on.
     */
    public String getHostName() {
        return hostName;
    }

    /**
     * Returns the user name that shall be used as the identification of the
     * user when connecting to a Channel Access server.
     * 
     * @return user name identifying the user that is using the client.
     */
    public String getUserName() {
        return userName;
    }

}
