/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.exception;

import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Simple error handler that logs the error using a
 * {@link java.util.logging.Logger}. Exceptions are logged with a level of
 * <code>WARNING</code>, other errors are logged with a level of
 * <code>INFO</code>. The logger is retrieved by getting a logger with the name
 * of the calling class.
 * 
 * @author Sebastian Marsching
 */
public class JavaUtilLoggingErrorHandler implements ErrorHandler {

    private final Level exceptionLogLevel;
    private final Level noExceptionLogLevel;

    /**
     * Default constructor. This constructor initializes the
     * <code>exceptionLogLevel</code> with <code>WARNING</code> and the
     * <code>noExceptionLogLevel</code> with <code>INFO</code>.
     */
    public JavaUtilLoggingErrorHandler() {
        this(Level.WARNING, Level.INFO);
    }

    /**
     * Creates an error handler using the specified log levels.
     *
     * @param exceptionLogLevel
     *            log level to be used for errors that include an exception.
     * @param noExceptionLogLevel
     *            log level to be used for errors that do not include an
     *            exception.
     * @since EPICS Jackie 3.1.0
     */
    public JavaUtilLoggingErrorHandler(Level exceptionLogLevel,
                                       Level noExceptionLogLevel) {
        Objects.requireNonNull(exceptionLogLevel,
                "exceptionLogLevel must not be null.");
        Objects.requireNonNull(noExceptionLogLevel,
                "noExceptionLogLevel must not be null.");
        this.exceptionLogLevel = exceptionLogLevel;
        this.noExceptionLogLevel = noExceptionLogLevel;
    }

    @Override
    public void handleError(Class<?> context, Throwable e, String description) {
        Logger logger = Logger.getLogger(context.getName());
        if (e != null) {
            logger.log(exceptionLogLevel, description != null ? description
                    : "Caught an exception.", e);
        } else {
            logger.log(noExceptionLogLevel, description);
        }
    }

}
