/*
 * Copyright 2014 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.exception;

import com.aquenos.epics.jackie.common.protocol.ChannelAccessCommand;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;

/**
 * Exception indicating that a message could not be serialized because it is not
 * supported by the specified protocol version.
 * 
 * @author Sebastian Marsching
 */
public class UnsupportedInThisVersionException extends
        MessageSerializationException {

    private static final long serialVersionUID = 8910836743059590647L;

    private final ChannelAccessVersion offendingVersion;
    private final ChannelAccessCommand offendingCommand;

    /**
     * Constructs a new exception with the specified detail message. The cause
     * is not initialized, and may subsequently be initialized by a call to
     * {@link #initCause}.
     * 
     * @param message
     *            the detail message. The detail message is saved for later
     *            retrieval by the {@link #getMessage()} method.
     * @param offendingVersion
     *            the Channel Access version that was specified and does not
     *            support the message.
     * @param offendingCommand
     *            the command of the message that could not be serialized
     *            because it is not supported by <code>offendingVersion</code>.
     */
    public UnsupportedInThisVersionException(String message,
            ChannelAccessVersion offendingVersion,
            ChannelAccessCommand offendingCommand) {
        super(message);

        this.offendingVersion = offendingVersion;
        this.offendingCommand = offendingCommand;
    }

    /**
     * Constructs a new exception with the specified Channel Access version and
     * command. This exception's message is initialized with a string containing
     * the version and command. The cause is not initialized, and may
     * subsequently be initialized by a call to {@link #initCause}.
     * 
     * @param offendingVersion
     *            the Channel Access version that was specified and does not
     *            support the message.
     * @param offendingCommand
     *            the command of the message that could not be serialized
     *            because it is not supported by <code>offendingVersion</code>.
     */
    public UnsupportedInThisVersionException(
            ChannelAccessVersion offendingVersion,
            ChannelAccessCommand offendingCommand) {
        super("Command " + offendingCommand.name()
                + " is not supported in Channel Access version "
                + offendingVersion.toString() + ".");

        this.offendingVersion = offendingVersion;
        this.offendingCommand = offendingCommand;
    }

    /**
     * Returns the Channel Access version that was requested and caused this
     * exception.
     * 
     * @return Channel Access version that caused this exception.
     */
    public ChannelAccessVersion getOffendingVersion() {
        return offendingVersion;
    }

    /**
     * Returns the command of the message that is not supported by the requested
     * Channel Access version and caused this exception.
     * 
     * @return command that caused this exception.
     */
    public ChannelAccessCommand getOffendingCommand() {
        return offendingCommand;
    }

}
