/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.InvalidPriorityException;
import com.aquenos.epics.jackie.common.exception.UnsupportedProtocolVersionException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Representation of a <code>CA_PROTO_CREATE_CHAN</code> message that is sent
 * from a server to a client. This message is sent in order to acknowledge the
 * connection of a channel and give the client information about the channel.
 * Multiple channels can be connected within the same TCP connection.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_CREATE_CHAN
 * @see ChannelAccessConnectChannelClientMessage
 */
public final class ChannelAccessConnectChannelServerMessage extends
        ChannelAccessConnectChannelMessage {

    private int cid;
    private int sid;
    private short dataType;
    private int count;

    /**
     * Creates a <code>CA_PROTO_CREATE_CHAN</code> message. This constructor
     * should be used to create a message that is supposed to be sent from a
     * server to a client.
     * 
     * @param cid
     *            channel CID used by the client. This is the identifier that
     *            the client will use to identify the channel in the response
     *            from the server. This number might be negative or zero.
     * @param sid
     *            channel SID used by the server. This is the identifier that
     *            the server will use to identify the channel in monitor, read
     *            or write requests from the client. This number might be
     *            negative or zero.
     * @param dataType
     *            the native data-type of the channel (never <code>null</code>).
     *            This is the data-type that can be processed by the server
     *            without any conversion.
     * @param count
     *            the native count of the channel. This is the number of
     *            elements that the channel's value can have on the server. Read
     *            or write operations to the channel cannot exceed this number
     *            of elements. However, a client might decide to request less
     *            elements.
     */
    public ChannelAccessConnectChannelServerMessage(int cid, int sid,
            ChannelAccessValueType dataType, int count) {
        super();

        this.cid = cid;
        this.sid = sid;
        this.dataType = dataType.toTypeCode();
        this.count = count;
    }

    private ChannelAccessConnectChannelServerMessage(
            ChannelAccessMessageHeader messageHeader, int cid, int sid,
            short dataType, int count) {
        super(messageHeader);

        this.cid = cid;
        this.sid = sid;
        this.dataType = dataType;
        this.count = count;
    }

    /**
     * Returns the channel CID used by the client. This is the identifier that
     * the client will use to identify the channel in the response from the
     * server. This number might be negative or zero.
     * 
     * @return channel CID assigned to the channel by the client.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the channel SID used by the server. This is the identifier that
     * the server will use to identify the channel in requests from the client.
     * This number might be negative or zero.
     * 
     * @return channel SID assigned to the channel by the server.
     */
    public int getChannelSID() {
        return sid;
    }

    /**
     * Returns the native data-type of this channel. This is the data type that
     * the server can process without any conversion. Returns <code>null</code>
     * if the specified data-type is unknown to this implementation.
     * 
     * @return native data-type of the channel or <code>null</code> if the
     *         numeric identifier cannot be mapped to a data type known by this
     *         implementation.
     */
    public ChannelAccessValueType getNativeDataType() {
        try {
            return ChannelAccessValueType.forTypeNumber(dataType);
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * <p>
     * Returns the native count of this channel. This is the number of elements
     * that the channel's value has on the server. The value in read and write
     * operations can never have more elements than this number. However, the
     * client can choose to use a smaller number of elements.
     * </p>
     * 
     * <p>
     * If the channel has more than {@link Integer#MAX_VALUE} elements, this
     * number can become negative. In this case, the true number of elements is
     * the returned number plus 2^32. However, channels with such a large number
     * of elements are not practical due to limitations of the Channel Access
     * protocol and thus should never appear in practice.
     * </p>
     * 
     * @return native number of elements for the channel. Might be negative for
     *         extremely large numbers of elements.
     */
    public int getNativeCount() {
        return count;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new UnsupportedProtocolVersionException(
                    "Connections from clients that do not at least support Channel Access version "
                            + ChannelAccessVersion.V4_4
                            + " are not supported by this library.");
        }
        serializeHeader(byteSink, 0, dataType, count, cid, sid, version,
                maxPayloadSize);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new UnsupportedProtocolVersionException(
                    "Connections from clients that do not at least support Channel Access version "
                            + ChannelAccessVersion.V4_4
                            + " are not supported by this library.");
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessConnectChannelServerMessage other = (ChannelAccessConnectChannelServerMessage) obj;
        return new EqualsBuilder().append(this.cid, other.cid)
                .append(this.sid, other.sid)
                .append(this.dataType, other.dataType)
                .append(this.count, other.count).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .append(sid).append(dataType).append(count).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append(", sid=");
        sb.append(sid);
        sb.append(", dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws InvalidPriorityException
     *             if the priority specified in the message header is out of
     *             range.
     */
    protected static ChannelAccessConnectChannelServerMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_CREATE_CHAN
                .getCommandNumber());
        byteSource.skip(messageHeader.getPayloadSize());
        return new ChannelAccessConnectChannelServerMessage(messageHeader,
                messageHeader.getCID(), messageHeader.getContextSpecific(),
                messageHeader.getDataType(), messageHeader.getCount());
    }

}
