/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.common.exception.UnsupportedInThisVersionException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;

/**
 * Representation of a <code>CA_PROTO_HOST_NAME</code> message. This message is
 * sent from a client to a server via TCP in order to communicate the name of
 * the client computer. This information might be used for access control.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_HOST_NAME
 */
public final class ChannelAccessHostNameMessage extends ChannelAccessMessage {

    private static final int OLD_PROTOCOL_MAX_PAYLOAD_SIZE = 0xfff8;

    private String hostName;

    /**
     * Constructor for creating a <code>CA_PROTO_HOST_NAME</code> message.
     * 
     * @param hostName
     *            name of the client computer.
     */
    public ChannelAccessHostNameMessage(String hostName) {
        super(ChannelAccessCommand.CA_PROTO_HOST_NAME);

        if (hostName == null) {
            throw new NullPointerException("Host name must not be null.");
        }

        this.hostName = hostName;
    }

    private ChannelAccessHostNameMessage(
            ChannelAccessMessageHeader messageHeader, String hostName) {
        super(ChannelAccessCommand.CA_PROTO_HOST_NAME, messageHeader);

        this.hostName = hostName;
    }

    /**
     * Returns the name of the host on which the Channel Access client is
     * running.
     * 
     * @return host name of the client computer. Might be <code>null</code> if
     *         this message instance has been reconstructed without its payload
     *         (as part of the payload of an error message).
     */
    public String getHostName() {
        return hostName;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_1)) {
            throw new UnsupportedInThisVersionException(version,
                    ChannelAccessCommand.CA_PROTO_HOST_NAME);
        }
        if (hostName == null) {
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
        // We limit the size of the string to the maximum payload size. Any
        // reasonable host name should be shorter anyway.
        Pair<byte[], Integer> hostNameBytesWithLength = NullTerminatedStringUtil
                .stringToTruncatedNullTerminatedBytes(hostName,
                        calculateMaxPayloadSize(version, maxPayloadSize),
                        ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT, charset);
        byte[] hostNameAsBytes = hostNameBytesWithLength.getLeft();
        int hostNameLength = hostNameBytesWithLength.getRight();
        serializeHeader(byteSink, hostNameLength, (short) 0, 0, 0, 0, version,
                maxPayloadSize);
        byteSink.putByteArray(hostNameAsBytes, 0, hostNameLength);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_1)) {
            throw new UnsupportedInThisVersionException(version,
                    ChannelAccessCommand.CA_PROTO_HOST_NAME);
        }
        if (hostName == null) {
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
        // We limit the size of the string to the maximum payload size.
        // Therefore, we can be sure that the payload is never too large and
        // can skip the header checks.
    }

    private static int calculateMaxPayloadSize(ChannelAccessVersion version,
            int maxPayloadSize) {
        if (!version.includes(ChannelAccessVersion.V4_9)
                && maxPayloadSize > OLD_PROTOCOL_MAX_PAYLOAD_SIZE) {
            return OLD_PROTOCOL_MAX_PAYLOAD_SIZE;
        } else {
            return maxPayloadSize;
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessHostNameMessage other = (ChannelAccessHostNameMessage) obj;
        return new EqualsBuilder().append(this.hostName, other.hostName)
                .isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(hostName).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("hostName=");
        if (hostName != null) {
            sb.append('"');
            sb.append(StringEscapeUtils.escapeJava(hostName));
            sb.append('"');
        } else {
            sb.append("null");
        }
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @param deserialize
     *            encoding to use for deserializing the host name. Usually, host
     *            names should consist of ASCII characters only, however this is
     *            not strictly enforced by the protocol, so the encoding might
     *            matter.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including the payload).
     */
    protected static ChannelAccessHostNameMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly, Charset deserialize) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_HOST_NAME
                .getCommandNumber());
        String hostName;
        if (!headerOnly) {
            byte[] hostNameAsBytes = byteSource.getByteArray(messageHeader
                    .getPayloadSize());
            hostName = NullTerminatedStringUtil.nullTerminatedBytesToString(
                    hostNameAsBytes, deserialize);
        } else {
            hostName = null;
        }
        return new ChannelAccessHostNameMessage(messageHeader, hostName);
    }

}
