/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_VERSION</code> message that is sent from a
 * server to a client via TCP. This message is usually send at the start of a
 * new TCP connection in order to communicate the protocol version supported by
 * the server.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_VERSION
 * @see ChannelAccessVersionUDPMessage
 * @see ChannelAccessVersionTCPClientMessage
 */
public final class ChannelAccessVersionTCPServerMessage extends
        ChannelAccessVersionMessage {

    /**
     * Creates a <code>CA_PROTO_VERSION</code> message. This constructor should
     * be used to create a message that is supposed to be sent from a server to
     * a client using the TCP protocol.
     * 
     * @param version
     *            Channel Access version supported by the server.
     */
    public ChannelAccessVersionTCPServerMessage(ChannelAccessVersion version) {
        super(version);

        if (version == null) {
            throw new NullPointerException(
                    "Channel Acccess version must not be null.");
        }
    }

    private ChannelAccessVersionTCPServerMessage(
            ChannelAccessMessageHeader messageHeader,
            ChannelAccessVersion version) {
        super(messageHeader, version);
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, (short) 0,
                this.version.getMinorVersionAsShort(), 0, 0, version,
                maxPayloadSize);
    }

    // This message has no additional state compare to its super class.
    // Therefore, there is no need to override equals and hashCode.

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("version=");
        sb.append(version);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     */
    protected static ChannelAccessVersionTCPServerMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_VERSION
                .getCommandNumber());
        byteSource.skip(messageHeader.getPayloadSize());
        return new ChannelAccessVersionTCPServerMessage(messageHeader,
                ChannelAccessVersion.forMinorVersion(messageHeader.getCount()));
    }

}
