/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.SerializedPayloadTooLargeException;
import com.aquenos.epics.jackie.common.exception.ShortPayloadException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueCodec;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Representation of a <code>CA_PROTO_WRITE</code> message that is sent from a
 * client to a server. This message is sent in order to write to a channel
 * without getting a response from the server.
 * 
 * @see ChannelAccessCommand#CA_PROTO_WRITE
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessWriteMessage extends ChannelAccessMessage {

    private static final byte[] PADDING_BYTES = new byte[ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT];

    private short dataType;
    private int count;
    private int sid;
    private int cid;
    private ChannelAccessPuttableValue<?> value;
    private byte[] rawValue;

    /**
     * Creates a <code>CA_PROTO_WRITE</code> message. This constructor should be
     * used to create a message that is supposed to be sent from a client to a
     * server.
     * 
     * @param dataType
     *            data type of the value in the message's payload (never
     *            <code>null</code>). This must match the type of
     *            <code>value</code>.
     * @param count
     *            number of value elements in the message's payload. This must
     *            not be negative and not exceed the number of elements in
     *            <code>value</code>.
     * @param sid
     *            server-assigned ID identifying the channel.
     * @param cid
     *            client-assigned ID identifying the channel.
     * @param value
     *            new value of the channel (never <code>null</code>).
     * @throws IllegalArgumentException
     *             if the specified data type does not match the value's data
     *             type or if <code>count</code> is negative or greater than the
     *             number of elements the value has.
     */
    public ChannelAccessWriteMessage(ChannelAccessValueType dataType,
            int count, int sid, int cid, ChannelAccessPuttableValue<?> value) {
        super(ChannelAccessCommand.CA_PROTO_WRITE);

        if (!dataType.equals(value.getType())) {
            throw new IllegalArgumentException("Specified data type ("
                    + dataType.toString()
                    + ") does not match value's data type ("
                    + value.getType().toString() + ").");
        }
        if (count < 0) {
            throw new IllegalArgumentException("Count must not be negative.");
        }
        if (count > value.getValueSize()) {
            throw new IllegalArgumentException(
                    "Number of elements to be sent must not exceed the number of elements the value has.");
        }

        this.dataType = dataType.toTypeCode();
        this.count = count;
        this.sid = sid;
        this.cid = cid;
        this.value = value.asReadOnlyValue();
    }

    private ChannelAccessWriteMessage(ChannelAccessMessageHeader messageHeader,
            short dataType, int count, int sid, int cid,
            ChannelAccessPuttableValue<?> value, byte[] rawValue) {
        super(ChannelAccessCommand.CA_PROTO_WRITE, messageHeader);

        this.dataType = dataType;
        this.count = count;
        this.sid = sid;
        this.cid = cid;
        if (value != null) {
            this.value = value.asReadOnlyValue();
        }
        this.rawValue = rawValue;
    }

    /**
     * Returns the data type of the value in this message's payload. Returns
     * <code>null</code> if the specified data-type is unknown to this
     * implementation.
     * 
     * @return data-type of the value or <code>null</code> if the numeric
     *         identifier cannot be mapped to a data type known by this
     *         implementation.
     */
    public ChannelAccessValueType getDataType() {
        try {
            ChannelAccessValueType type = ChannelAccessValueType
                    .forTypeNumber(this.dataType);
            // We treat a data type that is not valid for this message the same
            // way we treat an unknown data type.
            if (type.isPuttable()) {
                return type;
            } else {
                return null;
            }
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * Returns the number of elements the value in this message's payload has.
     * 
     * @return number of value elements.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the channel SID. This is the identifier that the server uses to
     * identify the channel. This number might be negative or zero.
     * 
     * @return channel SID assigned to the channel by the server.
     */
    public int getChannelSID() {
        return sid;
    }

    /**
     * Returns the channel CID. This is the identifier that the client uses to
     * identify the channel. This number might be negative or zero.
     * 
     * @return channel CID assigned to the channel by the client.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the new value for the channel. This method returns
     * <code>null</code> if the data-type sent by the client is not supported by
     * this implementation ({@link #getDataType()} returns <code>null</code>).
     * 
     * @return new value for the channel or <code>null</code> if the data type
     *         is not supported. The value's data type is the same that is
     *         returned by {@link #getDataType()} and the value's element count
     *         is equal to or greater than the number returned by
     *         {@link #getCount()}.
     */
    public ChannelAccessPuttableValue<?> getValue() {
        if (value == null) {
            return null;
        } else {
            return value;
        }
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        // maxPayloadSize should be aligned.
        assert (maxPayloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT == 0);
        ChannelAccessValueType dataTypeEnum = ChannelAccessValueType
                .forTypeNumber(dataType);
        if (value != null) {
            // This is the usual case. The data-type is valid and we have a
            // value.
            int maxCount = ChannelAccessValueCodec.calculateMaxCount(
                    dataTypeEnum, maxPayloadSize);
            if (count > maxCount || count < 0) {
                throw new SerializedPayloadTooLargeException(
                        "Cannot send message with value type "
                                + dataTypeEnum.toString() + " and count "
                                + (count & 0xffffffffL)
                                + " because payload size limit is set to "
                                + maxPayloadSize + ".");
            }
            int payloadSize = ChannelAccessValueCodec.calculatePayloadSize(
                    ChannelAccessValueType.forTypeNumber(dataType), count);
            int paddingSize = calculatePaddingSize(payloadSize);
            serializeHeader(byteSink, payloadSize + paddingSize, dataType,
                    count, sid, cid, version, maxPayloadSize);
            ChannelAccessValueCodec.encodePuttableValue(byteSink, value, count);
            byteSink.putByteArray(PADDING_BYTES, 0, paddingSize);
        } else if (rawValue != null) {
            // We do not have a value. This means that we could not handle the
            // type and thus we could not deserialize the value. However, we can
            // simply use the raw value.
            serializeHeader(byteSink, rawValue.length, dataType, count, sid,
                    cid, version, maxPayloadSize);
            byteSink.putByteArray(rawValue);
        } else {
            // We neither have a value nor a raw value. This means that this
            // message has been constructed from the header only and thus cannot
            // be serialized.
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        // maxPayloadSize should be aligned.
        assert (maxPayloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT == 0);
        ChannelAccessValueType dataTypeEnum = ChannelAccessValueType
                .forTypeNumber(dataType);
        if (value != null) {
            // This is the usual case. The data-type is valid and we have a
            // value.
            int maxCount = ChannelAccessValueCodec.calculateMaxCount(
                    dataTypeEnum, maxPayloadSize);
            if (count > maxCount || count < 0) {
                throw new SerializedPayloadTooLargeException(
                        "Cannot send message with value type "
                                + dataTypeEnum.toString() + " and count "
                                + (count & 0xffffffffL)
                                + " because payload size limit is set to "
                                + maxPayloadSize + ".");
            }
            int payloadSize = ChannelAccessValueCodec.calculatePayloadSize(
                    ChannelAccessValueType.forTypeNumber(dataType), count);
            int paddingSize = calculatePaddingSize(payloadSize);
            verifyHeader(payloadSize + paddingSize, count, version,
                    maxPayloadSize);
        } else if (rawValue != null) {
            // We do not have a value. This means that we could not handle the
            // type and thus we could not deserialize the value. However, we can
            // simply use the raw value.
            verifyHeader(rawValue.length, count, version, maxPayloadSize);
        } else {
            // We neither have a value nor a raw value. This means that this
            // message has been constructed from the header only and thus cannot
            // be serialized.
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessWriteMessage other = (ChannelAccessWriteMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count).append(this.sid, other.sid)
                .append(this.cid, other.cid).append(this.value, other.value)
                .append(this.rawValue, other.rawValue).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(sid).append(cid)
                .append(value).append(rawValue).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append(", sid=");
        sb.append(sid);
        sb.append(", cid=");
        sb.append(cid);
        if (value != null) {
            sb.append(", value=");
            sb.append(value);
        }
        if (rawValue != null) {
            sb.append(", rawValue=");
            addByteArrayToStringBuilder(sb, rawValue);
        }
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @param charset
     *            encoding to use for deserializing the host name. Usually, host
     *            names should consist of ASCII characters only, however this is
     *            not strictly enforced by the protocol, so the encoding might
     *            matter.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws ShortPayloadException
     *             if the message's payload size is less than the size needed
     *             for the data type and count.
     */
    protected static ChannelAccessWriteMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly, Charset charset) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_WRITE
                .getCommandNumber());

        short dataTypeNumber = messageHeader.getDataType();
        ChannelAccessValueType dataType;
        try {
            dataType = ChannelAccessValueType.forTypeNumber(dataTypeNumber);
            // We treat a data type that is not valid for this message the same
            // way we treat an unknown data type.
            if (!dataType.isPuttable()) {
                dataType = null;
            }
        } catch (IllegalArgumentException e) {
            dataType = null;
        }
        int count = messageHeader.getCount();
        int sid = messageHeader.getCID();
        int cid = messageHeader.getContextSpecific();
        ChannelAccessPuttableValue<?> value;
        byte[] rawValue;
        if (headerOnly) {
            value = null;
            rawValue = null;
        } else {
            int actualPayloadSize = messageHeader.getPayloadSize();
            if (dataType == null) {
                value = null;
                rawValue = byteSource.getByteArray(actualPayloadSize);
            } else {
                int expectedPayloadSize;
                try {
                    // calculatePayloadSize throws an IllegalArgumentException
                    // if count is negative or too large to represent the
                    // resulting number as an int.
                    expectedPayloadSize = ChannelAccessValueCodec
                            .calculatePayloadSize(dataType, count);
                } catch (IllegalArgumentException e) {
                    throw new ShortPayloadException("Payload size of "
                            + actualPayloadSize + " is too small to hold a "
                            + dataType.toString() + " with "
                            + (count & 0xffffffffL) + " elements.");
                }
                if (actualPayloadSize < expectedPayloadSize) {
                    // Some Channel Access clients might shorten single-element
                    // strings to the effectively needed payload size (aligned to
                    // 8 bytes).
                    if (count == 1
                            && dataType.toSimpleType().equals(ChannelAccessValueType.DBR_STRING)
                            && actualPayloadSize >= ChannelAccessValueCodec.calculatePayloadSize(
                                    dataType, 0)) {
                        expectedPayloadSize = actualPayloadSize;
                    } else {
                        throw new ShortPayloadException(
                                "Received a CA_PROTO_WRITE message with a payload size of "
                                        + actualPayloadSize
                                        + ". However, the message should have a payload of at least "
                                        + expectedPayloadSize + " bytes.");
                    }
                }
                value = ChannelAccessValueCodec.decodePuttableValue(byteSource,
                        dataType, expectedPayloadSize, count, charset);
                byteSource.skip(actualPayloadSize - expectedPayloadSize);
                rawValue = null;
            }
        }
        return new ChannelAccessWriteMessage(messageHeader, dataTypeNumber,
                count, sid, cid, value, rawValue);
    }

}
