/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.net.Inet4Address;
import java.net.InetSocketAddress;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.ExceptionSafeErrorHandler;
import com.aquenos.epics.jackie.common.exception.NullErrorHandler;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.text.StrMatcher;
import org.apache.commons.lang3.text.StrTokenizer;
import org.apache.commons.lang3.tuple.Pair;

/**
 * Utility functions for handling environment variables.
 * 
 * @author Sebastian Marsching
 */
public final class EnvironmentUtil {

    /**
     * Private constructor. All methods in this class are static.
     */
    private EnvironmentUtil() {
    }

    /**
     * Converts the specified string to a port number. Returns -1 if the string
     * is not a valid port number.
     * 
     * @param portString
     *            string to be converted.
     * @return port number from <code>portString</code> or -1 if
     *         <code>portString</code> is not a valid port number.
     */
    private static int toPortNumber(String portString) {
        if (!StringUtils.isNumeric(portString) || portString.isEmpty()
                || portString.length() > 5) {
            return -1;
        }
        int port = Integer.parseInt(portString, 10);
        if (port > 65535) {
            return -1;
        }
        return port;
    }

    /**
     * Parses an environment variable that contains a list of IPv4 address with
     * (optional) port numbers. The format is expected to use spaces, tabs,
     * carriage returns, line feeds, or form feeds as delimiters. Each token is
     * expected to have the format "N.N.N.N:M" or "N.N.N.N" where "N.N.N.N" is a
     * valid IPv4 address and "M" is a valid TCP or UDP port number. Tokens that
     * do not match the expected format are silently ignored. If no port number
     * is specified, the <code>defaultPort</code> is used.
     * 
     * @param environmentVariableName
     *            name of the environment variable that shall be parsed.
     * @param defaultPort
     *            port number to assume for addresses that do not specify a port
     *            number explicitly.
     * @return list of IP-address port-number pairs created from the content of
     *         the specified environment variable. If the environment variable
     *         is undefined or does not contain a correctly formatted token, the
     *         empty list is returned.
     * @throws IllegalArgumentException
     *             if <code>defaultPort</code> is less than zero or greater than
     *             65535.
     * @deprecated
     *     since EPICS Jackie 3.1.0. Consider using
     *     {@link #inet4SocketAddressListFromEnvironment(String, int, boolean, ErrorHandler)}
     *     instead.
     */
    @Deprecated
    public static List<Pair<Inet4Address, Integer>> addressAndPortListFromEnvironment(
            String environmentVariableName, int defaultPort) {
        if (defaultPort < 0 || defaultPort > 65535) {
            throw new IllegalArgumentException("Not a valid port number: "
                    + defaultPort);
        }
        LinkedList<Pair<Inet4Address, Integer>> addressList = new LinkedList<Pair<Inet4Address, Integer>>();
        String environmentVariable = System.getenv(environmentVariableName);
        if (environmentVariable == null) {
            return Collections.emptyList();
        }
        StrTokenizer tokenizer = new StrTokenizer(environmentVariable);
        tokenizer.setTrimmerMatcher(StrMatcher.splitMatcher());
        while (tokenizer.hasNext()) {
            String token = tokenizer.next();
            String addressPart;
            String portPart;
            int indexOfColon = token.indexOf(':');
            if (indexOfColon == -1) {
                addressPart = token;
                portPart = null;
            } else {
                addressPart = token.substring(0, indexOfColon);
                portPart = token.substring(indexOfColon + 1, token.length());
            }
            Inet4Address address;
            try {
                address = Inet4AddressUtil.stringToInetAddress(addressPart);
            } catch (IllegalArgumentException e) {
                // The token is not a valid IPv4 address, so we ignore it.
                continue;
            }
            int port = defaultPort;
            if (portPart != null) {
                port = toPortNumber(portPart);
                if (port < 1) {
                    // The port number is invalid, so ignore this token.
                    continue;
                }
            }
            addressList.add(Pair.of(address, port));
        }
        return addressList;
    }

    /**
     * Parses an environment variable that contains a boolean. The string "YES"
     * is interpreted as <code>true</code> and the string "NO" is interpreted as
     * <code>false</code>. The comparison is ignoring case. If the value of the
     * environment variable matches neither string nor the environment variable
     * is not set, <code>defaultValue</code> is returned.
     * 
     * @param environmentVariableName
     *            name of the environment variable that shall be parsed.
     * @param defaultValue
     *            default value that is used if the environment variable is not
     *            set or its content cannot clearly be mapped to a boolean
     *            value.
     * @return boolean value from the specified environment variable or
     *         <code>defaultValue</code> if value could be determined from the
     *         environment variable.
     */
    public static boolean booleanFromEnvironment(
            String environmentVariableName, boolean defaultValue) {
        String environmentVariable = System.getenv(environmentVariableName);
        if (environmentVariable == null) {
            return defaultValue;
        }
        if (environmentVariable.equalsIgnoreCase("YES")) {
            return true;
        } else if (environmentVariable.equalsIgnoreCase("NO")) {
            return false;
        } else {
            return defaultValue;
        }
    }

    /**
     * Parses an environment variable that contains a TCP or UDP port number. If
     * the environment variable is undefined or does not contain a valid port
     * number (e.g. not a number or a number that is less than 0 or greater than
     * 65535), the <code>defaultValue</code> is returned. If the port number is
     * zero and <code>allowZero</code> is <code>false</code>,
     * <code>defaultValue</code> is returned.
     * 
     * @param environmentVariableName
     *            name of the environment variable that shall be parsed.
     * @param defaultValue
     *            default value that is used if the environment variable is not
     *            set or its content is not a valid port number.
     * @param allowZero
     *            if <code>true</code> use the default value if the port number
     *            is zero, if <code>false</code> allow a zero port number.
     * @return port number from the specified environment variable or
     *         <code>defaultValue</code>.
     * @deprecated since EPICS Jackie 3.1.0.
     */
    @Deprecated
    public static int portNumberFromEnvironment(String environmentVariableName,
            int defaultValue, boolean allowZero) {
        return portNumberFromEnvironment(environmentVariableName, defaultValue,
                allowZero, null);
    }

    /**
     * Parses an environment variable that contains a TCP or UDP port number. If
     * the environment variable is undefined or does not contain a valid port
     * number (e.g. not a number or a number that is less than 0 or greater than
     * 65535), the <code>defaultValue</code> is returned. If the port number is
     * zero and <code>allowZero</code> is <code>false</code>,
     * <code>defaultValue</code> is returned.
     *
     * @param environmentVariableName
     *            name of the environment variable that shall be parsed.
     * @param defaultValue
     *            default value that is used if the environment variable is not
     *            set or its content is not a valid port number.
     * @param allowZero
     *            if <code>true</code> use the default value if the port number
     *            is zero, if <code>false</code> allow a zero port number.
     * @param errorHandler
     *            error handler which is notified when there is a problem with
     *            the value of the environment variable. If <code>null</code>,
     *            such problems are corrected silently.
     * @return port number from the specified environment variable or
     *         <code>defaultValue</code>.
     * @since EPICS Jackie 3.1.0
     */
    public static int portNumberFromEnvironment(String environmentVariableName,
            int defaultValue, boolean allowZero, ErrorHandler errorHandler) {
        // Ensure that we have an error handler and that it does not throw
        // exceptions.
        if (errorHandler == null) {
            errorHandler = new NullErrorHandler();
        } else if (!(errorHandler instanceof ExceptionSafeErrorHandler)) {
            errorHandler = new ExceptionSafeErrorHandler(errorHandler);
        }
        String environmentVariable = System.getenv(environmentVariableName);
        if (environmentVariable == null) {
            return defaultValue;
        }
        int port = toPortNumber(environmentVariable);
        if (port == -1 || (port == 0 && !allowZero)) {
            errorHandler.handleError(EnvironmentUtil.class, null,
                    "Using default value " + defaultValue + " for "
                            + environmentVariableName
                            + " because specified value is invalid: "
                            + environmentVariable);
            return defaultValue;
        }
        return port;
    }

    /**
     * Parses an environment variable that contains a double-precision
     * floating-point number. If the environment variable is undefined or does
     * not contain a valid number, the <code>defaultValue</code> is returned.
     * 
     * @param environmentVariableName
     *            name of the environment variable that shall be parsed.
     * @param defaultValue
     *            default value that is used if the environment variable is not
     *            set or its content is not a valid floating-point number.
     * @return floating-point number from the specified environment variable or
     *         <code>defaultValue</code> if the environment variable is not set
     *         or does not contain a valid floating-point number.
     * @deprecated since EPICS Jackie 3.1.0.
     */
    @Deprecated
    public static double doubleFromEnvironment(String environmentVariableName,
            double defaultValue) {
        return doubleFromEnvironment(environmentVariableName, defaultValue,
                null);
    }

    /**
     * Parses an environment variable that contains a double-precision
     * floating-point number. If the environment variable is undefined or does
     * not contain a valid number, the <code>defaultValue</code> is returned.
     *
     * @param environmentVariableName
     *            name of the environment variable that shall be parsed.
     * @param defaultValue
     *            default value that is used if the environment variable is not
     *            set or its content is not a valid floating-point number.
     * @param errorHandler
     *            error handler which is notified when there is a problem with
     *            the value of the environment variable. If <code>null</code>,
     *            such problems are corrected silently.
     * @return floating-point number from the specified environment variable or
     *         <code>defaultValue</code> if the environment variable is not set
     *         or does not contain a valid, finite floating-point number.
     */
    public static double doubleFromEnvironment(String environmentVariableName,
            double defaultValue, ErrorHandler errorHandler) {
        // Ensure that we have an error handler and that it does not throw
        // exceptions.
        if (errorHandler == null) {
            errorHandler = new NullErrorHandler();
        } else if (!(errorHandler instanceof ExceptionSafeErrorHandler)) {
            errorHandler = new ExceptionSafeErrorHandler(errorHandler);
        }
        String environmentVariable = System.getenv(environmentVariableName);
        if (environmentVariable == null) {
            return defaultValue;
        }
        double value;
        try {
            value = Double.parseDouble(environmentVariable);
        } catch (NumberFormatException e) {
            value = Double.NaN;
        }
        if (!Double.isFinite(value)) {
            errorHandler.handleError(EnvironmentUtil.class, null,
                    "Using default value " + defaultValue + " for "
                            + environmentVariableName
                            + " because specified value is invalid: "
                            + environmentVariable);
            return defaultValue;
        } else {
            return value;
        }
    }

    /**
     * <p>
     * Parses an environment variable that contains a whitespace-separated list
     * of host-port pairs. If an entry specified a port number, it must be
     * separated from the host part by a colon (e.g.
     * <code>host.example.com:12345</code>, <code>127.0.0.1:4567</code>).
     * </p>
     *
     * <p>
     * Only IPv4 addresses are supported by this function. If a hostname
     * resolves to both IPv4 and IPv6 addresses, the IPv6 addresses are silently
     * ignored. If it only resolves to IPv6 addresses, an error message is
     * logged using the provided <code>errorHandler</code>.
     * </p>
     *
     * @param environmentVariableName
     *            name of the environment variable from which the value is
     *            retrieved.
     * @param defaultPort
     *            port number that is used for entries that do not explicitly
     *            specify a port number.
     * @param allowZero
     *            controls whether a port number of zero is allowed or not.
     * @param errorHandler
     *            error handler that is called when an invalid token is
     *            encountered (e.g. a token with an invalid port number or a
     *            hostname that cannot be resolved). If <code>null</code>, such
     *            errors are silently ignored.
     * @return
     *     list of socket addresses retrieved from the specified environment
     *     variable.
     * @throws IllegalArgumentException
     *     if <code>defaultPort</code> does not represent a valid port number.
     * @since EPICS Jackie 3.1.0
     * @see Inet4AddressUtil#stringToInet4SocketAddressList(String, int, boolean, ErrorHandler)
     */
    public static List<InetSocketAddress> inet4SocketAddressListFromEnvironment(
            String environmentVariableName, int defaultPort, boolean allowZero,
            ErrorHandler errorHandler) {
        String environmentVariable = System.getenv(environmentVariableName);
        if (environmentVariable == null) {
            return Collections.emptyList();
        }
        ErrorHandler wrappingErrorHandler = null;
        if (errorHandler != null) {
            wrappingErrorHandler = (context, e, description) -> {
                description = "Error while parsing environment variable "
                        + environmentVariableName
                        + ": " + description;
                errorHandler.handleError(context, e, description);
            };
        }
        return Inet4AddressUtil.stringToInet4SocketAddressList(
                environmentVariable, defaultPort, allowZero,
                wrappingErrorHandler);
    }

    /**
     * Parses an environment variable that contains an integer number. If the
     * environment variable is undefined or does not contain a valid number, the
     * <code>defaultValue</code> is returned. If the environment variable
     * contains an integer number but this number exceeds the specified limits,
     * it is coerced to the limits.
     * 
     * @param environmentVariableName
     *            name of the environment variable that shall be parsed.
     * @param defaultValue
     *            default value that is used if the environment variable is not
     *            set or its content is not a valid integer number.
     * @param lowerLimit
     *            lower limit for the number returned (inclusive). The returned
     *            number is never less than this limit.
     * @param upperLimit
     *            upper limit for the number returned (inclusive). The returned
     *            number is never greater than this limit.
     * @return integer number from the specified environment variable or
     *         <code>defaultValue</code> if the environment variable is not set
     *         or does not contain a valid integer number.
     * @throws IllegalArgumentException
     *             if the lower limit is greater than the upper limit or if the
     *             default value is not within the interval defined by the lower
     *             and upper limit.
     * @deprecated since EPICS Jackie 3.1.0.
     */
    @Deprecated
    public static int integerFromEnvironment(String environmentVariableName,
            int defaultValue, int lowerLimit, int upperLimit) {
        return integerFromEnvironment(environmentVariableName, defaultValue,
                lowerLimit, upperLimit, null);
    }

    /**
     * Parses an environment variable that contains an integer number. If the
     * environment variable is undefined or does not contain a valid number, the
     * <code>defaultValue</code> is returned. If the environment variable
     * contains an integer number but this number exceeds the specified limits,
     * it is coerced to the limits.
     *
     * @param environmentVariableName
     *            name of the environment variable that shall be parsed.
     * @param defaultValue
     *            default value that is used if the environment variable is not
     *            set or its content is not a valid integer number.
     * @param lowerLimit
     *            lower limit for the number returned (inclusive). The returned
     *            number is never less than this limit.
     * @param upperLimit
     *            upper limit for the number returned (inclusive). The returned
     *            number is never greater than this limit.
     * @return integer number from the specified environment variable or
     *         <code>defaultValue</code> if the environment variable is not set
     *         or does not contain a valid integer number.
     * @throws IllegalArgumentException
     *             if the lower limit is greater than the upper limit or if the
     *             default value is not within the interval defined by the lower
     *             and upper limit.
     */
    public static int integerFromEnvironment(String environmentVariableName,
            int defaultValue, int lowerLimit, int upperLimit,
            ErrorHandler errorHandler) {
        if (lowerLimit > upperLimit) {
            throw new IllegalArgumentException(
                    "Lower limit must not be greater than upperLimit.");
        }
        if (defaultValue < lowerLimit || defaultValue > upperLimit) {
            throw new IllegalArgumentException(
                    "The default value must be within the interval defined by the lower and upper limit.");
        }
        // Ensure that we have an error handler and that it does not throw
        // exceptions.
        if (errorHandler == null) {
            errorHandler = new NullErrorHandler();
        } else if (!(errorHandler instanceof ExceptionSafeErrorHandler)) {
            errorHandler = new ExceptionSafeErrorHandler(errorHandler);
        }
        String environmentVariable = System.getenv(environmentVariableName);
        if (environmentVariable == null) {
            return defaultValue;
        }
        int value;
        try {
            value = Integer.parseInt(environmentVariable);
        } catch (NumberFormatException e) {
            // We do not parse the exception to the error handler because this
            // might result in a higher log level than we want.
            errorHandler.handleError(EnvironmentUtil.class, null,
                    "Using default value " + defaultValue + " for "
                            + environmentVariableName
                            + " because specified value is invalid: "
                            + environmentVariable);
            return defaultValue;
        }
        if (value < lowerLimit) {
            errorHandler.handleError(EnvironmentUtil.class, null,
                    "Using value " + lowerLimit + " for "
                            + environmentVariableName
                            + " because specified value is invalid: "
                            + environmentVariable);
            return lowerLimit;
        }
        if (value > upperLimit) {
            errorHandler.handleError(EnvironmentUtil.class, null,
                    "Using value " + upperLimit + " for "
                            + environmentVariableName
                            + " because specified value is invalid: "
                            + environmentVariable);
            return upperLimit;
        }
        return value;
    }

}
