/*
 * Copyright 2014-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Pattern;

import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.ExceptionSafeErrorHandler;
import com.aquenos.epics.jackie.common.exception.NullErrorHandler;
import org.apache.commons.lang3.StringUtils;

/**
 * Utility methods for dealing with IPv4 addresses.
 * 
 * @author Sebastian Marsching
 */
public final class Inet4AddressUtil {

    private static final Pattern WHITESPACE_PATTERN = Pattern.compile("[ \t\n\r\f]");

    private Inet4AddressUtil() {
    }

    /**
     * The loopback address (127.0.0.1).
     */
    public static final Inet4Address LOOPBACK_ADDRESS;
    static {
        try {
            LOOPBACK_ADDRESS = (Inet4Address) InetAddress
                    .getByAddress(new byte[] { 127, 0, 0, 1 });
        } catch (UnknownHostException e) {
            throw new RuntimeException(
                    "Unexpected exception when trying to generate loopback address.");
        }
    }

    /**
     * Converts an IPv4 address from a series of four bytes to an integer. The
     * bytes must be in network byte-order (big endian).
     * 
     * @param address
     *            IPv4 address as a series of four bytes.
     * @return IPv4 address as an integer.
     */
    public static int addressToInt(byte[] address) {
        if (address.length != 4) {
            throw new IllegalArgumentException(
                    "An IPv4 address must have 4 bytes, but this address has "
                            + address.length + " bytes.");
        }
        int intAddress = address[3] & 0xFF;
        intAddress |= ((address[2] << 8) & 0xFF00);
        intAddress |= ((address[1] << 16) & 0xFF0000);
        intAddress |= ((address[0] << 24) & 0xFF000000);
        return intAddress;
    }

    /**
     * Converts from an IPv4 address from an integer to a series of four bytes.
     * The bytes will be in network byte-order (big endian).
     * 
     * @param address
     *            IPv4 address as an integer.
     * @return IPv4 address as a series of four bytes.
     */
    public static byte[] addressToBytes(int address) {
        byte[] byteAddress = new byte[4];
        byteAddress[0] = (byte) ((address >>> 24) & 0xFF);
        byteAddress[1] = (byte) ((address >>> 16) & 0xFF);
        byteAddress[2] = (byte) ((address >>> 8) & 0xFF);
        byteAddress[3] = (byte) (address & 0xFF);
        return byteAddress;
    }

    /**
     * <p>
     * Parses a string that contains a whitespace-separated list of host-port
     * pairs. If an entry specified a port number, it must be separated from the
     * host part by a colon (e.g. <code>host.example.com:12345</code>,
     * <code>127.0.0.1:4567</code>).
     * </p>
     *
     * <p>
     * Only IPv4 addresses are supported by this function. If a hostname
     * resolves to both IPv4 and IPv6 addresses, the IPv6 addresses are silently
     * ignored. If it only resolves to IPv6 addresses, an error message is
     * logged using the provided <code>errorHandler</code>.
     * </p>
     *
     * @param addressListString
     *            string that shall be parsed.
     * @param defaultPort
     *            port number that is used for entries that do not explicitly
     *            specify a port number.
     * @param allowZero
     *            controls whether a port number of zero is allowed or not.
     * @param errorHandler
     *            error handler that is called when an invalid token is
     *            encountered (e.g. a token with an invalid port number or a
     *            hostname that cannot be resolved). If <code>null</code>, such
     *            errors are silently ignored.
     * @return
     *     list of socket addresses converted from the specified string.
     * @throws IllegalArgumentException
     *     if <code>defaultPort</code> does not represent a valid port number.
     * @since EPICS Jackie 3.1.0
     */
    public static List<InetSocketAddress> stringToInet4SocketAddressList(
            String addressListString, int defaultPort, boolean allowZero,
            ErrorHandler errorHandler) {
        if (defaultPort < 0 || defaultPort > 65535
                || (!allowZero && defaultPort < 1)) {
            throw new IllegalArgumentException("Not a valid port number: "
                    + defaultPort);
        }
        // Ensure that we have an error handler and that it does not throw
        // exceptions.
        if (errorHandler == null) {
            errorHandler = new NullErrorHandler();
        } else if (!(errorHandler instanceof ExceptionSafeErrorHandler)) {
            errorHandler = new ExceptionSafeErrorHandler(errorHandler);
        }
        LinkedList<InetSocketAddress> addressList = new LinkedList<>();
        for (String token : WHITESPACE_PATTERN.split(addressListString)) {
            if (token.isEmpty()) {
                continue;
            }
            String hostPart;
            String portPart;
            // We do not support IPv6 addresses here, but if we encounter an
            // IPv6 address, we still want the host and port part to be
            // separated correctly, so that the error message that might be
            // generated is not misleading.
            if (token.startsWith("[") && token.endsWith("]")) {
                hostPart = token;
                portPart = null;
            } else {
                int indexOfColon = token.lastIndexOf(':');
                if (indexOfColon == -1) {
                    hostPart = token;
                    portPart = null;
                } else {
                    hostPart = token.substring(0, indexOfColon);
                    portPart = token.substring(indexOfColon + 1);
                }
            }
            int port;
            if (portPart != null) {
                try {
                    port = Integer.parseInt(portPart);
                } catch (NumberFormatException e) {
                    port = -1;
                }
                if (port > 65535 || port < 0 || (port == 0 && !allowZero)) {
                    errorHandler.handleError(Inet4AddressUtil.class, null,
                            "Ignoring token with invalid port number: "
                                    + token);
                    continue;
                }
            } else {
                port = defaultPort;
            }
            boolean foundValidAddress = false;
            try {
                InetAddress[] allAddresses = InetAddress.getAllByName(hostPart);
                for (InetAddress address : allAddresses) {
                    if (address instanceof Inet4Address) {
                        addressList.add(new InetSocketAddress(address, port));
                        foundValidAddress = true;
                    }
                }
            } catch (UnknownHostException e) {
                errorHandler.handleError(EnvironmentUtil.class, null,
                        "Ignoring token with unresolvable hostname: " + token);
                continue;
            }
            if (!foundValidAddress) {
                errorHandler.handleError(EnvironmentUtil.class, null,
                        "Ignoring token that does not resolve to an IPv4 address: "
                                + token);
            }
        }
        return addressList;
    }

    /**
     * Parses a string representing an IPv4 address. The address must be a valid
     * IPv4 address. In particular, it must only contain digits and the period
     * (.), must consist of exactly four parts divided by periods, and each part
     * must represent a number between 0 and 255. Leading zeros in parts are
     * allowed, however there may not be more than three digits per part. Each
     * part must have at least one digit.
     * 
     * @param addressString
     *            IPv4 address in typical notation (NNN.NNN.NNN.NNN).
     * @return IPv4 address created from string.
     * @throws IllegalArgumentException
     *             if <code>addressString</code> is not a valid IPv4 address.
     */
    public static Inet4Address stringToInetAddress(String addressString) {
        String[] addressParts = StringUtils.split(addressString, '.');
        if (addressParts.length != 4) {
            throw new IllegalArgumentException("Not a valid IPv4 address: "
                    + addressString);
        }
        byte[] addressBytes = new byte[4];
        for (int i = 0; i < addressBytes.length; i++) {
            String addressPart = addressParts[i];
            if (addressPart.length() < 1 || addressPart.length() > 3
                    || !StringUtils.isNumeric(addressPart)) {
                throw new IllegalArgumentException("Not a valid IPv4 address: "
                        + addressString);
            }
            int numericPart = Integer.parseInt(addressPart, 10);
            // We do not have to check for negative values because we know that
            // the string only contains digits.
            if (numericPart > 255) {
                throw new IllegalArgumentException("Not a valid IPv4 address: "
                        + addressString);
            }
            addressBytes[i] = (byte) numericPart;
        }
        try {
            return (Inet4Address) InetAddress.getByAddress(addressBytes);
        } catch (UnknownHostException e) {
            // This should never happen, because we verified the address string.
            throw new RuntimeException(e);
        }
    }

}
