/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicIntegerFieldUpdater;

/**
 * <p>
 * Pool of integer IDs. The pool is both thread safe and lock free. IDs are
 * acquired as objects and the life-cycle of the numeric ID is bound to the
 * life-cycle of this object. Numeric IDs are automatically returned to the pool
 * when their associated ID object is finalized. However, as the time of
 * finalization is not predictable, ID objects should be released explicitly if
 * possible in order to avoid exhaustion of the ID pool.
 * </p>
 * 
 * <p>
 * The pool generates integer IDs from a configurable range, ensuring that each
 * numeric ID is only handed out once. In order to accommodate for applications
 * that need a large number of IDs over their life-cycle, the pool supports the
 * reuse of numeric IDs that have been released.
 * </p>
 * 
 * <p>
 * Optionally, the pool can be configured to block released IDs from immediate
 * reuse for a certain time. This is useful in environments where IDs leave the
 * scope of the application and thus external entities might have dangling
 * references to those IDs. In this case, configuring a time for which the ID is
 * blocked from reuse after having been released can help in avoiding a
 * situation in which an external entity uses an ID that has now been reused for
 * a different purpose.
 * </p>
 * 
 * <p>
 * Internally, this implementation uses a {@link BasicIntegerIdPool} for
 * managing numeric IDs. Therefore the performance considerations in the
 * documentation of that class also apply to this class.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class IntegerIdPool {

    /**
     * <p>
     * Integer identifier. The numeric value (as returned by the {@link #get()}
     * method) is guaranteed to be unique for the whole life-time of this
     * object, meaning that no other ID acquired from the same
     * {@link IntegerIdPool} will return the same numeric identifier. However,
     * once this ID has been released, it might be reused by a newly acquired
     * ID, depending on the ID pool's configuration.
     * </p>
     * 
     * <p>
     * This ID (and its associated numeric identifier) are automatically
     * released when this object is finalized. However, due to the inherently
     * unpredictable nature of garbage collection, it is strongly recommended to
     * explicitly release the ID once it is not needed any longer in order to
     * avoid exhaustion of the associated ID pool. The ID can be released
     * explicitly by calling its {@link #release()} method.
     * </p>
     * 
     * <p>
     * The {@link #equals(Object)} method of this class actually compares for
     * object equality, not equality of the numeric identifier. This is
     * reasonable because two different ID objects with the same numeric
     * identifier can only exist when at least one of these IDs has been
     * released. A released ID, however, is never considered the same ID as a
     * different released or alive ID.
     * </p>
     * 
     * @author Sebastian Marsching
     */
    public static final class Id {

        /**
         * Field updater for the {@link #released} field. We use a volatile
         * field in combination with an updater instead of an
         * {@link AtomicBoolean} in order to keep the number of objects as small
         * as possible.
         */
        private static final AtomicIntegerFieldUpdater<Id> RELEASED_UPDATER = AtomicIntegerFieldUpdater
                .newUpdater(Id.class, "released");

        /**
         * The ID pool that actually managed the ID. This reference is needed so
         * that we can release the ID back to the pool.
         */
        private BasicIntegerIdPool idPool;

        /**
         * Flag indicating whether this ID has been released. This field is used
         * indirectly through {@link #RELEASED_UPDATER}.
         */
        @SuppressWarnings("unused")
        private volatile int released;

        /**
         * Numeric ID that has been acquired from the {@link BasicIntegerIdPool}
         * .
         */
        private int id;

        /**
         * Private constructor for use by {@link IntegerIdPool#acquireId()}.
         * 
         * @param idPool
         *            the actual ID pool for acquiring and releasing the numeric
         *            ID.
         */
        private Id(BasicIntegerIdPool idPool) {
            this.idPool = idPool;
            this.id = this.idPool.acquireId();
            this.released = 0;
        }

        /**
         * <p>
         * Returns the numeric identifier for this ID. The numeric identifier
         * can be used if the ID has to be passed to a component where this ID
         * object cannot be passed (e.g. over the network). However, the numeric
         * identifier might be reused for other identifiers in the future,
         * depending on the configuration of the ID pool this ID has been
         * acquired from.
         * </p>
         * 
         * <p>
         * Calling this method after calling {@link #release()} results in an
         * {@link IllegalStateException} being thrown.
         * 
         * @return numeric identifier for this ID.
         * @throws IllegalStateException
         *             if this ID has been released.
         */
        public int get() {
            if (isAlive()) {
                return id;
            } else {
                throw new IllegalStateException(
                        "This ID has already been released.");
            }
        }

        /**
         * <p>
         * Releases this ID, returning it to the pool it has been acquired from.
         * Calling this method multiple times is not an error, however, the ID
         * will only be released the first time this method is called.
         * Subsequent calls to this method have no effect.
         * </p>
         * 
         * <p>
         * After calling this method, {@link #isAlive()} will always return
         * <code>false</code> and {@link #get()} will always throw an
         * {@link IllegalStateException}.
         * </p>
         */
        public void release() {
            if (!isAlive()) {
                // ID has already been released.
                return;
            }
            if (atomicMarkAsReleased()) {
                // We successfully set the released flag to one. This means that
                // no other thread released or will release the ID.
                idPool.releaseId(id);
            }
        }

        /**
         * Returns whether this ID is still alive or has already been released.
         * 
         * @return <code>true</code> if this ID is still alive,
         *         <code>false</code> if it has been released by calling
         *         {@link #release()}.
         */
        public boolean isAlive() {
            return RELEASED_UPDATER.get(this) == 0;
        }

        /**
         * Atomically marks this ID as released.
         * 
         * @return <code>true</code> if this ID has not been released before and
         *         has now been marked as released. <code>false</code> if this
         *         has already been released before.
         */
        private boolean atomicMarkAsReleased() {
            return RELEASED_UPDATER.compareAndSet(this, 0, 1);
        }

        /**
         * Returns a string representation of this ID. The string representation
         * is the string representation of the numeric ID (as returned by
         * <code>get()</code>) or the string "&lt;released&gt;" if this ID has
         * already been released.
         * 
         * @return string representation of this ID.
         */
        @Override
        public String toString() {
            if (isAlive()) {
                return Integer.toString(id);
            } else {
                return "<released>";
            }
        }

        /**
         * <p>
         * Finalizes this object, releasing the ID. This method calls
         * {@link #release()} followed by <code>super.finalize()</code>. This
         * ensures that an ID is released when it is not used any longer.
         * However, the point in time at which <code>finalize()</code> is called
         * is non-deterministic. For this reason, the <code>release()</code>
         * method should be called explicitly whenever this is feasible.
         * </p>
         * 
         * <p>
         * From the documentation of <code>Object.finalize()</code>:
         * </p>
         * 
         * <p>
         * Called by the garbage collector on an object when garbage collection
         * determines that there are no more references to the object. A
         * subclass overrides the <code>finalize</code> method to dispose of
         * system resources or to perform other cleanup.
         * </p>
         * 
         * <p>
         * The general contract of <tt>finalize</tt> is that it is invoked if
         * and when the Java™ virtual machine has determined that there is no
         * longer any means by which this object can be accessed by any thread
         * that has not yet died, except as a result of an action taken by the
         * finalization of some other object or class which is ready to be
         * finalized. The <tt>finalize</tt> method may take any action,
         * including making this object available again to other threads; the
         * usual purpose of <tt>finalize</tt>, however, is to perform cleanup
         * actions before the object is irrevocably discarded. For example, the
         * finalize method for an object that represents an input/output
         * connection might perform explicit I/O transactions to break the
         * connection before the object is permanently discarded.
         * </p>
         * 
         * <p>
         * The Java programming language does not guarantee which thread will
         * invoke the <tt>finalize</tt> method for any given object. It is
         * guaranteed, however, that the thread that invokes finalize will not
         * be holding any user-visible synchronization locks when finalize is
         * invoked. If an uncaught exception is thrown by the finalize method,
         * the exception is ignored and finalization of that object terminates.
         * </p>
         * 
         * <p>
         * After the <tt>finalize</tt> method has been invoked for an object, no
         * further action is taken until the Java virtual machine has again
         * determined that there is no longer any means by which this object can
         * be accessed by any thread that has not yet died, including possible
         * actions by other objects or classes which are ready to be finalized,
         * at which point the object may be discarded.
         * </p>
         * 
         * <p>
         * The <tt>finalize</tt> method is never invoked more than once by a
         * Java virtual machine for any given object.
         * </p>
         * 
         * <p>
         * Any exception thrown by the <code>finalize</code> method causes the
         * finalization of this object to be halted, but is otherwise ignored.
         *
         * @throws Throwable
         *             the <code>Exception</code> raised by this method
         */
        @Override
        protected void finalize() throws Throwable {
            try {
                // If the ID has not been released explicitly, we want to
                // release it latest when it is garbage collected.
                release();
            } finally {
                super.finalize();
            }
        }

    }

    /**
     * The ID pool that actually manages the IDs.
     */
    private BasicIntegerIdPool idPool;

    /**
     * <p>
     * Creates a new pool for integer IDs. All IDs returned by the pools
     * {@link #acquireId()} method are guaranteed to be in the range from
     * <code>minimumId</code> to <code>maximumId</code> (both inclusive).
     * </p>
     * 
     * <p>
     * If <code>minimumId</code> is greater than <code>maximumId</code>, the
     * pool will start at <code>minimumId</code> counting upwards until
     * {@link Integer#MAX_VALUE} is reached. Then, the counter will wrap around
     * continuing with negative numbers until <code>maximumId</code> is reached.
     * </p>
     * 
     * <p>
     * <code>minimumId</code> and <code>maximumId</code> must be chosen so that
     * there is at least one integer number that is not covered by the range.
     * This number is needed because it will be internally used as a marker when
     * the counter has been exhausted.
     * </p>
     * 
     * <p>
     * The <code>blockFromReuseTime</code> parameter specifies the amount of
     * time in milliseconds that must pass between an ID being returned by
     * calling its {@link Id#release() release()} method and
     * {@link #acquireId()} returning the same ID. This is useful for
     * applications where an external entity might keep an ID for some time and
     * it is important to distinguish the now invalid ID from newly acquired
     * IDs.
     * </p>
     * 
     * <p>
     * If <code>earlyReuseOnExhaustion</code> is <code>true</code>,
     * {@link #acquireId()} will not fail immediately if no ID can be reused yet
     * and no new ID can be assigned because all possible IDs are already in
     * use. Instead of that, it will try to find an ID that has been released,
     * but for which the <code>blockFromReuseTime</code> has not passed yet.
     * This can be used as a last resort, when it is more important to make sure
     * that IDs can be acquired than it is important to ensure that released IDs
     * are not reused for a certain time.
     * </p>
     * 
     * <p>
     * Effectively, all parameters of this constructor are simply passed on to
     * the constructor
     * {@link BasicIntegerIdPool#BasicIntegerIdPool(int, int, long, boolean)}.
     * </p>
     * 
     * @param minimumId
     *            minimum number used as an ID (inclusive). If
     *            <code>minimumId</code> is greater than <code>maximumId</code>,
     *            IDs that are less than <code>minimumId</code> are used if they
     *            are also less than or equal to <code>maximumId</code>.
     * @param maximumId
     *            maximum number used as an ID (inclusive). If
     *            <code>maximumId</code> is less than <code>minimumId</code>,
     *            IDs that are greater than <code>maximumId</code> are used if
     *            they are also greater than or equal to <code>minimumId</code>.
     * @param blockFromReuseTime
     *            number of milliseconds to wait before reusing an ID after it
     *            has been released. A value of <code>0</code> means that IDs
     *            can be reused immediately.
     * @param earlyReuseOnExhaustion
     *            if <code>true</code> IDs which have been released less than
     *            <code>blockFromReuseTime</code> milliseconds ago are still
     *            used if otherwise the pool would be exhausted.
     * @throws IllegalArgumentException
     *             if <code>minimumId</code> and <code>maximumId</code> have not
     *             been chosen so that there is at least one unusable ID or if
     *             blockFromReuseTime is negative.
     */
    public IntegerIdPool(int minimumId, int maximumId, long blockFromReuseTime,
            boolean earlyReuseOnExhaustion) {
        this.idPool = new BasicIntegerIdPool(minimumId, maximumId,
                blockFromReuseTime, earlyReuseOnExhaustion);
    }

    /**
     * Acquires an ID that is currently not in use. The ID is marked as used so
     * that subsequent calls to this method will not return the ID. It will only
     * be used again after it has been released using its {@link Id#release()
     * release()} method.
     * 
     * @return unique ID. This ID should be returned to the pool by calling its
     *         {@link Id#release() release()} method when it is not needed any
     *         longer.
     * @throws IllegalStateException
     *             if this pool is exhausted and no unused ID can be assigned at
     *             the moment.
     */
    public Id acquireId() {
        return new Id(idPool);
    }

}
