/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

/**
 * Utility methods for converting between Java strings and the null-terminated
 * byte-sequences used over the wire.
 * 
 * @author Sebastian Marsching
 */
public final class NullTerminatedStringUtil {

    /**
     * Private constructor. This class should never be instantiated. All methods
     * are static.
     */
    private NullTerminatedStringUtil() {
    }

    /**
     * Converts a null-terminated part of an array of bytes to a string. This
     * method is suitable for converting a string received from the network to a
     * Java string. Only bytes that appear before the first null-byte are used.
     * If there is no null-byte, the behavior depends on the <code>strict</code>
     * option.
     * 
     * @param bytes
     *            array of bytes
     * @param offset
     *            offset into the <code>bytes</code> array. This method will
     *            start reading at this position.
     * @param length
     *            maximum number of bytes that should be read from the
     *            <code>bytes</code> array.
     * @param encoding
     *            character set to use for converting the bytes to a string.
     * @param strict
     *            if <code>true</code>, this method throws an
     *            IllegalArgumentException if the string is not null terminated
     *            (there is no null byte in the range denoted by
     *            <code>offset</code> and <code>length</code>. If
     *            <code>false</code>, all bytes are used for the string (an
     *            implicit null byte after the end of the byte range is
     *            assumed).
     * @return string representing the specified elements of the byte array that
     *         precede the first null-byte.
     * @throws IndexOutOfBoundsException
     *             if <code>offset</code> + <code>length</code> exceed the size
     *             of <code>bytes</code>.
     */
    public static String nullTerminatedBytesToString(byte[] bytes, int offset,
            int length, Charset encoding, boolean strict) {
        int end = offset + length;
        int usedLength = length;
        for (int i = offset; i < end; i++) {
            if (bytes[i] == 0) {
                usedLength = i - offset;
                break;
            }
        }
        if (strict && usedLength == length) {
            throw new IllegalArgumentException(
                    "The string is not null terminated.");
        }
        return new String(bytes, offset, usedLength, encoding);
    }

    /**
     * Converts a null-terminated array of bytes to a string. This method is
     * suitable for converting a string received from the network to a Java
     * string. Only bytes that appear before the first null-byte are used. If
     * there is no null-byte, the behavior depends on the <code>strict</code>
     * option.
     * 
     * @param bytes
     *            array of bytes
     * @param encoding
     *            character set to use for converting the bytes to a string.
     * @param strict
     *            if <code>true</code>, this method throws an
     *            IllegalArgumentException if the string is not null terminated.
     *            If <code>false</code>, all bytes are used for the string (an
     *            implicit null byte after the end of the byte array is
     *            assumed).
     * @return string representing the elements of the byte array that precede
     *         the first null-byte.
     */
    public static String nullTerminatedBytesToString(byte[] bytes,
            Charset encoding, boolean strict) {
        return nullTerminatedBytesToString(bytes, 0, bytes.length, encoding,
                strict);
    }

    /**
     * Converts a null-terminated array of bytes to a string. This method is
     * suitable for converting a string received from the network to a Java
     * string. Only bytes that appear before the first null-byte are used. If
     * there is no null-byte, all bytes are used.
     * 
     * @param bytes
     *            array of bytes
     * @param encoding
     *            character set to use for converting the bytes to a string.
     * @return string representing the elements of the byte array that precede
     *         the first null-byte.
     */
    public static String nullTerminatedBytesToString(byte[] bytes,
            Charset encoding) {
        return nullTerminatedBytesToString(bytes, encoding, false);
    }

    /**
     * Converts a string to an array of null-terminated bytes, truncating the
     * string if necessary. The resulting array is never longer than
     * <code>maximumLength</code> and always has at least one terminating
     * null-byte.
     * 
     * @param string
     *            the string that shall be converted to bytes.
     * @param maximumLength
     *            the maximum length of the resulting byte array (including the
     *            terminating null-byte).
     * @param alignTo
     *            number of bytes that the string length (including the
     *            terminating null-bytes) should be aligned to. This means that
     *            some additional null-bytes will be added to match the
     *            alignment if needed. The total length is still guaranteed to
     *            be less than or equal to <code>maximumLength</code>.
     * @param encoding
     *            the character set to be used for converting the string.
     * @return a pair of a byte array containing the null-terminated and
     *         truncated (if needed) string and an integer specifying the number
     *         of bytes that are needed to store the string (including the first
     *         null-byte).
     * @throws IllegalArgumentException
     *             if <code>maximumLength</code> is negative,
     *             <code>alignTo</code> is less than or equal to zero, or
     *             <code>maximumLength</code> is less than <code>alignTo</code>.
     */
    public static Pair<byte[], Integer> stringToTruncatedNullTerminatedBytes(
            String string, int maximumLength, int alignTo, Charset encoding) {
        if (maximumLength < 0) {
            throw new IllegalArgumentException(
                    "Maxium length must not be negative.");
        }
        if (alignTo <= 0) {
            throw new IllegalArgumentException("Alignment must be at least 1.");
        }
        if (maximumLength < alignTo) {
            throw new IllegalArgumentException(
                    "The maximum length must be greater than or equal to the requested alignment.");
        }
        // Make sure that the maximum length is aligned to the requested
        // alignment. There is no sense in using a non-aligned maximum length
        // because the extra bytes will not be used anyway.
        if (maximumLength % alignTo != 0) {
            maximumLength -= maximumLength % alignTo;
        }
        CharsetEncoder encoder = encoding.newEncoder();
        double maximumBytesLength = encoder.maxBytesPerChar() * string.length();
        int bytesLength;
        if (maximumBytesLength > (double) (maximumLength - 1)) {
            bytesLength = maximumLength;
        } else {
            bytesLength = ((int) Math.ceil(maximumBytesLength)) + 1;
        }
        // If bytesLength (the maximum number of bytes we might need) is less
        // than maximumLength (the limit for the total string size), we can
        // reduce maxiumLength so that we do not allocate an unnecessarily large
        // array. However, we still want to keep the size of this array aligned.
        if (bytesLength < maximumLength) {
            int paddingLength = alignTo - bytesLength % alignTo;
            // If bytesLength is already aligned, the result of modulo operation
            // is zero and thus paddingLength is alignTo.
            if (paddingLength == alignTo) {
                paddingLength = 0;
            }
            maximumLength = bytesLength + paddingLength;
        }
        // We want to be sure that bytesLength never exceed maximumLength,
        // even after casting.
        if (bytesLength > maximumLength) {
            bytesLength = maximumLength;
        }
        byte[] bytes = new byte[maximumLength];
        ByteBuffer buffer = ByteBuffer.wrap(bytes);
        char[] characters = string.toCharArray();
        int numberOfCharacters = characters.length;
        // In each (sensible) encoding, at least one byte is needed per
        // character. Therefore we can safely cut a long string.
        if (numberOfCharacters > maximumLength) {
            numberOfCharacters = maximumLength;
        }
        int maxPosition;
        try {
            CoderResult cr;
            cr = encoder.encode(
                    CharBuffer.wrap(characters, 0, numberOfCharacters), buffer,
                    true);
            if (!cr.isUnderflow() && !cr.isOverflow()) {
                cr.throwException();
            }
            if (cr.isUnderflow()) {
                cr = encoder.flush(buffer);
                if (!cr.isUnderflow() && !cr.isOverflow()) {
                    cr.throwException();
                }
            }
            // We have to safe the maximum position that we used, so that we can
            // later reset those bytes to zeros.
            maxPosition = buffer.position();
            // The string might have been too long and thus we might written
            // more bytes than we wanted. We continue by trying to write one
            // character less. We have to reserve one byte for the terminating
            // null-byte.
            while (cr.isOverflow() || buffer.remaining() == 0) {
                encoder.reset();
                buffer.clear();
                numberOfCharacters--;
                cr = encoder.encode(
                        CharBuffer.wrap(characters, 0, numberOfCharacters),
                        buffer, true);
                if (!cr.isUnderflow() && !cr.isOverflow()) {
                    cr.throwException();
                }
                if (cr.isUnderflow()) {
                    cr = encoder.flush(buffer);
                    if (!cr.isUnderflow() && !cr.isOverflow()) {
                        cr.throwException();
                    }
                }
            }
        } catch (CharacterCodingException e) {
            throw new Error(e);
        }
        int numberOfBytes = buffer.position() + 1;
        // If there have been more bytes in the buffer than there are now, we
        // have to overwrite the old bytes with zeros. All remaining bytes
        // should be zeros because the byte array is initialized with zeros on
        // construction.
        while (buffer.position() < maxPosition) {
            buffer.put((byte) 0);
        }
        // The returned number of bytes has to be aligned. Therefore we add some
        // more (null-)bytes if needed. We can do this, because we know that
        // numberOfBytes is less than or equal to maximumLength and
        // maximumLength is aligned.
        if (numberOfBytes % alignTo != 0) {
            numberOfBytes += alignTo - numberOfBytes % alignTo;
        }
        return new ImmutablePair<byte[], Integer>(bytes, numberOfBytes);
    }

    /**
     * Converts a string to an array of null-terminated bytes, truncating the
     * string if necessary. The resulting array is never longer than
     * <code>maximumLength</code> and always has at least one terminating
     * null-byte.
     * 
     * @param string
     *            the string that shall be converted to bytes.
     * @param maximumLength
     *            the maximum length of the resulting byte array (including the
     *            terminating null-byte).
     * @param encoding
     *            the character set to be used for converting the string.
     * @return a pair of a byte array containing the null-terminated and
     *         truncated (if needed) string and an integer specifying the number
     *         of bytes that are needed to store the string (including the first
     *         null-byte).
     * @throws IllegalArgumentException
     *             if <code>maximumLength</code> is less than or equal to zero.
     */
    public static Pair<byte[], Integer> stringToTruncatedNullTerminatedBytes(
            String string, int maximumLength, Charset encoding) {
        return stringToTruncatedNullTerminatedBytes(string, maximumLength, 1,
                encoding);
    }

    /**
     * Converts a string to an array of null-terminated bytes, truncating the
     * string if necessary. The resulting array has exactly
     * <code>fixedLength</code> bytes. If the string is shorter than the space
     * available in the array, the remaining space is filled with null bytes.
     * 
     * @param string
     *            the string that shall be converted to bytes.
     * @param fixedLength
     *            the length of the resulting byte array (including the
     *            terminating null-byte).
     * @param encoding
     *            the character set to be used for converting the string.
     * @return a byte array containing the null-terminated and truncated (if
     *         needed) string
     * @throws IllegalArgumentException
     *             if <code>fixedLength</code> is less than or equal to zero.
     */
    public static byte[] stringToTruncatedNullTerminatedFixedBytes(
            String string, int fixedLength, Charset encoding) {
        return stringToTruncatedNullTerminatedBytes(string, fixedLength,
                fixedLength, encoding).getLeft();
    }

    /**
     * Converts a string to an array of null-terminated bytes. The resulting
     * array is never longer than <code>maximumLength</code> and always has at
     * least one terminating null-byte. If the string is too long to fit into
     * <code>maximumLength</code> bytes, <code>null</code> is returned.
     * 
     * @param string
     *            the string that shall be converted to bytes.
     * @param maximumLength
     *            the maximum length of the resulting byte array (including the
     *            terminating null-byte).
     * @param alignTo
     *            number of bytes that the string length (including the
     *            terminating null-bytes) should be aligned to. This means that
     *            some additional null-bytes will be added to match the
     *            alignment if needed. The total length is still guaranteed to
     *            be less than or equal to <code>maximumLength</code>.
     * @param encoding
     *            the character set to be used for converting the string.
     * @return a pair of a byte array containing the null-terminated string and
     *         an integer specifying the number of bytes that are needed to
     *         store the string (including the first null-byte). The return
     *         value is <code>null</code> if the byte representation of the
     *         string (including one terminating null-byte) is longer than
     *         <code>maximumLength</code>.
     * @throws IllegalArgumentException
     *             if <code>maximumLength</code> is negative or
     *             <code>alignTo</code> is less than or equal to zero.
     */
    public static Pair<byte[], Integer> stringToSizeLimitedNullTerminatedBytesOrNull(
            String string, int maximumLength, int alignTo, Charset encoding) {
        if (maximumLength < 0) {
            throw new IllegalArgumentException(
                    "Maxium length must not be negative.");
        }
        // Make sure that the maximum length is aligned to requested alignment.
        // There is no sense in use a non-aligned maximum length because the
        // extra bytes will not be used anyway.
        if (maximumLength % alignTo != 0) {
            maximumLength -= maximumLength % alignTo;
        }
        if (alignTo <= 0) {
            throw new IllegalArgumentException("Alignment must be at least 1.");
        }
        CharsetEncoder encoder = encoding.newEncoder();
        double maximumBytesLength = encoder.maxBytesPerChar() * string.length();
        int bytesLength;
        if (maximumBytesLength > (double) (maximumLength - 1)) {
            bytesLength = maximumLength;
        } else {
            bytesLength = ((int) Math.ceil(maximumBytesLength)) + 1;
        }
        // If bytesLength (the maximum number of bytes we might need) is less
        // than maximumLength (the limit for the total string size), we can
        // reduce maxiumLength so that we do not allocate an unnecessarily large
        // array. However, we still want to keep the size of this array aligned.
        if (bytesLength < maximumLength) {
            int paddingLength = alignTo - bytesLength % alignTo;
            // If bytesLength is already aligned, the result of modulo operation
            // is zero and thus paddingLength is alignTo.
            if (paddingLength == alignTo) {
                paddingLength = 0;
            }
            maximumLength = bytesLength + paddingLength;
        }
        // We want to be sure that bytesLength never exceed maximumLength,
        // even after casting.
        if (bytesLength > maximumLength) {
            bytesLength = maximumLength;
        }
        byte[] bytes = new byte[maximumLength];
        ByteBuffer buffer = ByteBuffer.wrap(bytes);
        char[] characters = string.toCharArray();
        try {
            CoderResult cr;
            cr = encoder.encode(
                    CharBuffer.wrap(characters, 0, characters.length), buffer,
                    true);
            if (cr.isOverflow()) {
                // The string is too long.
                return null;
            }
            if (!cr.isUnderflow()) {
                cr.throwException();
            }
            cr = encoder.flush(buffer);
            if (cr.isOverflow()) {
                // The string is too long.
                return null;
            }
            if (!cr.isUnderflow()) {
                cr.throwException();
            }
            // We have to make sure that there is enough space left for at least
            // one null byte.
            if (buffer.remaining() == 0) {
                // There is no space remaining in the buffer. This means that we
                // filled the buffer completely and there is no trailing
                // null-byte, thus the string is too long.
                return null;
            }
        } catch (CharacterCodingException e) {
            throw new Error(e);
        }
        // We return the size of the string including the terminating null-byte.
        // This is equal to the current buffer position plus one.
        // The returned number of bytes has to be aligned. Therefore we add some
        // more (null-)bytes if needed. We can d this, because we know that
        // numberOfBytes is less than or equal to maximumLength and
        // maximumLength is aligned.
        int numberOfBytes = buffer.position() + 1;
        if (numberOfBytes % alignTo != 0) {
            numberOfBytes += alignTo - numberOfBytes % alignTo;
        }
        return new ImmutablePair<byte[], Integer>(bytes, numberOfBytes);
    }

    /**
     * Converts a string to an array of null-terminated bytes. The resulting
     * array is never longer than <code>maximumLength</code> and always has at
     * least one terminating null-byte. If the string is too long to fit into
     * <code>maximumLength</code> bytes, <code>null</code> is returned.
     * 
     * @param string
     *            the string that shall be converted to bytes.
     * @param maximumLength
     *            the maximum length of the resulting byte array (including the
     *            terminating null-byte).
     * @param encoding
     *            the character set to be used for converting the string.
     * @return a pair of a byte array containing the null-terminated string and
     *         an integer specifying the number of bytes that are needed to
     *         store the string (including the first null-byte). The return
     *         value is <code>null</code> if the byte representation of the
     *         string (including one terminating null-byte) is longer than
     *         <code>maximumLength</code>.
     * @throws IllegalArgumentException
     *             if <code>maximumLength</code> is negative.
     */
    public static Pair<byte[], Integer> stringToSizeLimitedNullTerminatedBytesOrNull(
            String string, int maximumLength, Charset encoding) {
        return stringToSizeLimitedNullTerminatedBytesOrNull(string,
                maximumLength, 1, encoding);
    }

}
