/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * User name utility. Provides a method for determining the login name of the
 * user that is running the program.
 * 
 * @author Sebastian Marsching
 */
public final class UserNameUtil {

    private static final Object SYSTEM_OBJECT;
    private static final Method USER_NAME_METHOD;

    static {
        Class<?> systemClass = null;
        Object systemObject = null;
        Method method = null;
        // We try to get an instance of the right class for the platform we are
        // running on. We catch all exceptions because we do not want any error
        // to block the initialization of this class.
        try {
            try {
                systemClass = Class
                        .forName("com.sun.security.auth.module.UnixSystem");
                method = systemClass.getMethod("getUsername");
                systemObject = systemClass.newInstance();
            } catch (ClassNotFoundException | LinkageError e) {
                // If the class cannot be found or loaded, we just try the next
                // one.
            }
            if (systemObject == null) {
                try {
                    systemClass = Class
                            .forName("com.sun.security.auth.module.NTSystem");
                    method = systemClass.getMethod("getName");
                    systemObject = systemClass.newInstance();
                } catch (ClassNotFoundException | LinkageError e) {
                    // If the class cannot be found or loaded, we just try the
                    // next one.
                }
            }
            if (systemObject == null) {
                try {
                    systemClass = Class
                            .forName("com.sun.security.auth.module.SolarisSystem");
                    method = systemClass.getMethod("getUsername");
                    systemObject = systemClass.newInstance();
                } catch (ClassNotFoundException | LinkageError e) {
                    // If the class cannot be found or loaded, this is not an
                    // error. It just means that we cannot use this service to
                    // determine the user name.
                }
            }
        } catch (IllegalAccessException e) {
        } catch (InstantiationException e) {
        } catch (NoSuchMethodException e) {
        } catch (SecurityException e) {
        }
        // Either both variables should be null or both should be non-null. We
        // also check the return type of the method so that we do not get a
        // ClassCastException later.
        if (systemObject != null && method != null
                && String.class.isAssignableFrom(method.getReturnType())) {
            SYSTEM_OBJECT = systemObject;
            USER_NAME_METHOD = method;
        } else {
            SYSTEM_OBJECT = null;
            USER_NAME_METHOD = null;
        }
    }

    private UserNameUtil() {
    }

    /**
     * Returns the user name of the user that is running the program. The user
     * name is determined by using platform-specific classes from the JAAS API.
     * If this fails (e.g. because the platform is not supported or the JRE does
     * not provide those classes), the "user.name" system property is used. If
     * this system property is not set, the "USERNAME" environment variable is
     * used. If the environment variable is not set either, <code>null</code> is
     * returned.
     * 
     * @return user name of the user that is running the program or
     *         <code>null</code> if the user name cannot be determined.
     */
    public static String getUserName() {
        String userName;
        // First, we try to use a system-specific class from JAAS. This might
        // fail, because none of the classes might be available (e.g. for a JRE
        // that is not from Oracle).
        userName = userNameFromSystem();
        if (userName != null) {
            return userName;
        }
        // Next, we try the "user.name" system property. This should work on
        // virtually all systems, but the information might be inaccurate.
        userName = System.getProperty("user.name");
        if (userName != null) {
            return userName;
        }
        // If the "user.name" system property is not set, we try the "USERNAME"
        // environment variable. If this variable is not set either, we return
        // null.
        userName = System.getenv("USERNAME");
        return userName;
    }

    private static String userNameFromSystem() {
        // If the SYSTEM_OBJECT and USER_NAME_METHOD have been initialized, we
        // try to use them to determine the user name. If this fails, we return
        // null because we want to use a different method.
        if (SYSTEM_OBJECT != null && USER_NAME_METHOD != null) {
            try {
                return (String) USER_NAME_METHOD.invoke(SYSTEM_OBJECT);
            } catch (IllegalAccessException e) {
            } catch (IllegalArgumentException e) {
            } catch (InvocationTargetException e) {
            }
        }
        return null;
    }

}
