/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.nio.charset.Charset;
import java.util.Collection;
import java.util.List;

/**
 * <p>
 * Channel Access value that represents the
 * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} type. In its
 * internal representation, this type is identical to the
 * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} type. However, this type
 * has a very specific use. It can only be used in get (read) operations and is
 * a way to query for the type of the record associated with a channel.
 * </p>
 * 
 * <p>
 * For all practical purposes, this type should store exactly one element.
 * However, the protocol technically allows for a different number of elements.
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @author Sebatian Marsching
 */
public interface ChannelAccessClassName extends
        ChannelAccessGettableValue<String> {

    /**
     * <p>
     * Returns the list storing the actual value elements. The list is returned
     * by reference, so subsequent changes to the list will affect this value
     * and vice-versa. Therefore, the list may not be modified after this value
     * has been passed to a different component. If this value is read-only, the
     * list returned is also read-only.
     * </p>
     * 
     * <p>
     * The list returned does not allow the insertion of <code>null</code>
     * elements.
     * </p>
     * 
     * @return list storing this value's elements (never <code>null</code>).
     *         Elements of the list are never <code>null</code>.
     * @see #getValueSize()
     */
    List<String> getValue();

    /**
     * <p>
     * Replaces this value's elements with the elements from the specified
     * collection. This has the same effect as calling <code>clear()</code>
     * followed by <code>addAll(value)</code> on the list returned by
     * {@link #getValue()}.
     * </p>
     * 
     * <p>
     * The string's raw representation is updated automatically. Strings that
     * are truncated while being converted to their raw representation still
     * maintain their original length in the list of element strings.
     * </p>
     * 
     * @param value
     *            collection of strings that should replace this value's
     *            elements.
     * @throws NullPointerException
     *             if <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setValue(Collection<String> value);

    /**
     * <p>
     * Returns the raw representation of the strings. This an array of N * 40
     * bytes, where N is the number of strings (as returned by
     * {@link #getValueSize()}). Each string in this array has exactly 40 bytes
     * and is null-terminated. This means that the string with the zero-based
     * index i starts at an array index of i * 40.
     * </p>
     * 
     * <p>
     * The array returned is a copy, so modifications to it will not affect this
     * value.
     * </p>
     * 
     * @return raw representation of the string values (never <code>null</code>
     *         ).
     * @see #getValue()
     * @see #getValueSize()
     * @see #getCharset()
     */
    byte[] getRawValue();

    /**
     * <p>
     * Updates the string values using the specified raw representations. The
     * size of the array passed must be an integer multiple of 40. Each sequence
     * of 40 bytes is interpreted as one null-terminated string. The strings
     * must be null-terminated, therefore the size of an individual string must
     * not exceed 39 characters (not including the terminating null-byte).
     * </p>
     * 
     * <p>
     * The array is copied, thus subsequent modifications of the array will not
     * affect this value. The {@link String Java strings} returned by
     * {@link #getValue()} are updated by converting the specified raw strings
     * using the specified {@link #getCharset() encoding}.
     * </p>
     * 
     * @param rawValue
     *            new raw representation of the strings (never <code>null</code>
     *            ). The size of the array must be a multiple of 40 and strings
     *            must be null terminated.
     * @throws IllegalArgumentException
     *             if the size of the array is not an integer multiple of 40 or
     *             the individual strings are not null terminated.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setRawValue(byte[] rawValue);

    /**
     * Returns the encoding that is used for converting {@linkplain String Java
     * strings} into raw (null-terminated) strings and vice-versa. The encoding
     * is defined when a value instance is created and never changes.
     * 
     * @return encoding used for encoding and decoding strings (never
     *         <code>null</code>).
     */
    Charset getCharset();

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessClassName asReadOnlyValue();

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessClassName clone();

}
