/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSink.AtomicPutOperation;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.io.ByteSource.AtomicGetOperation;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessAcknowledgeAlarmImpl;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessAlarmAcknowledgementStatusImpl;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessCharBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessClassNameImpl;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessConfigureAcknowledgeTransientAlarmsImpl;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessDoubleBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessEnumBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessFloatBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessLongBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessShortBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessStringBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessValueBase;

/**
 * <p>
 * Encoder and decoder for {@link ChannelAccessValue}s.
 * </p>
 * 
 * <p>
 * Values can be encoded (serialized) using one of the <code>encode</code>
 * methods. They can be decoded (deserialized) using one of the
 * <code>decode</code> methods.
 * </p>
 * 
 * <p>
 * The {@link #calculatePayloadSize(ChannelAccessValueType, int)} method can be
 * used to determine the space the serialized form of a value will need in a
 * Channel Access message. The
 * {@link #calculateMaxCount(ChannelAccessValueType, int)} method can be used to
 * determine the maximum number of elements that will fit into a message of a
 * limited size.
 * </p>
 * 
 * <p>
 * Encoding and decoding is not completely symmetrical: Decoding and then
 * encoding a binary representation of a value should result in the same binary
 * representation. However, encoding and then decoding a value might result in a
 * different value because due to limitations of the Channel Access protocol
 * some information (in particular long strings) cannot be transported.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessValueCodec {

    /**
     * Size for the fields of a type that do not depend on the number of
     * elements. In order to get the minimum size for a type, the size of a
     * single element has to be added to this value, because at least one
     * element must be present in the serialized representation (even if the
     * count is zero).
     */
    private static final int[] BASE_SIZE = new int[] { 0, // DBR_STRING
            0, // DBR_SHORT
            0, // DBR_FLOAT
            0, // DBR_ENUM
            0, // DBR_CHAR
            0, // DBR_LONG
            0, // DBR_DOUBLE
            4, // DBR_STS_STRING
            4, // DBR_STS_SHORT
            4, // DBR_STS_FLOAT
            4, // DBR_STS_ENUM
            5, // DBR_STS_CHAR
            4, // DBR_STS_LONG
            8, // DBR_STS_DOUBLE
            12, // DBR_TIME_STRING
            14, // DBR_TIME_SHORT
            12, // DBR_TIME_FLOAT
            14, // DBR_TIME_ENUM
            15, // DBR_TIME_CHAR
            12, // DBR_TIME_LONG
            16, // DBR_TIME_DOUBLE
            4, // DBR_GR_STRING
            24, // DBR_GR_SHORT
            40, // DBR_GR_FLOAT
            422, // DBR_GR_ENUM
            19, // DBR_GR_CHAR
            36, // DBR_GR_LONG
            64, // DBR_GR_DOUBLE
            4, // DBR_CTRL_STRING
            28, // DBR_CTRL_SHORT
            48, // DBR_CTRL_FLOAT
            422, // DBR_CTRL_ENUM
            21, // DBR_CTRL_CHAR
            44, // DBR_CTRL_LONG
            80, // DBR_CTRL_DOUBLE
            0, // DBR_PUT_ACKT
            0, // DBR_PUT_ACKS
            8, // DBR_STSACK_STRING
            0 // DBR_CLASS_NAME
    };

    /**
     * Size of the per-element field of a type. This number should be multiplied
     * with the number of elements and added to the base size in order to get
     * the total size.
     */
    private static final int[] ELEMENT_SIZE = new int[] { 40, // DBR_STRING
            2, // DBR_SHORT
            4, // DBR_FLOAT
            2, // DBR_ENUM
            1, // DBR_CHAR
            4, // DBR_LONG
            8, // DBR_DOUBLE
            40, // DBR_STS_STRING
            2, // DBR_STS_SHORT
            4, // DBR_STS_FLOAT
            2, // DBR_STS_ENUM
            1, // DBR_STS_CHAR
            4, // DBR_STS_LONG
            8, // DBR_STS_DOUBLE
            40, // DBR_TIME_STRING
            2, // DBR_TIME_SHORT
            4, // DBR_TIME_FLOAT
            2, // DBR_TIME_ENUM
            1, // DBR_TIME_CHAR
            4, // DBR_TIME_LONG
            8, // DBR_TIME_DOUBLE
            40, // DBR_GR_STRING
            2, // DBR_GR_SHORT
            4, // DBR_GR_FLOAT
            2, // DBR_GR_ENUM
            1, // DBR_GR_CHAR
            4, // DBR_GR_LONG
            8, // DBR_GR_DOUBLE
            40, // DBR_CTRL_STRING
            2, // DBR_CTRL_SHORT
            4, // DBR_CTRL_FLOAT
            2, // DBR_CTRL_ENUM
            1, // DBR_CTRL_CHAR
            4, // DBR_CTRL_LONG
            8, // DBR_CTRL_DOUBLE
            2, // DBR_PUT_ACKT
            2, // DBR_PUT_ACKS
            40, // DBR_STSACK_STRING
            40 // DBR_CLASS_NAME
    };

    /**
     * Maximum supported number of elements for each type. We precalculate this
     * limits because we need them very often.
     */
    private static final int[] MAX_COUNT;
    static {
        ChannelAccessValueType[] allTypes = ChannelAccessValueType.values();
        MAX_COUNT = new int[allTypes.length];
        for (ChannelAccessValueType type : ChannelAccessValueType.values()) {
            int typeNumber = type.toTypeCode();
            int baseSize = BASE_SIZE[typeNumber];
            int elementSize = ELEMENT_SIZE[typeNumber];
            MAX_COUNT[typeNumber] = (Integer.MAX_VALUE - baseSize)
                    / elementSize;
        }
    }

    /**
     * Private constructor. This class only has static methods, so there is no
     * reason to create an instance.
     */
    private ChannelAccessValueCodec() {
    }

    /**
     * Decodes a gettable ChannelAccess value from its serialized form. Gettable
     * values are values that can be used in get operations (sending data from a
     * server to a client). If this method throws an exception, the byte source
     * will remain in a state as if no data had been read.
     * 
     * @param byteSource
     *            byte source for reading the value's date.
     * @param type
     *            type of the value (never <code>null</code>).
     * @param dataSize
     *            size of the available data (in bytes). If this is greater than
     *            the amount of data needed for a value of the specified type
     *            with the specified number of elements, the extra data is read
     *            and discarded. If this is less than the data needed for a
     *            value of the specified type with the specified number of
     *            elements, an {@link IllegalArgumentException} is thrown.
     * @param count
     *            number of elements in the value. Must not be negative.
     * @param charset
     *            character-set to use for decoding strings (never
     *            <code>null</code>).
     * @return value of the specified <code>type</code>, created by reading
     *         <code>dataSize</code> bytes from <code>byteSource</code>.
     * @throws BufferUnderflowException
     *             if the byteSource provides less than <code>dataSize</code>
     *             bytes of data and thus throws a
     *             <code>BufferUnderflowException</code>.
     * @throws IllegalArgumentException
     *             if <code>count</code> is negative, if <code>dataSize</code>
     *             is too small for the specified type and number of elements,
     *             and if <code>type</code> is not a gettable type.
     */
    public static ChannelAccessGettableValue<?> decodeGettableValue(
            ByteSource byteSource, ChannelAccessValueType type, int dataSize,
            int count, Charset charset) {
        if (type == ChannelAccessValueType.DBR_PUT_ACKT
                || type == ChannelAccessValueType.DBR_PUT_ACKS) {
            throw new IllegalArgumentException("Type " + type
                    + " is not gettable.");
        } else {
            return (ChannelAccessGettableValue<?>) decodeValue(byteSource,
                    type, dataSize, count, charset);
        }
    }

    /**
     * Decodes a puttable ChannelAccess value from its serialized form. Puttable
     * values are values that can be used in put operations (sending data from a
     * client to a server). If this method throws an exception, the byte source
     * will remain in a state as if no data had been read.
     * 
     * @param byteSource
     *            byte source for reading the value's date.
     * @param type
     *            type of the value (never <code>null</code>).
     * @param dataSize
     *            size of the available data (in bytes). If this is greater than
     *            the amount of data needed for a value of the specified type
     *            with the specified number of elements, the extra data is read
     *            and discarded. If this is less than the data needed for a
     *            value of the specified type with the specified number of
     *            elements, an {@link IllegalArgumentException} is thrown.
     * @param count
     *            number of elements in the value. Must not be negative.
     * @param charset
     *            character-set to use for decoding strings (never
     *            <code>null</code>).
     * @return value of the specified <code>type</code>, created by reading
     *         <code>dataSize</code> bytes from <code>byteSource</code>.
     * @throws BufferUnderflowException
     *             if the byteSource provides less than <code>dataSize</code>
     *             bytes of data and thus throws a
     *             <code>BufferUnderflowException</code>.
     * @throws IllegalArgumentException
     *             if <code>count</code> is negative, if <code>dataSize</code>
     *             is too small for the specified type and number of elements,
     *             and if <code>type</code> is not a puttable type.
     */
    public static ChannelAccessPuttableValue<?> decodePuttableValue(
            ByteSource byteSource, ChannelAccessValueType type, int dataSize,
            int count, Charset charset) {
        if (type.isSimpleType() || type == ChannelAccessValueType.DBR_PUT_ACKT
                || type == ChannelAccessValueType.DBR_PUT_ACKS) {
            return (ChannelAccessPuttableValue<?>) decodeValue(byteSource,
                    type, dataSize, count, charset);
        } else {
            throw new IllegalArgumentException("Type " + type
                    + " is not puttable.");
        }
    }

    /**
     * Decodes a ChannelAccess value from its serialized form. If this method
     * throws an exception, the byte source will remain in a state as if no data
     * had been read.
     * 
     * @param byteSource
     *            byte source for reading the value's date.
     * @param type
     *            type of the value (never <code>null</code>).
     * @param dataSize
     *            size of the available data (in bytes). If this is greater than
     *            the amount of data needed for a value of the specified type
     *            with the specified number of elements, the extra data is read
     *            and discarded. If this is less than the data needed for a
     *            value of the specified type with the specified number of
     *            elements, an {@link IllegalArgumentException} is thrown.
     * @param count
     *            number of elements in the value. Must not be negative.
     * @param charset
     *            character-set to use for decoding strings (never
     *            <code>null</code>).
     * @return value of the specified <code>type</code>, created by reading
     *         <code>dataSize</code> bytes from <code>byteSource</code>.
     * @throws BufferUnderflowException
     *             if the byteSource provides less than <code>dataSize</code>
     *             bytes of data and thus throws a
     *             <code>BufferUnderflowException</code>.
     * @throws IllegalArgumentException
     *             if <code>count</code> is negative and if
     *             <code>dataSize</code> is too small for the specified type and
     *             number of elements.
     */
    public static ChannelAccessValue<?> decodeValue(
            final ByteSource byteSource, final ChannelAccessValueType type,
            final int dataSize, final int count, final Charset charset) {
        if (count < 0) {
            throw new IllegalArgumentException(
                    "Number of expected elements must not be negative.");
        }
        int valueSize = calculatePayloadSize(type, count);
        if (dataSize < valueSize) {
            if (count == 1
                    && type.toSimpleType().equals(ChannelAccessValueType.DBR_STRING)
                    && dataSize >= calculatePayloadSize(type, 0)) {
                valueSize = dataSize;
            } else {
                throw new IllegalArgumentException("A value of type " + type
                        + " needs more than " + dataSize + " bytes.");
            }
        }
        final int valueSizeFinal = valueSize;
        return byteSource
                .atomicGet(new AtomicGetOperation<ChannelAccessValue<?>>() {
                    @Override
                    public ChannelAccessValue<?> get() {
                        ChannelAccessValue<?> value = decodeInternal(
                                byteSource, valueSizeFinal, type, count,
                                charset);
                        // Discard any extra bytes.
                        byteSource.skip(dataSize - valueSizeFinal);
                        return value;
                    }
                });
    }

    /**
     * Reads a Channel Access value from a byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @param numberOfBytes
     *            number of bytes that contain the value and shall be read from
     *            the byte source.
     * @param type
     *            type of the value to be read.
     * @param count
     *            number of elements the value has.
     * @param charset
     *            character-set to use for decoding strings (never
     *            <code>null</code>).
     * @return value of the specified type constructed from the data in the byte
     *         source.
     * @throws BufferUnderflowException
     *             if not enough data can be read.
     */
    private static ChannelAccessValue<?> decodeInternal(ByteSource byteSource,
            int numberOfBytes, ChannelAccessValueType type, int count,
            Charset charset) {
        switch (type) {
        case DBR_STRING:
            return ChannelAccessStringBase.ChannelAccessStringImpl.deserialize(
                    byteSource, numberOfBytes, count, charset);
        case DBR_SHORT:
            return ChannelAccessShortBase.ChannelAccessShortImpl.deserialize(
                    byteSource, count);
        case DBR_FLOAT:
            return ChannelAccessFloatBase.ChannelAccessFloatImpl.deserialize(
                    byteSource, count);
        case DBR_ENUM:
            return ChannelAccessEnumBase.ChannelAccessEnumImpl.deserialize(
                    byteSource, count);
        case DBR_CHAR:
            return ChannelAccessCharBase.ChannelAccessCharImpl.deserialize(
                    byteSource, count);
        case DBR_LONG:
            return ChannelAccessLongBase.ChannelAccessLongImpl.deserialize(
                    byteSource, count);
        case DBR_DOUBLE:
            return ChannelAccessDoubleBase.ChannelAccessDoubleImpl.deserialize(
                    byteSource, count);
        case DBR_STS_STRING:
            return ChannelAccessStringBase.ChannelAccessAlarmStringImpl
                    .deserialize(type, byteSource, numberOfBytes, count,
                            charset);
        case DBR_STS_SHORT:
            return ChannelAccessShortBase.ChannelAccessAlarmShortImpl
                    .deserialize(byteSource, count);
        case DBR_STS_FLOAT:
            return ChannelAccessFloatBase.ChannelAccessAlarmFloatImpl
                    .deserialize(byteSource, count);
        case DBR_STS_ENUM:
            return ChannelAccessEnumBase.ChannelAccessAlarmEnumImpl
                    .deserialize(byteSource, count);
        case DBR_STS_CHAR:
            return ChannelAccessCharBase.ChannelAccessAlarmCharImpl
                    .deserialize(byteSource, count);
        case DBR_STS_LONG:
            return ChannelAccessLongBase.ChannelAccessAlarmLongImpl
                    .deserialize(byteSource, count);
        case DBR_STS_DOUBLE:
            return ChannelAccessDoubleBase.ChannelAccessAlarmDoubleImpl
                    .deserialize(byteSource, count);
        case DBR_TIME_STRING:
            return ChannelAccessStringBase.ChannelAccessTimeStringImpl
                    .deserialize(byteSource, numberOfBytes, count, charset);
        case DBR_TIME_SHORT:
            return ChannelAccessShortBase.ChannelAccessTimeShortImpl
                    .deserialize(byteSource, count);
        case DBR_TIME_FLOAT:
            return ChannelAccessFloatBase.ChannelAccessTimeFloatImpl
                    .deserialize(byteSource, count);
        case DBR_TIME_ENUM:
            return ChannelAccessEnumBase.ChannelAccessTimeEnumImpl.deserialize(
                    byteSource, count);
        case DBR_TIME_CHAR:
            return ChannelAccessCharBase.ChannelAccessTimeCharImpl.deserialize(
                    byteSource, count);
        case DBR_TIME_LONG:
            return ChannelAccessLongBase.ChannelAccessTimeLongImpl.deserialize(
                    byteSource, count);
        case DBR_TIME_DOUBLE:
            return ChannelAccessDoubleBase.ChannelAccessTimeDoubleImpl
                    .deserialize(byteSource, count);
        case DBR_GR_STRING:
            return ChannelAccessStringBase.ChannelAccessAlarmStringImpl
                    .deserialize(type, byteSource, numberOfBytes, count,
                            charset);
        case DBR_GR_SHORT:
            return ChannelAccessShortBase.ChannelAccessGraphicsShortImpl
                    .deserialize(byteSource, count, charset);
        case DBR_GR_FLOAT:
            return ChannelAccessFloatBase.ChannelAccessGraphicsFloatImpl
                    .deserialize(byteSource, count, charset);
        case DBR_GR_ENUM:
            return ChannelAccessEnumBase.ChannelAccessGraphicsEnumImpl
                    .deserialize(type, byteSource, count, charset);
        case DBR_GR_CHAR:
            return ChannelAccessCharBase.ChannelAccessGraphicsCharImpl
                    .deserialize(byteSource, count, charset);
        case DBR_GR_LONG:
            return ChannelAccessLongBase.ChannelAccessGraphicsLongImpl
                    .deserialize(byteSource, count, charset);
        case DBR_GR_DOUBLE:
            return ChannelAccessDoubleBase.ChannelAccessGraphicsDoubleImpl
                    .deserialize(byteSource, count, charset);
        case DBR_CTRL_STRING:
            return ChannelAccessStringBase.ChannelAccessAlarmStringImpl
                    .deserialize(type, byteSource, numberOfBytes, count,
                            charset);
        case DBR_CTRL_SHORT:
            return ChannelAccessShortBase.ChannelAccessControlsShortImpl
                    .deserialize(byteSource, count, charset);
        case DBR_CTRL_FLOAT:
            return ChannelAccessFloatBase.ChannelAccessControlsFloatImpl
                    .deserialize(byteSource, count, charset);
        case DBR_CTRL_ENUM:
            return ChannelAccessEnumBase.ChannelAccessGraphicsEnumImpl
                    .deserialize(type, byteSource, count, charset);
        case DBR_CTRL_CHAR:
            return ChannelAccessCharBase.ChannelAccessControlsCharImpl
                    .deserialize(byteSource, count, charset);
        case DBR_CTRL_LONG:
            return ChannelAccessLongBase.ChannelAccessControlsLongImpl
                    .deserialize(byteSource, count, charset);
        case DBR_CTRL_DOUBLE:
            return ChannelAccessDoubleBase.ChannelAccessControlsDoubleImpl
                    .deserialize(byteSource, count, charset);
        case DBR_PUT_ACKT:
            return ChannelAccessConfigureAcknowledgeTransientAlarmsImpl
                    .deserialize(byteSource, count);
        case DBR_PUT_ACKS:
            return ChannelAccessAcknowledgeAlarmImpl.deserialize(byteSource,
                    count);
        case DBR_STSACK_STRING:
            return ChannelAccessAlarmAcknowledgementStatusImpl.deserialize(
                    byteSource, numberOfBytes, count, charset);
        case DBR_CLASS_NAME:
            return ChannelAccessClassNameImpl.deserialize(byteSource,
                    numberOfBytes, count, charset);
        default:
            throw new RuntimeException("Unhandled type " + type);
        }
    }

    /**
     * <p>
     * Encodes a gettable ChannelAccess value into its serialized form. Gettable
     * values are values that can be used in get operations (sending data from a
     * server to a client). If strings have to be serialized, the value's
     * encoding as specified by the return value of its
     * <code>getCharset()</code> method is used.
     * </p>
     * 
     * <p>
     * The specified value must have been created by
     * {@link ChannelAccessValueCodec} or {@link ChannelAccessValueFactory},
     * because this method expects a certain implementation. If the
     * implementation is not the expected one, a {@link ClassCastException} is
     * thrown.
     * </p>
     * 
     * <p>
     * The serialization is limited to the specified number elements, even if
     * the value contains more elements. Therefore, the number of bytes written
     * is always equal to the number returned by
     * {@link #calculatePayloadSize(ChannelAccessValueType, int)}.
     * </p>
     * 
     * <p>
     * This method throws an {@link BufferOverflowException} if there is not
     * enough space left in the byte sink to write the complete value. If this
     * method throws an exception, the byte sink will remain in a state as if no
     * data had been written.
     * </p>
     * 
     * <p>
     * This method simply delegates to
     * {@link #encodeValue(ByteSink, ChannelAccessValue, int)} and is mainly
     * provided for symmetry with
     * {@link #decodeGettableValue(ByteSource, ChannelAccessValueType, int, int, Charset)}
     * .
     * </p>
     * 
     * @param byteSink
     *            byte sink the value's data is written to (never
     *            <code>null</code>).
     * @param value
     *            value that should be serialized to the byte sink (never
     *            <code>null</code>).
     * @param count
     *            number of elements that should be serialized. This number must
     *            not be negative and must be less than or equal to the number
     *            of elements that the value has.
     * @throws BufferOverflowException
     *             if the value cannot be written to the byte source completely,
     *             because it throws a {@link BufferOverflowException}.
     * @throws ClassCastException
     *             if the specified <code>value</code> has not been created by
     *             {@link ChannelAccessValueCodec} or
     *             {@link ChannelAccessValueFactory} and thus is an
     *             implementation unknown to this class.
     * @throws IllegalArgumentException
     *             if count is negative or greater than the number of elements
     *             in the value.
     */
    public static void encodeGettableValue(ByteSink byteSink,
            ChannelAccessGettableValue<?> value, int count) {
        encodeValue(byteSink, value, count);
    }

    /**
     * <p>
     * Encodes a puttable ChannelAccess value into its serialized form. Puttable
     * values are values that can be used in put operations (sending data from a
     * client to a server). If strings have to be serialized, the value's
     * encoding as specified by the return value of its
     * <code>getCharset()</code> method is used.
     * </p>
     * 
     * <p>
     * The specified value must have been created by
     * {@link ChannelAccessValueCodec} or {@link ChannelAccessValueFactory},
     * because this method expects a certain implementation. If the
     * implementation is not the expected one, a {@link ClassCastException} is
     * thrown.
     * </p>
     * 
     * <p>
     * The serialization is limited to the specified number elements, even if
     * the value contains more elements. Therefore, the number of bytes written
     * is always equal to the number returned by
     * {@link #calculatePayloadSize(ChannelAccessValueType, int)}.
     * </p>
     * 
     * <p>
     * This method throws an {@link BufferOverflowException} if there is not
     * enough space left in the byte sink to write the complete value. If this
     * method throws an exception, the byte sink will remain in a state as if no
     * data had been written.
     * </p>
     * 
     * <p>
     * This method simply delegates to
     * {@link #encodeValue(ByteSink, ChannelAccessValue, int)} and is mainly
     * provided for symmetry with
     * {@link #decodePuttableValue(ByteSource, ChannelAccessValueType, int, int, Charset)}
     * .
     * </p>
     * 
     * @param byteSink
     *            byte sink the value's data is written to (never
     *            <code>null</code>).
     * @param value
     *            value that should be serialized to the byte sink (never
     *            <code>null</code>).
     * @param count
     *            number of elements that should be serialized. This number must
     *            not be negative and must be less than or equal to the number
     *            of elements that the value has.
     * @throws BufferOverflowException
     *             if the value cannot be written to the byte source completely,
     *             because it throws a {@link BufferOverflowException}.
     * @throws ClassCastException
     *             if the specified <code>value</code> has not been created by
     *             {@link ChannelAccessValueCodec} or
     *             {@link ChannelAccessValueFactory} and thus is an
     *             implementation unknown to this class.
     * @throws IllegalArgumentException
     *             if count is negative or greater than the number of elements
     *             in the value.
     */
    public static void encodePuttableValue(ByteSink byteSink,
            ChannelAccessPuttableValue<?> value, int count) {
        encodeValue(byteSink, value, count);
    }

    /**
     * <p>
     * Encodes a ChannelAccess value into its serialized form. If strings have
     * to be serialized, the value's encoding as specified by the return value
     * of its <code>getCharset()</code> method is used.
     * </p>
     * 
     * <p>
     * The specified value must have been created by
     * {@link ChannelAccessValueCodec} or {@link ChannelAccessValueFactory},
     * because this method expects a certain implementation. If the
     * implementation is not the expected one, a {@link ClassCastException} is
     * thrown.
     * </p>
     * 
     * <p>
     * The serialization is limited to the specified number elements, even if
     * the value contains more elements. Therefore, the number of bytes written
     * is always equal to the number returned by
     * {@link #calculatePayloadSize(ChannelAccessValueType, int)}.
     * </p>
     * 
     * <p>
     * This method throws an {@link BufferOverflowException} if there is not
     * enough space left in the byte sink to write the complete value. If this
     * method throws an exception, the byte sink will remain in a state as if no
     * data had been written.
     * </p>
     * 
     * @param byteSink
     *            byte sink the value's data is written to (never
     *            <code>null</code>).
     * @param value
     *            value that should be serialized to the byte sink (never
     *            <code>null</code>).
     * @param count
     *            number of elements that should be serialized. This number must
     *            not be negative and must be less than or equal to the number
     *            of elements that the value has.
     * @throws BufferOverflowException
     *             if the value cannot be written to the byte source completely,
     *             because it throws a {@link BufferOverflowException}.
     * @throws ClassCastException
     *             if the specified <code>value</code> has not been created by
     *             {@link ChannelAccessValueCodec} or
     *             {@link ChannelAccessValueFactory} and thus is an
     *             implementation unknown to this class.
     * @throws IllegalArgumentException
     *             if count is negative or greater than the number of elements
     *             in the value.
     */
    public static void encodeValue(final ByteSink byteSink,
            final ChannelAccessValue<?> value, final int count) {
        if (count < 0) {
            throw new IllegalArgumentException(
                    "Number of elements must not be negative.");
        }
        if (count > value.getValueSize()) {
            throw new IllegalArgumentException(
                    "Number of elements to be serialized ("
                            + count
                            + ") must not be greater than number of elements in the value ("
                            + value.getValueSize() + ").");
        }
        // This might cause a ClassCastException, but this is okay, because we
        // specified it in the interface description.
        final ChannelAccessValueBase<?, ?> valueImpl = (ChannelAccessValueBase<?, ?>) value;
        byteSink.atomicPut(new AtomicPutOperation<Void>() {
            @Override
            public Void put() {
                valueImpl.serialize(byteSink, count);
                return null;
            }
        });
    }

    /**
     * Calculates the serialized size of a value of the specified type having
     * the specified number of elements.
     * 
     * @param type
     *            value type to calculate size for (never <code>null</code>).
     * @param count
     *            number of elements in the value. Must not be negative.
     * @return space that is needed for the serialized form of the value in
     *         bytes.
     * @throws IllegalArgumentException
     *             if <code>count</code> is negative or the return value would
     *             exceed {@link Integer#MAX_VALUE}.
     */
    public static int calculatePayloadSize(ChannelAccessValueType type,
            int count) {
        if (count < 0) {
            throw new IllegalArgumentException(
                    "Element count must not be negative.");
        }
        int typeNumber = type.toTypeCode();
        int baseSize = BASE_SIZE[typeNumber];
        int elementSize = ELEMENT_SIZE[typeNumber];
        if (count > MAX_COUNT[typeNumber]) {
            throw new IllegalArgumentException(
                    "Cannot calculate the size of a "
                            + type
                            + " with "
                            + count
                            + " elements, because the size would exceed the maximum integer value.");
        }
        return baseSize + elementSize * count;
    }

    /**
     * Calculates the maximum number of elements a value of the specified type
     * can have, if it has to fit in the specified space. If a value of the
     * specified type cannot fit in the specified space because even its
     * smallest representation is too large, zero is returned.
     * 
     * @param type
     *            ChannelAccess value type to check.
     * @param maxDataSize
     *            size of the space (in bytes) in which the serialized value has
     *            to fit.
     * @return the number of elements that fit in the specified space or
     *         <code>0</code> if the specified type cannot fit into the
     *         specified space.
     */
    public static int calculateMaxCount(ChannelAccessValueType type,
            int maxDataSize) {
        int typeNumber = type.toTypeCode();
        int baseSize = BASE_SIZE[typeNumber];
        int elementSize = ELEMENT_SIZE[typeNumber];
        if (maxDataSize < baseSize) {
            // If there is not even enough space for the base, the type cannot
            // fit. We make this check to catch corner cases with negative
            // numbers.
            return 0;
        }
        int maxCount = (maxDataSize - baseSize) / elementSize;
        return maxCount;
    }

}
