/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.ShortBuffer;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessAcknowledgeAlarm;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Implementation of DBR_PUT_ACKS.
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessAcknowledgeAlarmImpl extends
        ChannelAccessValueBase<Short, ChannelAccessAcknowledgeAlarmImpl>
        implements ChannelAccessAcknowledgeAlarm {

    protected short[] value;

    /**
     * Creates a DBR_PUT_ACKS.
     * 
     * @param acknowledgeAlarmSeverity
     *            severity of the highest alarm to be acknowledged.
     */
    public ChannelAccessAcknowledgeAlarmImpl(
            ChannelAccessAlarmSeverity acknowledgeAlarmSeverity) {
        this.value = new short[] { acknowledgeAlarmSeverity.toSeverityCode() };
    }

    /**
     * Creates a read-only DBR_PUT_ACKS. All read operations are delegated to
     * <code>delegate</code>.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessAcknowledgeAlarmImpl(
            ChannelAccessAcknowledgeAlarmImpl delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     */
    private ChannelAccessAcknowledgeAlarmImpl() {
    }

    @Override
    public ChannelAccessValueType getType() {
        return ChannelAccessValueType.DBR_PUT_ACKS;
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.length;
    }

    @Override
    public Short getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value[index];
    }

    @Override
    public ChannelAccessAlarmSeverity getAcknowledgeAlarmSeverity() {
        if (delegate != null) {
            return delegate.getAcknowledgeAlarmSeverity();
        }
        if (value.length == 0) {
            return ChannelAccessAlarmSeverity.NO_ALARM;
        } else {
            try {
                return ChannelAccessAlarmSeverity.forSeverityCode(value[0]);
            } catch (IllegalArgumentException e) {
                // An IllegalArgumentException means that the severity code was
                // out of range. This can only mean that it was higher than the
                // highest known severity, which is the INVALID_ALARM severity.
                return ChannelAccessAlarmSeverity.INVALID_ALARM;
            }
        }
    }

    @Override
    public void setAcknowledgeAlarmSeverity(
            ChannelAccessAlarmSeverity acknowledgeAlarmSeverity) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value.length != 1) {
            value = new short[1];
        }
        value[0] = acknowledgeAlarmSeverity.toSeverityCode();
    }

    @Override
    public ShortBuffer getValue() {
        if (delegate != null) {
            return delegate.getValue().asReadOnlyBuffer();
        }
        return ShortBuffer.wrap(value);
    }

    @Override
    public void setValue(short[] value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value == null) {
            throw new NullPointerException();
        }
        this.value = value;
    }

    @Override
    public void serialize(ByteSink byteSink, int count) {
        if (delegate != null) {
            delegate.serialize(byteSink, count);
            return;
        }
        assert (count >= 0);
        assert (count <= value.length);
        if (count != 0) {
            byteSink.putShortArray(value, 0, count);
        }
    }

    @Override
    public ChannelAccessAcknowledgeAlarm asReadOnlyValue() {
        if (delegate != null) {
            return this;
        }
        return new ChannelAccessAcknowledgeAlarmImpl(this);
    }

    @Override
    public boolean equals(Object obj) {
        if (delegate != null) {
            return delegate.equals(obj);
        }
        obj = extractDelegate(obj);
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been checked.
        ChannelAccessAcknowledgeAlarmImpl other = (ChannelAccessAcknowledgeAlarmImpl) obj;
        return new EqualsBuilder().append(this.value, other.value).isEquals();
    }

    @Override
    public int hashCode() {
        if (delegate != null) {
            return delegate.hashCode();
        }
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(this.value).toHashCode();
    }

    @Override
    public ChannelAccessAcknowledgeAlarm clone() {
        if (delegate != null) {
            return delegate.clone();
        }
        ChannelAccessAcknowledgeAlarmImpl copy = (ChannelAccessAcknowledgeAlarmImpl) super
                .clone();
        copy.value = this.value.clone();
        return copy;
    }

    /**
     * Creates a DBR_PUT_ACKS from its serialized representation.
     * 
     * @param byteSource
     *            byte source to read data from.
     * @param count
     *            number of value elements to read.
     * @return DBR_PUT_ACKS instance created from the serialized data read from
     *         <code>byteSource</code>.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    public static ChannelAccessAcknowledgeAlarmImpl deserialize(
            ByteSource byteSource, int count) {
        assert (count >= 0);
        ChannelAccessAcknowledgeAlarmImpl value = new ChannelAccessAcknowledgeAlarmImpl();
        if (count == 0) {
            value.value = ArrayUtils.EMPTY_SHORT_ARRAY;
        } else {
            value.value = byteSource.getShortArray(count);
        }
        return value;
    }

}
