/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;
import java.util.Collection;

import com.aquenos.epics.jackie.common.value.ChannelAccessValueCodec;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmAcknowledgementStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Implementation of DBR_STSACK_STRING.
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessAlarmAcknowledgementStatusImpl extends
        AbstractStringBase<ChannelAccessAlarmAcknowledgementStatusImpl>
        implements ChannelAccessAlarmAcknowledgementStatus {

    private static final int META_DATA_SIZE = ChannelAccessValueCodec.calculatePayloadSize(
            ChannelAccessValueType.DBR_STSACK_STRING, 0);

    /**
     * Current alarm status.
     */
    protected ChannelAccessAlarmStatus alarmStatus;

    /**
     * Current alarm severity.
     */
    protected ChannelAccessAlarmSeverity currentAlarmSeverity;

    /**
     * Explicitly acknowledge transient alarms?
     */
    protected boolean explicitlyAcknowledgeTransientAlarms;

    /**
     * Highest unacknowledged alarm severity.
     */
    protected ChannelAccessAlarmSeverity unacknowledgedAlarmSeverity;

    /**
     * Creates a DBR_STSACK_STRING.
     * 
     * @param value
     *            string value.
     * @param alarmStatus
     *            current alarm status
     * @param currentAlarmSeverity
     *            current alarm severity.
     * @param explicitlyAcknowledgeTransientAlarms
     *            Do transient alarms have to be acknowledged explicitly?
     * @param unackowledgedAlarmSeverity
     *            severity of highest unacknowledged alarm.
     * @param charset
     *            encoding used for encoding/decoding strings.
     */
    public ChannelAccessAlarmAcknowledgementStatusImpl(
            Collection<String> value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity currentAlarmSeverity,
            boolean explicitlyAcknowledgeTransientAlarms,
            ChannelAccessAlarmSeverity unackowledgedAlarmSeverity,
            Charset charset) {
        super(value, charset);
        assert (alarmStatus != null);
        this.alarmStatus = alarmStatus;
        assert (currentAlarmSeverity != null);
        this.currentAlarmSeverity = currentAlarmSeverity;
        this.explicitlyAcknowledgeTransientAlarms = explicitlyAcknowledgeTransientAlarms;
        assert (unackowledgedAlarmSeverity != null);
        this.unacknowledgedAlarmSeverity = unackowledgedAlarmSeverity;
    }

    /**
     * Creates a read-only DBR_STSACK_STRING. All read operations are delegated
     * to <code>delegate</code>.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessAlarmAcknowledgementStatusImpl(
            ChannelAccessAlarmAcknowledgementStatusImpl delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     * 
     * @param charset
     *            encoding used for encoding/decoding strings.
     */
    private ChannelAccessAlarmAcknowledgementStatusImpl(Charset charset) {
        super(charset);
    }

    @Override
    public ChannelAccessValueType getType() {
        return ChannelAccessValueType.DBR_STSACK_STRING;
    }

    @Override
    public ChannelAccessAlarmStatus getAlarmStatus() {
        if (delegate != null) {
            return delegate.getAlarmStatus();
        }
        return alarmStatus;
    }

    @Override
    public void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        this.alarmStatus = alarmStatus;
    }

    @Override
    public ChannelAccessAlarmSeverity getAlarmSeverity() {
        if (delegate != null) {
            return delegate.getAlarmSeverity();
        }
        return currentAlarmSeverity;
    }

    @Override
    public void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        this.currentAlarmSeverity = alarmSeverity;
    }

    @Override
    public boolean isExplicitlyAcknowledgeTransientAlarms() {
        if (delegate != null) {
            return delegate.isExplicitlyAcknowledgeTransientAlarms();
        }
        return explicitlyAcknowledgeTransientAlarms;
    }

    @Override
    public void setExplicitlyAcknowledgeTransientAlarms(
            boolean explicitlyAcknowledgeTransientAlarms) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        this.explicitlyAcknowledgeTransientAlarms = explicitlyAcknowledgeTransientAlarms;
    }

    @Override
    public ChannelAccessAlarmSeverity getUnacknowledgedAlarmSeverity() {
        if (delegate != null) {
            return delegate.getUnacknowledgedAlarmSeverity();
        }
        return unacknowledgedAlarmSeverity;
    }

    @Override
    public void setUnacknowledgedAlarmSeverity(
            ChannelAccessAlarmSeverity unacknowledgedAlarmSeverity) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        this.unacknowledgedAlarmSeverity = unacknowledgedAlarmSeverity;
    }

    @Override
    public void serialize(ByteSink byteSink, int count) {
        if (delegate != null) {
            delegate.serialize(byteSink, count);
            return;
        }
        assert (count >= 0);
        assert (count <= value.size());
        ValueCodecUtils.serializeAlarms(byteSink, alarmStatus,
                currentAlarmSeverity);
        ValueCodecUtils.serializeBooleanAsShort(byteSink,
                explicitlyAcknowledgeTransientAlarms);
        ValueCodecUtils.serializeAlarmSeverity(byteSink,
                unacknowledgedAlarmSeverity);
        // Serialize strings.
        serializeValue(byteSink, count);
    }

    @Override
    public ChannelAccessAlarmAcknowledgementStatus asReadOnlyValue() {
        if (delegate != null) {
            return this;
        }
        return new ChannelAccessAlarmAcknowledgementStatusImpl(this);
    }

    @Override
    public boolean equals(Object obj) {
        if (delegate != null) {
            return delegate.equals(obj);
        }
        obj = extractDelegate(obj);
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been checked.
        ChannelAccessAlarmAcknowledgementStatusImpl other = (ChannelAccessAlarmAcknowledgementStatusImpl) obj;
        return new EqualsBuilder()
                .append(this.alarmStatus, other.alarmStatus)
                .append(this.currentAlarmSeverity, other.currentAlarmSeverity)
                .append(this.explicitlyAcknowledgeTransientAlarms,
                        other.explicitlyAcknowledgeTransientAlarms)
                .append(this.unacknowledgedAlarmSeverity,
                        other.unacknowledgedAlarmSeverity).isEquals();
    }

    @Override
    public int hashCode() {
        if (delegate != null) {
            return delegate.hashCode();
        }
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(this.alarmStatus).append(this.currentAlarmSeverity)
                .append(this.explicitlyAcknowledgeTransientAlarms)
                .append(this.unacknowledgedAlarmSeverity).toHashCode();
    }

    @Override
    public ChannelAccessAlarmAcknowledgementStatus clone() {
        if (delegate != null) {
            return delegate.clone();
        }
        ChannelAccessAlarmAcknowledgementStatusImpl copy = (ChannelAccessAlarmAcknowledgementStatusImpl) super
                .clone();
        return copy;
    }

    /**
     * Creates a DBR_STSACK_STRING from its serialized representation.
     * 
     * @param byteSource
     *            byte source to read data from.
     * @param numberOfBytes
     *            number of bytes that contain the value and shall be read
     *            from the byte source.
     * @param count
     *            number of value elements to read.
     * @param charset
     *            character-set used for encoding and decoding strings.
     * @return DBR_STSTACK_STRING instance created from the serialized data read
     *         from <code>byteSource</code>.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    public static ChannelAccessAlarmAcknowledgementStatusImpl deserialize(
            ByteSource byteSource, int numberOfBytes, int count,
            Charset charset) {
        ChannelAccessAlarmAcknowledgementStatusImpl value = new ChannelAccessAlarmAcknowledgementStatusImpl(
                charset);
        value.alarmStatus = ValueCodecUtils.deserializeAlarmStatus(byteSource);
        value.currentAlarmSeverity = ValueCodecUtils
                .deserializeAlarmSeverity(byteSource);
        value.explicitlyAcknowledgeTransientAlarms = ValueCodecUtils
                .deserializeBooleanFromShort(byteSource);
        value.unacknowledgedAlarmSeverity = ValueCodecUtils
                .deserializeAlarmSeverity(byteSource);
        value.deserializeValue(byteSource, numberOfBytes - META_DATA_SIZE,
                count);
        return value;
    }

}
