/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeStamp;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Base class for Channel Access char type. This class contains the various
 * specializations of the char type as inlined classes.
 * 
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessCharBase<ImplementationType extends ChannelAccessCharBase<ImplementationType>>
        extends ChannelAccessValueBase<Byte, ImplementationType> implements
        ChannelAccessChar {

    /**
     * Implementation of DBR_CHAR.
     */
    public static final class ChannelAccessCharImpl extends
            ChannelAccessCharBase<ChannelAccessCharImpl> implements
            ChannelAccessSimpleOnlyChar {

        /**
         * Creates a DBR_CHAR.
         * 
         * @param value
         *            char value.
         */
        public ChannelAccessCharImpl(byte[] value) {
            super(value);
        }

        /**
         * Creates a read-only DBR_DOUBLE. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessCharImpl(ChannelAccessCharImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessCharImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_CHAR;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessSimpleOnlyChar asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessCharImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessSimpleOnlyChar clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessSimpleOnlyChar) super.clone();
        }

        /**
         * Creates a DBR_CHAR from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_CHAR instance created from the serialized data read from
         *         <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessCharImpl deserialize(ByteSource byteSource,
                int count) {
            ChannelAccessCharImpl value = new ChannelAccessCharImpl();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for char with alarms.
     */
    public static abstract class ChannelAccessAlarmCharBase<ImplementationType extends ChannelAccessAlarmCharBase<ImplementationType>>
            extends ChannelAccessCharBase<ImplementationType> implements
            ChannelAccessAlarmChar {

        /**
         * Alarm status.
         */
        protected ChannelAccessAlarmStatus alarmStatus;

        /**
         * Alarm severity.
         */
        protected ChannelAccessAlarmSeverity alarmSeverity;

        /**
         * Base constructor.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        private ChannelAccessAlarmCharBase(byte[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value);
            this.alarmStatus = alarmStatus;
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmCharBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmCharBase() {
        }

        @Override
        public ChannelAccessAlarmStatus getAlarmStatus() {
            if (delegate != null) {
                return delegate.getAlarmStatus();
            }
            return this.alarmStatus;
        }

        @Override
        public void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmStatus != null);
            this.alarmStatus = alarmStatus;
        }

        @Override
        public ChannelAccessAlarmSeverity getAlarmSeverity() {
            if (delegate != null) {
                return delegate.getAlarmSeverity();
            }
            return this.alarmSeverity;
        }

        @Override
        public void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmSeverity != null);
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Deserializes the alarm status from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeAlarms(ByteSource byteSource) {
            this.alarmStatus = ValueCodecUtils
                    .deserializeAlarmStatus(byteSource);
            this.alarmSeverity = ValueCodecUtils
                    .deserializeAlarmSeverity(byteSource);
        }

        /**
         * Serializes the alarm status to the byte sink.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeAlarms(ByteSink byteSink) {
            ValueCodecUtils.serializeAlarms(byteSink, alarmStatus,
                    alarmSeverity);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessAlarmCharBase<ImplementationType> other = (ChannelAccessAlarmCharBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.alarmStatus, other.alarmStatus)
                    .append(this.alarmSeverity, other.alarmSeverity).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.alarmStatus).append(this.alarmSeverity)
                    .toHashCode();
        }

        @Override
        public ChannelAccessAlarmChar clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return (ChannelAccessAlarmChar) super.clone();
        }

    }

    /**
     * Implementation of DBR_STS_CHAR.
     */
    public static final class ChannelAccessAlarmCharImpl extends
            ChannelAccessAlarmCharBase<ChannelAccessAlarmCharImpl> implements
            ChannelAccessAlarmOnlyChar {

        /**
         * Creates a DBR_STS_CHAR.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        public ChannelAccessAlarmCharImpl(byte[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value, alarmStatus, alarmSeverity);
        }

        /**
         * Creates a read-only DBR_STS_CHAR. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmCharImpl(ChannelAccessAlarmCharImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmCharImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_STS_CHAR;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            // Insert extra padding byte.
            byteSink.putByte((byte) 0);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessAlarmOnlyChar asReadOnlyValue() {
            if (delegate != null) {
                return this;
            } else {
                return new ChannelAccessAlarmCharImpl(this);
            }
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessAlarmOnlyChar clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessAlarmOnlyChar) super.clone();
        }

        /**
         * Creates a DBR_STS_CHAR from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_STS_CHAR instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessAlarmCharImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessAlarmCharImpl value = new ChannelAccessAlarmCharImpl();
            value.deserializeAlarms(byteSource);
            // Skip extra padding byte.
            byteSource.skip(1);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_TIME_CHAR.
     */
    public static final class ChannelAccessTimeCharImpl extends
            ChannelAccessAlarmCharBase<ChannelAccessTimeCharImpl> implements
            ChannelAccessTimeChar {

        private static final byte[] THREE_BYTES = new byte[3];

        /**
         * Seconds part of time-stamp.
         */
        protected int timeStampSeconds;

        /**
         * Nanoseconds part of time-stamp.
         */
        protected int timeStampNanoseconds;

        /**
         * Creates a DBR_TIME_CHAR.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param timeStampSeconds
         *            seconds part of time-stamp.
         * @param timeStampNanoSeconds
         *            nanoseconds part of time-stamp.
         */
        public ChannelAccessTimeCharImpl(byte[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
                int timeStampNanoSeconds) {
            super(value, alarmStatus, alarmSeverity);
            this.timeStampSeconds = timeStampSeconds;
            if (timeStampNanoSeconds < 0 || timeStampNanoSeconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of timestamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoSeconds;
        }

        /**
         * Creates a read-only DBR_TIME_CHAR. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessTimeCharImpl(ChannelAccessTimeCharImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessTimeCharImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_TIME_CHAR;
        }

        @Override
        public ChannelAccessTimeStamp getTimeStamp() {
            if (delegate != null) {
                return delegate.getTimeStamp();
            }
            return ChannelAccessValueFactory.createTimeStamp(timeStampSeconds,
                    timeStampNanoseconds);
        }

        @Override
        public int getTimeSeconds() {
            if (delegate != null) {
                return delegate.getTimeSeconds();
            }
            return timeStampSeconds;
        }

        @Override
        public int getTimeNanoseconds() {
            if (delegate != null) {
                return delegate.getTimeNanoseconds();
            }
            return timeStampNanoseconds;
        }

        @Override
        public void setTimeStamp(ChannelAccessTimeStamp timeStamp) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStamp.getSeconds();
            this.timeStampNanoseconds = timeStamp.getNanoseconds();
        }

        @Override
        public void setTimeSeconds(int timeStampSeconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStampSeconds;
        }

        @Override
        public void setTimeNanoseconds(int timeStampNanoseconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (timeStampNanoseconds < 0 || timeStampNanoseconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of time-stamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoseconds;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            ValueCodecUtils.serializeTime(byteSink, timeStampSeconds,
                    timeStampNanoseconds);
            // Insert three extra padding bytes.
            byteSink.putByteArray(THREE_BYTES);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessTimeChar asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessTimeCharImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessTimeCharImpl other = (ChannelAccessTimeCharImpl) obj;
            return new EqualsBuilder()
                    .append(this.timeStampSeconds, other.timeStampSeconds)
                    .append(this.timeStampNanoseconds,
                            other.timeStampNanoseconds).isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.timeStampSeconds)
                    .append(this.timeStampNanoseconds).toHashCode();
        }

        @Override
        public ChannelAccessTimeChar clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessTimeChar) super.clone();
        }

        /**
         * Creates a DBR_TIME_CHAR from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_TIME_CHAR instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessTimeCharImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessTimeCharImpl value = new ChannelAccessTimeCharImpl();
            value.deserializeAlarms(byteSource);
            // Deserialize time-stamp.
            value.timeStampSeconds = ValueCodecUtils
                    .deserializeTimeSeconds(byteSource);
            value.timeStampNanoseconds = ValueCodecUtils
                    .deserializeTimeNanoseconds(byteSource);
            // Skip three extra padding bytes.
            byteSource.skip(3);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for char with display information.
     */
    public static abstract class ChannelAccessGraphicsCharBase<ImplementationType extends ChannelAccessGraphicsCharBase<ImplementationType>>
            extends ChannelAccessAlarmCharBase<ImplementationType> implements
            ChannelAccessGraphicsChar {

        private static final int ENGINEERING_UNITS_STRING_SIZE = 8;

        /**
         * Engineering units.
         */
        protected String engineeringUnits;

        /**
         * Raw representation of engineering units.
         */
        protected byte[] rawEngineeringUnits;

        /**
         * Upper display limit.
         */
        protected byte upperDisplayLimit;

        /**
         * Lower display limit.
         */
        protected byte lowerDisplayLimit;

        /**
         * Upper alarm limit.
         */
        protected byte upperAlarmLimit;

        /**
         * Upper warning limit.
         */
        protected byte upperWarningLimit;

        /**
         * Lower warning limit.
         */
        protected byte lowerWarningLimit;

        /**
         * Lower alarm limit.
         */
        protected byte lowerAlarmLimit;

        /**
         * Charset used for encoding/decoding strings.
         */
        protected Charset charset;

        /**
         * Base constructor.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsCharBase(byte[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, byte upperDisplayLimit,
                byte lowerDisplayLimit, byte upperAlarmLimit,
                byte upperWarningLimit, byte lowerWarningLimit,
                byte lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity);
            this.engineeringUnits = engineeringUnits;
            assert (charset != null);
            this.charset = charset;
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
            this.upperDisplayLimit = upperDisplayLimit;
            this.lowerDisplayLimit = lowerDisplayLimit;
            this.upperAlarmLimit = upperAlarmLimit;
            this.upperWarningLimit = upperWarningLimit;
            this.lowerWarningLimit = lowerWarningLimit;
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsCharBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsCharBase(Charset charset) {
            this.charset = charset;
        }

        @Override
        public String getUnits() {
            if (delegate != null) {
                return delegate.getUnits();
            }
            return engineeringUnits;
        }

        @Override
        public byte[] getRawUnits() {
            if (delegate != null) {
                return delegate.getRawUnits();
            }
            return rawEngineeringUnits.clone();
        }

        @Override
        public void setUnits(String units) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.engineeringUnits = units;
            // Update raw representation.
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
        }

        @Override
        public void setRawUnits(byte[] rawUnits) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            // We use a new array for storing the raw units. This way, we do
            // not have to copy the array in the clone() method and we do not
            // have to fill the array with null bytes explicitly.
            this.rawEngineeringUnits = new byte[ENGINEERING_UNITS_STRING_SIZE];
            // Copy first seven bytes.
            int copyLength = Math.min(rawUnits.length,
                    ENGINEERING_UNITS_STRING_SIZE - 1);
            System.arraycopy(rawUnits, 0, this.rawEngineeringUnits, 0,
                    copyLength);
            // Update string.
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
        }

        @Override
        public Byte getGenericUpperDisplayLimit() {
            return getUpperDisplayLimit();
        }

        @Override
        public Byte getGenericLowerDisplayLimit() {
            return getLowerDisplayLimit();
        }

        @Override
        public Byte getGenericUpperAlarmLimit() {
            return getUpperAlarmLimit();
        }

        @Override
        public Byte getGenericUpperWarningLimit() {
            return getUpperWarningLimit();
        }

        @Override
        public Byte getGenericLowerWarningLimit() {
            return getLowerWarningLimit();
        }

        @Override
        public Byte getGenericLowerAlarmLimit() {
            return getLowerAlarmLimit();
        }

        @Override
        public Charset getCharset() {
            if (delegate != null) {
                return delegate.getCharset();
            }
            return charset;
        }

        @Override
        public byte getUpperDisplayLimit() {
            if (delegate != null) {
                return delegate.getUpperDisplayLimit();
            }
            return upperDisplayLimit;
        }

        @Override
        public void setUpperDisplayLimit(byte upperDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperDisplayLimit = upperDisplayLimit;
        }

        @Override
        public byte getLowerDisplayLimit() {
            if (delegate != null) {
                return delegate.getLowerDisplayLimit();
            }
            return lowerDisplayLimit;
        }

        @Override
        public void setLowerDisplayLimit(byte lowerDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerDisplayLimit = lowerDisplayLimit;
        }

        @Override
        public byte getUpperAlarmLimit() {
            if (delegate != null) {
                return delegate.getUpperAlarmLimit();
            }
            return upperAlarmLimit;
        }

        @Override
        public void setUpperAlarmLimit(byte upperAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperAlarmLimit = upperAlarmLimit;
        }

        @Override
        public byte getUpperWarningLimit() {
            if (delegate != null) {
                return delegate.getUpperWarningLimit();
            }
            return upperWarningLimit;
        }

        @Override
        public void setUpperWarningLimit(byte upperWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperWarningLimit = upperWarningLimit;
        }

        @Override
        public byte getLowerWarningLimit() {
            if (delegate != null) {
                return delegate.getLowerWarningLimit();
            }
            return lowerWarningLimit;
        }

        @Override
        public void setLowerWarningLimit(byte lowerWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerWarningLimit = lowerWarningLimit;
        }

        @Override
        public byte getLowerAlarmLimit() {
            if (delegate != null) {
                return delegate.getLowerAlarmLimit();
            }
            return lowerAlarmLimit;
        }

        @Override
        public void setLowerAlarmLimit(byte lowerAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        /**
         * Deserializes the display information from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeGraphics(ByteSource byteSource) {
            this.rawEngineeringUnits = byteSource
                    .getByteArray(ENGINEERING_UNITS_STRING_SIZE);
            // Ensure that engineering units are null terminated.
            this.rawEngineeringUnits[ENGINEERING_UNITS_STRING_SIZE - 1] = 0;
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
            this.upperDisplayLimit = byteSource.getByte();
            this.lowerDisplayLimit = byteSource.getByte();
            this.upperAlarmLimit = byteSource.getByte();
            this.upperWarningLimit = byteSource.getByte();
            this.lowerWarningLimit = byteSource.getByte();
            this.lowerAlarmLimit = byteSource.getByte();
        }

        /**
         * Serializes the display information.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeGraphics(ByteSink byteSink) {
            byteSink.putByteArray(rawEngineeringUnits);
            byteSink.putByte(upperDisplayLimit);
            byteSink.putByte(lowerDisplayLimit);
            byteSink.putByte(upperAlarmLimit);
            byteSink.putByte(upperWarningLimit);
            byteSink.putByte(lowerWarningLimit);
            byteSink.putByte(lowerAlarmLimit);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessGraphicsCharBase<ImplementationType> other = (ChannelAccessGraphicsCharBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.engineeringUnits, other.engineeringUnits)
                    .append(this.rawEngineeringUnits, other.rawEngineeringUnits)
                    .append(this.upperDisplayLimit, other.upperDisplayLimit)
                    .append(this.lowerDisplayLimit, other.lowerDisplayLimit)
                    .append(this.upperAlarmLimit, other.upperAlarmLimit)
                    .append(this.upperWarningLimit, other.upperWarningLimit)
                    .append(this.lowerWarningLimit, other.lowerWarningLimit)
                    .append(this.lowerAlarmLimit, other.lowerAlarmLimit)
                    .append(this.charset, other.charset).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.engineeringUnits)
                    .append(this.rawEngineeringUnits)
                    .append(this.upperDisplayLimit)
                    .append(this.lowerDisplayLimit)
                    .append(this.upperAlarmLimit)
                    .append(this.upperWarningLimit)
                    .append(this.lowerWarningLimit)
                    .append(this.lowerAlarmLimit).append(this.charset)
                    .toHashCode();
        }

        @Override
        public ChannelAccessGraphicsChar clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            // We do not have to do anything special because we never modify
            // the rawEngineeringUnits array.
            return (ChannelAccessGraphicsChar) super.clone();
        }

    }

    /**
     * Implementation of DBR_GR_CHAR.
     */
    public static final class ChannelAccessGraphicsCharImpl extends
            ChannelAccessGraphicsCharBase<ChannelAccessGraphicsCharImpl>
            implements ChannelAccessGraphicsOnlyChar {

        /**
         * Creates a DBR_GR_CHAR.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessGraphicsCharImpl(byte[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, byte upperDisplayLimit,
                byte lowerDisplayLimit, byte upperAlarmLimit,
                byte upperWarningLimit, byte lowerWarningLimit,
                byte lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, engineeringUnits,
                    upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                    upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                    charset);
        }

        /**
         * Creates a read-only DBR_GR_CHAR. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsCharImpl(
                ChannelAccessGraphicsCharImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsCharImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_GR_CHAR;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            // Insert extra padding byte.
            byteSink.putByte((byte) 0);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessGraphicsOnlyChar asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessGraphicsCharImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessGraphicsOnlyChar clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessGraphicsOnlyChar) super.clone();
        }

        /**
         * Creates a DBR_GR_CHAR from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_GR_CHAR instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessGraphicsCharImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessGraphicsCharImpl value = new ChannelAccessGraphicsCharImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            // Skip extra padding byte.
            byteSource.skip(1);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_CTRL_CHAR.
     */
    public static final class ChannelAccessControlsCharImpl extends
            ChannelAccessGraphicsCharBase<ChannelAccessControlsCharImpl>
            implements ChannelAccessControlsChar {

        /**
         * Upper control limit.
         */
        protected byte upperControlLimit;

        /**
         * Lower control limit.
         */
        protected byte lowerControlLimit;

        /**
         * Creates a DBR_CTRL_CHAR.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param upperControlLimit
         *            upper control limit.
         * @param lowerControlLimit
         *            lower control limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessControlsCharImpl(byte[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, byte upperDisplayLimit,
                byte lowerDisplayLimit, byte upperAlarmLimit,
                byte upperWarningLimit, byte lowerWarningLimit,
                byte lowerAlarmLimit, byte upperControlLimit,
                byte lowerControlLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, engineeringUnits,
                    upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                    upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                    charset);
            this.upperControlLimit = upperControlLimit;
            this.lowerControlLimit = lowerControlLimit;
        }

        /**
         * Creates a read-only DBR_CTRL_DOUBLE. All read operations are
         * delegated to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessControlsCharImpl(
                ChannelAccessControlsCharImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessControlsCharImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_CTRL_CHAR;
        }

        @Override
        public Byte getGenericUpperControlLimit() {
            return getUpperControlLimit();
        }

        @Override
        public Byte getGenericLowerControlLimit() {
            return getLowerControlLimit();
        }

        @Override
        public byte getUpperControlLimit() {
            if (delegate != null) {
                return delegate.getUpperControlLimit();
            }
            return upperControlLimit;
        }

        @Override
        public void setUpperControlLimit(byte upperControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperControlLimit = upperControlLimit;
        }

        @Override
        public byte getLowerControlLimit() {
            if (delegate != null) {
                return delegate.getLowerControlLimit();
            }
            return lowerControlLimit;
        }

        @Override
        public void setLowerControlLimit(byte lowerControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerControlLimit = lowerControlLimit;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            byteSink.putByte(upperControlLimit);
            byteSink.putByte(lowerControlLimit);
            // Insert extra padding byte.
            byteSink.putByte((byte) 0);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessControlsChar asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessControlsCharImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessControlsCharImpl other = (ChannelAccessControlsCharImpl) obj;
            return new EqualsBuilder()
                    .append(this.upperControlLimit, other.upperControlLimit)
                    .append(this.lowerControlLimit, other.lowerControlLimit)
                    .isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.upperControlLimit)
                    .append(this.lowerControlLimit).toHashCode();
        }

        @Override
        public ChannelAccessControlsChar clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessControlsChar) super.clone();
        }

        /**
         * Creates a DBR_CTRL_CHAR from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_CTRL_CHAR instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessControlsCharImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessControlsCharImpl value = new ChannelAccessControlsCharImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.upperControlLimit = byteSource.getByte();
            value.lowerControlLimit = byteSource.getByte();
            // Skip extra padding byte.
            byteSource.skip(1);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    protected byte[] value;

    /**
     * Base constructor.
     * 
     * @param value
     *            char value.
     */
    private ChannelAccessCharBase(byte[] value) {
        assert (value != null);
        this.value = value;
    }

    /**
     * Base constructor for read-only variant.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessCharBase(ImplementationType delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     */
    private ChannelAccessCharBase() {
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.length;
    }

    @Override
    public Byte getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value[index];
    }

    @Override
    public ByteBuffer getValue() {
        if (delegate != null) {
            return delegate.getValue().asReadOnlyBuffer();
        }
        return ByteBuffer.wrap(value);
    }

    @Override
    public void setValue(byte[] value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value == null) {
            throw new NullPointerException();
        }
        this.value = value;
    }

    /**
     * Deserializes the value elements from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @param count
     *            number of elements to deserialize.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    protected void deserializeValue(ByteSource byteSource, int count) {
        assert (count >= 0);
        if (count == 0) {
            value = ArrayUtils.EMPTY_BYTE_ARRAY;
        } else {
            value = byteSource.getByteArray(count);
        }
    }

    /**
     * Serializes the value elements to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write to.
     * @param count
     *            number of elements to be serialized.
     */
    protected void serializeValue(ByteSink byteSink, int count) {
        if (count != 0) {
            byteSink.putByteArray(value, 0, count);
        }
    }

    @Override
    public boolean equals(Object obj) {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been
        // checked.
        @SuppressWarnings("unchecked")
        ChannelAccessCharBase<ImplementationType> other = (ChannelAccessCharBase<ImplementationType>) obj;
        return new EqualsBuilder().append(this.value, other.value).isEquals();
    }

    @Override
    public int hashCode() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(value).toHashCode();
    }

    @Override
    public ChannelAccessChar clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        @SuppressWarnings("unchecked")
        ChannelAccessCharBase<ImplementationType> copy = (ChannelAccessCharBase<ImplementationType>) super
                .clone();
        copy.value = this.value.clone();
        return copy;
    }

}
