/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.ShortBuffer;
import java.nio.charset.Charset;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeStamp;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Base class for Channel Access enum type. This class contains the various
 * specializations of the enum type as inlined classes.
 * 
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessEnumBase<ImplementationType extends ChannelAccessEnumBase<ImplementationType>>
        extends ChannelAccessValueBase<Short, ImplementationType> implements
        ChannelAccessEnum {

    /**
     * Implementation of DBR_ENUM.
     */
    public static final class ChannelAccessEnumImpl extends
            ChannelAccessEnumBase<ChannelAccessEnumImpl> implements
            ChannelAccessSimpleOnlyEnum {

        /**
         * Creates a DBR_ENUM.
         * 
         * @param value
         *            enum value.
         */
        public ChannelAccessEnumImpl(short[] value) {
            super(value);
        }

        /**
         * Creates a read-only DBR_ENUM. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessEnumImpl(ChannelAccessEnumImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessEnumImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_ENUM;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessSimpleOnlyEnum asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessEnumImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessSimpleOnlyEnum clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessSimpleOnlyEnum) super.clone();
        }

        /**
         * Creates a DBR_ENUM from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_ENUM instance created from the serialized data read from
         *         <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessEnumImpl deserialize(ByteSource byteSource,
                int count) {
            ChannelAccessEnumImpl value = new ChannelAccessEnumImpl();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for enum with alarms.
     */
    public static abstract class ChannelAccessAlarmEnumBase<ImplementationType extends ChannelAccessAlarmEnumBase<ImplementationType>>
            extends ChannelAccessEnumBase<ImplementationType> implements
            ChannelAccessAlarmEnum {

        /**
         * Alarm status.
         */
        protected ChannelAccessAlarmStatus alarmStatus;

        /**
         * Alarm severity.
         */
        protected ChannelAccessAlarmSeverity alarmSeverity;

        /**
         * Base constructor.
         * 
         * @param value
         *            enum value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        private ChannelAccessAlarmEnumBase(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value);
            this.alarmStatus = alarmStatus;
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmEnumBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmEnumBase() {
        }

        @Override
        public ChannelAccessAlarmStatus getAlarmStatus() {
            if (delegate != null) {
                return delegate.getAlarmStatus();
            }
            return this.alarmStatus;
        }

        @Override
        public void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmStatus != null);
            this.alarmStatus = alarmStatus;
        }

        @Override
        public ChannelAccessAlarmSeverity getAlarmSeverity() {
            if (delegate != null) {
                return delegate.getAlarmSeverity();
            }
            return this.alarmSeverity;
        }

        @Override
        public void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmSeverity != null);
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Deserializes the alarm status from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeAlarms(ByteSource byteSource) {
            this.alarmStatus = ValueCodecUtils
                    .deserializeAlarmStatus(byteSource);
            this.alarmSeverity = ValueCodecUtils
                    .deserializeAlarmSeverity(byteSource);
        }

        /**
         * Serializes the alarm status to the byte sink.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeAlarms(ByteSink byteSink) {
            ValueCodecUtils.serializeAlarms(byteSink, alarmStatus,
                    alarmSeverity);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessAlarmEnumBase<ImplementationType> other = (ChannelAccessAlarmEnumBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.alarmStatus, other.alarmStatus)
                    .append(this.alarmSeverity, other.alarmSeverity).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.alarmStatus).append(this.alarmSeverity)
                    .toHashCode();
        }

        @Override
        public ChannelAccessAlarmEnum clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return (ChannelAccessAlarmEnum) super.clone();
        }

    }

    /**
     * Implementation of DBR_STS_ENUM.
     */
    public static final class ChannelAccessAlarmEnumImpl extends
            ChannelAccessAlarmEnumBase<ChannelAccessAlarmEnumImpl> implements
            ChannelAccessAlarmOnlyEnum {

        /**
         * Creates a DBR_STS_ENUM.
         * 
         * @param value
         *            enum value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        public ChannelAccessAlarmEnumImpl(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value, alarmStatus, alarmSeverity);
        }

        /**
         * Creates a read-only DBR_STS_ENUM. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmEnumImpl(ChannelAccessAlarmEnumImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmEnumImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_STS_ENUM;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessAlarmOnlyEnum asReadOnlyValue() {
            if (delegate != null) {
                return this;
            } else {
                return new ChannelAccessAlarmEnumImpl(this);
            }
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessAlarmOnlyEnum clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessAlarmOnlyEnum) super.clone();
        }

        /**
         * Creates a DBR_STS_ENUM from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_STS_ENUM instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessAlarmEnumImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessAlarmEnumImpl value = new ChannelAccessAlarmEnumImpl();
            value.deserializeAlarms(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_TIME_ENUM.
     */
    public static final class ChannelAccessTimeEnumImpl extends
            ChannelAccessAlarmEnumBase<ChannelAccessTimeEnumImpl> implements
            ChannelAccessTimeEnum {

        /**
         * Seconds part of time-stamp.
         */
        protected int timeStampSeconds;

        /**
         * Nanoseconds part of time-stamp.
         */
        protected int timeStampNanoseconds;

        /**
         * Creates a DBR_TIME_ENUM.
         * 
         * @param value
         *            enum value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param timeStampSeconds
         *            seconds part of time-stamp.
         * @param timeStampNanoSeconds
         *            nanoseconds part of time-stamp.
         */
        public ChannelAccessTimeEnumImpl(short[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
                int timeStampNanoSeconds) {
            super(value, alarmStatus, alarmSeverity);
            this.timeStampSeconds = timeStampSeconds;
            if (timeStampNanoSeconds < 0 || timeStampNanoSeconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of timestamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoSeconds;
        }

        /**
         * Creates a read-only DBR_TIME_ENUM. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessTimeEnumImpl(ChannelAccessTimeEnumImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessTimeEnumImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_TIME_ENUM;
        }

        @Override
        public ChannelAccessTimeStamp getTimeStamp() {
            if (delegate != null) {
                return delegate.getTimeStamp();
            }
            return ChannelAccessValueFactory.createTimeStamp(timeStampSeconds,
                    timeStampNanoseconds);
        }

        @Override
        public int getTimeSeconds() {
            if (delegate != null) {
                return delegate.getTimeSeconds();
            }
            return timeStampSeconds;
        }

        @Override
        public int getTimeNanoseconds() {
            if (delegate != null) {
                return delegate.getTimeNanoseconds();
            }
            return timeStampNanoseconds;
        }

        @Override
        public void setTimeStamp(ChannelAccessTimeStamp timeStamp) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStamp.getSeconds();
            this.timeStampNanoseconds = timeStamp.getNanoseconds();
        }

        @Override
        public void setTimeSeconds(int timeStampSeconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStampSeconds;
        }

        @Override
        public void setTimeNanoseconds(int timeStampNanoseconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (timeStampNanoseconds < 0 || timeStampNanoseconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of time-stamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoseconds;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            ValueCodecUtils.serializeTime(byteSink, timeStampSeconds,
                    timeStampNanoseconds);
            // Insert two extra padding bytes.
            byteSink.putShort((short) 0);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessTimeEnum asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessTimeEnumImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessTimeEnumImpl other = (ChannelAccessTimeEnumImpl) obj;
            return new EqualsBuilder()
                    .append(this.timeStampSeconds, other.timeStampSeconds)
                    .append(this.timeStampNanoseconds,
                            other.timeStampNanoseconds).isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.timeStampSeconds)
                    .append(this.timeStampNanoseconds).toHashCode();
        }

        @Override
        public ChannelAccessTimeEnum clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessTimeEnum) super.clone();
        }

        /**
         * Creates a DBR_TIME_ENUM from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_TIME_ENUM instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessTimeEnumImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessTimeEnumImpl value = new ChannelAccessTimeEnumImpl();
            value.deserializeAlarms(byteSource);
            // Deserialize time-stamp.
            value.timeStampSeconds = ValueCodecUtils
                    .deserializeTimeSeconds(byteSource);
            value.timeStampNanoseconds = ValueCodecUtils
                    .deserializeTimeNanoseconds(byteSource);
            // Skip two extra padding bytes.
            byteSource.skip(2);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for short with display information.
     */
    public static final class ChannelAccessGraphicsEnumImpl extends
            ChannelAccessAlarmEnumBase<ChannelAccessGraphicsEnumImpl> implements
            ChannelAccessGraphicsEnum {

        private class WrappedStringList extends AbstractList<String> {

            @Override
            public String get(int index) {
                return labels.get(index);
            }

            @Override
            public int size() {
                return labels.size();
            }

            @Override
            public String set(int index, String element) {
                if (element == null) {
                    throw new NullPointerException();
                }
                String old = labels.set(index, element);
                updateRawLabel(index, element);
                return old;
            }

            @Override
            public void add(int index, String element) {
                if (labels.size() == MAX_NUM_LABELS) {
                    throw new IllegalStateException(
                            "The list cannot contain more than 16 labels.");
                }
                if (element == null) {
                    throw new NullPointerException();
                }
                labels.add(index, element);
                int numberOfLabels = labels.size();
                for (int labelIndex = index; labelIndex < numberOfLabels; labelIndex++) {
                    updateRawLabel(labelIndex, labels.get(labelIndex));
                }
            }

            @Override
            public String remove(int index) {
                String removed = labels.remove(index);
                int numberOfLabels = labels.size();
                for (int labelIndex = index; labelIndex < numberOfLabels; labelIndex++) {
                    updateRawLabel(labelIndex, labels.get(labelIndex));
                }
                // Clear the bytes of the label that is now empty.
                System.arraycopy(EMPTY_RAW_LABEL, 0, rawLabels, numberOfLabels
                        * LABEL_STRING_SIZE, LABEL_STRING_SIZE);
                return removed;
            }

            // Overridden for an optimized implementation.
            @Override
            public void clear() {
                labels.clear();
                // We also clear the raw array so that we will never compare or
                // send garbage.
                System.arraycopy(EMPTY_RAW_LABELS, 0, rawLabels,
                        MAX_NUM_LABELS, TOTAL_LABEL_BYTES);
            }

            // Overridden for an optimized implementation.
            @Override
            public boolean addAll(Collection<? extends String> c) {
                return addAll(labels.size(), c);
            }

            // Overridden for an optimized implementation.
            @Override
            public boolean addAll(int index, Collection<? extends String> c) {
                if (c.size() + labels.size() > MAX_NUM_LABELS) {
                    throw new IllegalStateException(
                            "The list cannot contain more than 16 labels.");
                }
                for (String s : c) {
                    if (s == null) {
                        throw new NullPointerException();
                    }
                }
                boolean modified = labels.addAll(index, c);
                if (modified) {
                    int newSize = labels.size();
                    for (int stringIndex = index; stringIndex < newSize; stringIndex++) {
                        updateRawLabel(stringIndex, labels.get(stringIndex));
                    }
                }
                return modified;
            }

            // Overridden for an optimized implementation.
            @Override
            public Object[] toArray() {
                return labels.toArray();
            }

            // Overridden for an optimized implementation.
            @Override
            public <T> T[] toArray(T[] a) {
                return labels.toArray(a);
            }

        };

        private static final int MAX_NUM_LABELS = 16;
        private static final int LABEL_STRING_SIZE = 26;
        private static final int TOTAL_LABEL_BYTES = MAX_NUM_LABELS
                * LABEL_STRING_SIZE;
        private static final byte[] EMPTY_RAW_LABEL = new byte[LABEL_STRING_SIZE];
        private static final byte[] EMPTY_RAW_LABELS = new byte[TOTAL_LABEL_BYTES];

        /**
         * Actual type represented by this instance.
         */
        protected ChannelAccessValueType type;

        /**
         * Enum state display labels.
         */
        protected ArrayList<String> labels;

        /**
         * Raw representation of labels.
         */
        protected byte[] rawLabels;

        /**
         * Charset used for encoding/decoding strings.
         */
        protected Charset charset;

        /**
         * Creates a DBR_GR_ENUM or DBR_CTRL_ENUM.
         * 
         * @param type
         *            actual type represented by this instance.
         * @param value
         *            short value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param labels
         *            enum state labels.
         * @param charset
         *            encoding used for encoding/decoding strings.
         * @throws IllegalArgumentException
         *             if <code>labels</code> has more than 16 elements.
         */
        public ChannelAccessGraphicsEnumImpl(ChannelAccessValueType type,
                short[] value, ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                Collection<String> labels, Charset charset) {
            super(value, alarmStatus, alarmSeverity);
            assert (type == ChannelAccessValueType.DBR_GR_ENUM || type == ChannelAccessValueType.DBR_CTRL_ENUM);
            this.type = type;
            // The charset must be set before calling setLabels.
            assert (charset != null);
            this.charset = charset;
            int numberOfLabels = labels.size();
            if (numberOfLabels > MAX_NUM_LABELS) {
                throw new IllegalArgumentException(
                        "The number of labels is limited to 16.");
            }
            this.labels = new ArrayList<String>(MAX_NUM_LABELS);
            this.rawLabels = new byte[TOTAL_LABEL_BYTES];
            int labelIndex = 0;
            for (String label : labels) {
                if (label == null) {
                    throw new NullPointerException();
                }
                this.labels.add(label);
                updateRawLabel(labelIndex, label);
                labelIndex++;
            }
        }

        /**
         * Creates a read-only DBR_GR_ENUM. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsEnumImpl(
                ChannelAccessGraphicsEnumImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param type
         *            actual type represented by this instance.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsEnumImpl(ChannelAccessValueType type,
                Charset charset) {
            this.type = type;
            this.charset = charset;
        }

        @Override
        public ChannelAccessValueType getType() {
            if (delegate != null) {
                return delegate.getType();
            }
            return type;
        }

        @Override
        public List<String> getLabels() {
            if (delegate != null) {
                return Collections.unmodifiableList(delegate.getLabels());
            }
            return new WrappedStringList();
        }

        public void setLabels(Collection<String> labels) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (labels.size() > MAX_NUM_LABELS) {
                throw new IllegalArgumentException(
                        "An enum cannot have more than 16 labels.");
            }
            // We check for null pointers first, so that we do not modify the
            // elements list if there are any.
            for (String element : labels) {
                if (element == null) {
                    throw new NullPointerException();
                }
            }
            this.labels.clear();
            int labelIndex = 0;
            for (String label : labels) {
                this.labels.add(label);
                updateRawLabel(labelIndex, label);
                labelIndex++;
            }
        }

        private void updateRawLabel(int labelIndex, String newLabel) {
            byte[] rawLabel = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(newLabel,
                            LABEL_STRING_SIZE, charset);
            System.arraycopy(rawLabel, 0, this.rawLabels, labelIndex
                    * LABEL_STRING_SIZE, LABEL_STRING_SIZE);
        }

        @Override
        public byte[] getRawLabels() {
            if (delegate != null) {
                return delegate.getRawLabels();
            }
            return rawLabels.clone();
        }

        @Override
        public void setRawLabels(byte[] rawLabels, int numberOfLabels) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (numberOfLabels < 0 || numberOfLabels > 16) {
                throw new IllegalArgumentException(
                        "Number of labels must be between 0 and 16.");
            }
            if (rawLabels.length != TOTAL_LABEL_BYTES) {
                throw new IllegalArgumentException(
                        "Raw representation of enum labels must have exactly 416 bytes.");
            }
            // Convert labels and check that all label strings are terminated by
            // a null byte.
            LinkedList<String> newLabels = new LinkedList<String>();
            for (int labelIndex = 0; labelIndex < numberOfLabels; labelIndex++) {
                newLabels.add(NullTerminatedStringUtil
                        .nullTerminatedBytesToString(rawLabels, labelIndex
                                * LABEL_STRING_SIZE, LABEL_STRING_SIZE,
                                charset, true));
            }
            System.arraycopy(rawLabels, 0, this.rawLabels, 0, TOTAL_LABEL_BYTES);
            // We remove all existing labels from the labels list and add the
            // new ones.
            this.labels.clear();
            this.labels.addAll(newLabels);
        }

        @Override
        public Charset getCharset() {
            if (delegate != null) {
                return delegate.getCharset();
            }
            return charset;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            // Serialize labels
            byteSink.putShort((short) labels.size());
            byteSink.putByteArray(rawLabels);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessGraphicsEnum asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessGraphicsEnumImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessGraphicsEnumImpl other = (ChannelAccessGraphicsEnumImpl) obj;
            return new EqualsBuilder().append(this.type, other.type)
                    .append(this.labels, other.labels)
                    .append(this.rawLabels, other.rawLabels)
                    .append(this.charset, other.charset).isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.type).append(this.labels)
                    .append(this.rawLabels).append(this.charset).toHashCode();
        }

        @SuppressWarnings("unchecked")
        @Override
        public ChannelAccessGraphicsEnum clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            ChannelAccessGraphicsEnumImpl copy = (ChannelAccessGraphicsEnumImpl) super
                    .clone();
            copy.labels = (ArrayList<String>) this.labels.clone();
            copy.rawLabels = this.rawLabels.clone();
            return copy;
        }

        /**
         * Creates a DBR_GR_ENUM from its serialized representation.
         * 
         * @param type
         *            actual type represented by this instance.
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_GR_ENUM instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessGraphicsEnumImpl deserialize(
                ChannelAccessValueType type, ByteSource byteSource, int count,
                Charset charset) {
            ChannelAccessGraphicsEnumImpl value = new ChannelAccessGraphicsEnumImpl(
                    type, charset);
            value.deserializeAlarms(byteSource);
            // Deserialize labels
            short numberOfLabels = byteSource.getShort();
            if (numberOfLabels < 0 || numberOfLabels > 16) {
                numberOfLabels = 16;
            }
            byte[] rawLabels = byteSource.getByteArray(TOTAL_LABEL_BYTES);
            ArrayList<String> labels = new ArrayList<String>(MAX_NUM_LABELS);
            for (int labelIndex = 0; labelIndex < numberOfLabels; labelIndex++) {
                // Ensure that each label is null terminated.
                rawLabels[(labelIndex + 1) * LABEL_STRING_SIZE - 1] = 0;
                labels.add(NullTerminatedStringUtil.nullTerminatedBytesToString(
                        rawLabels, labelIndex * LABEL_STRING_SIZE,
                        LABEL_STRING_SIZE, charset, true));
            }
            value.rawLabels = rawLabels;
            value.labels = labels;
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    protected short[] value;

    /**
     * Base constructor.
     * 
     * @param value
     *            enum value.
     */
    private ChannelAccessEnumBase(short[] value) {
        assert (value != null);
        this.value = value;
    }

    /**
     * Base constructor for read-only variant.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessEnumBase(ImplementationType delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     */
    private ChannelAccessEnumBase() {
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.length;
    }

    @Override
    public Short getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value[index];
    }

    @Override
    public ShortBuffer getValue() {
        if (delegate != null) {
            return delegate.getValue().asReadOnlyBuffer();
        }
        return ShortBuffer.wrap(value);
    }

    @Override
    public void setValue(short[] value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value == null) {
            throw new NullPointerException();
        }
        this.value = value;
    }

    /**
     * Deserializes the value elements from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @param count
     *            number of elements to deserialize.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    protected void deserializeValue(ByteSource byteSource, int count) {
        assert (count >= 0);
        if (count == 0) {
            value = ArrayUtils.EMPTY_SHORT_ARRAY;
        } else {
            value = byteSource.getShortArray(count);
        }
    }

    /**
     * Serializes the value elements to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write to.
     * @param count
     *            number of elements to be serialized.
     */
    protected void serializeValue(ByteSink byteSink, int count) {
        if (count != 0) {
            byteSink.putShortArray(value, 0, count);
        }
    }

    @Override
    public boolean equals(Object obj) {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been
        // checked.
        @SuppressWarnings("unchecked")
        ChannelAccessEnumBase<ImplementationType> other = (ChannelAccessEnumBase<ImplementationType>) obj;
        return new EqualsBuilder().append(this.value, other.value).isEquals();
    }

    @Override
    public int hashCode() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(value).toHashCode();
    }

    @Override
    public ChannelAccessEnum clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        @SuppressWarnings("unchecked")
        ChannelAccessEnumBase<ImplementationType> copy = (ChannelAccessEnumBase<ImplementationType>) super
                .clone();
        copy.value = this.value.clone();
        return copy;
    }

}
