/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.FloatBuffer;
import java.nio.charset.Charset;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeStamp;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Base class for Channel Access float type. This class contains the various
 * specializations of the float type as inlined classes.
 * 
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessFloatBase<ImplementationType extends ChannelAccessFloatBase<ImplementationType>>
        extends ChannelAccessValueBase<Float, ImplementationType> implements
        ChannelAccessFloat {

    /**
     * Implementation of DBR_FLOAT.
     */
    public static final class ChannelAccessFloatImpl extends
            ChannelAccessFloatBase<ChannelAccessFloatImpl> implements
            ChannelAccessSimpleOnlyFloat {

        /**
         * Creates a DBR_FLOAT.
         * 
         * @param value
         *            float value.
         */
        public ChannelAccessFloatImpl(float[] value) {
            super(value);
        }

        /**
         * Creates a read-only DBR_FLOAT. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessFloatImpl(ChannelAccessFloatImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessFloatImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_FLOAT;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessSimpleOnlyFloat asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessFloatImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessSimpleOnlyFloat clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessSimpleOnlyFloat) super.clone();
        }

        /**
         * Creates a DBR_FLOAT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_FLOAT instance created from the serialized data read from
         *         <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessFloatImpl deserialize(ByteSource byteSource,
                int count) {
            ChannelAccessFloatImpl value = new ChannelAccessFloatImpl();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for float with alarms.
     */
    public static abstract class ChannelAccessAlarmFloatBase<ImplementationType extends ChannelAccessAlarmFloatBase<ImplementationType>>
            extends ChannelAccessFloatBase<ImplementationType> implements
            ChannelAccessAlarmFloat {

        /**
         * Alarm status.
         */
        protected ChannelAccessAlarmStatus alarmStatus;

        /**
         * Alarm severity.
         */
        protected ChannelAccessAlarmSeverity alarmSeverity;

        /**
         * Base constructor.
         * 
         * @param value
         *            float value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        private ChannelAccessAlarmFloatBase(float[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value);
            this.alarmStatus = alarmStatus;
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmFloatBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmFloatBase() {
        }

        @Override
        public ChannelAccessAlarmStatus getAlarmStatus() {
            if (delegate != null) {
                return delegate.getAlarmStatus();
            }
            return this.alarmStatus;
        }

        @Override
        public void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmStatus != null);
            this.alarmStatus = alarmStatus;
        }

        @Override
        public ChannelAccessAlarmSeverity getAlarmSeverity() {
            if (delegate != null) {
                return delegate.getAlarmSeverity();
            }
            return this.alarmSeverity;
        }

        @Override
        public void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmSeverity != null);
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Deserializes the alarm status from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeAlarms(ByteSource byteSource) {
            this.alarmStatus = ValueCodecUtils
                    .deserializeAlarmStatus(byteSource);
            this.alarmSeverity = ValueCodecUtils
                    .deserializeAlarmSeverity(byteSource);
        }

        /**
         * Serializes the alarm status to the byte sink.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeAlarms(ByteSink byteSink) {
            ValueCodecUtils.serializeAlarms(byteSink, alarmStatus,
                    alarmSeverity);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessAlarmFloatBase<ImplementationType> other = (ChannelAccessAlarmFloatBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.alarmStatus, other.alarmStatus)
                    .append(this.alarmSeverity, other.alarmSeverity).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.alarmStatus).append(this.alarmSeverity)
                    .toHashCode();
        }

        @Override
        public ChannelAccessAlarmFloat clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return (ChannelAccessAlarmFloat) super.clone();
        }

    }

    /**
     * Implementation of DBR_STS_FLOAT.
     */
    public static final class ChannelAccessAlarmFloatImpl extends
            ChannelAccessAlarmFloatBase<ChannelAccessAlarmFloatImpl> implements
            ChannelAccessAlarmOnlyFloat {

        /**
         * Creates a DBR_STS_FLOAT.
         * 
         * @param value
         *            float value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        public ChannelAccessAlarmFloatImpl(float[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value, alarmStatus, alarmSeverity);
        }

        /**
         * Creates a read-only DBR_STS_FLOAT. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmFloatImpl(ChannelAccessAlarmFloatImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmFloatImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_STS_FLOAT;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessAlarmOnlyFloat asReadOnlyValue() {
            if (delegate != null) {
                return this;
            } else {
                return new ChannelAccessAlarmFloatImpl(this);
            }
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessAlarmOnlyFloat clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessAlarmOnlyFloat) super.clone();
        }

        /**
         * Creates a DBR_STS_FLOAT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_STS_FLOAT instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessAlarmFloatImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessAlarmFloatImpl value = new ChannelAccessAlarmFloatImpl();
            value.deserializeAlarms(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_TIME_FLOAT.
     */
    public static final class ChannelAccessTimeFloatImpl extends
            ChannelAccessAlarmFloatBase<ChannelAccessTimeFloatImpl> implements
            ChannelAccessTimeFloat {

        /**
         * Seconds part of time-stamp.
         */
        protected int timeStampSeconds;

        /**
         * Nanoseconds part of time-stamp.
         */
        protected int timeStampNanoseconds;

        /**
         * Creates a DBR_TIME_FLOAT.
         * 
         * @param value
         *            float value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param timeStampSeconds
         *            seconds part of time-stamp.
         * @param timeStampNanoSeconds
         *            nanoseconds part of time-stamp.
         */
        public ChannelAccessTimeFloatImpl(float[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
                int timeStampNanoSeconds) {
            super(value, alarmStatus, alarmSeverity);
            this.timeStampSeconds = timeStampSeconds;
            if (timeStampNanoSeconds < 0 || timeStampNanoSeconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of timestamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoSeconds;
        }

        /**
         * Creates a read-only DBR_TIME_FLOAT. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessTimeFloatImpl(ChannelAccessTimeFloatImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessTimeFloatImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_TIME_FLOAT;
        }

        @Override
        public ChannelAccessTimeStamp getTimeStamp() {
            if (delegate != null) {
                return delegate.getTimeStamp();
            }
            return ChannelAccessValueFactory.createTimeStamp(timeStampSeconds,
                    timeStampNanoseconds);
        }

        @Override
        public int getTimeSeconds() {
            if (delegate != null) {
                return delegate.getTimeSeconds();
            }
            return timeStampSeconds;
        }

        @Override
        public int getTimeNanoseconds() {
            if (delegate != null) {
                return delegate.getTimeNanoseconds();
            }
            return timeStampNanoseconds;
        }

        @Override
        public void setTimeStamp(ChannelAccessTimeStamp timeStamp) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStamp.getSeconds();
            this.timeStampNanoseconds = timeStamp.getNanoseconds();
        }

        @Override
        public void setTimeSeconds(int timeStampSeconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStampSeconds;
        }

        @Override
        public void setTimeNanoseconds(int timeStampNanoseconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (timeStampNanoseconds < 0 || timeStampNanoseconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of time-stamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoseconds;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            ValueCodecUtils.serializeTime(byteSink, timeStampSeconds,
                    timeStampNanoseconds);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessTimeFloat asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessTimeFloatImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessTimeFloatImpl other = (ChannelAccessTimeFloatImpl) obj;
            return new EqualsBuilder()
                    .append(this.timeStampSeconds, other.timeStampSeconds)
                    .append(this.timeStampNanoseconds,
                            other.timeStampNanoseconds).isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.timeStampSeconds)
                    .append(this.timeStampNanoseconds).toHashCode();
        }

        @Override
        public ChannelAccessTimeFloat clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessTimeFloat) super.clone();
        }

        /**
         * Creates a DBR_TIME_FLOAT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_TIME_FLOAT instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessTimeFloatImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessTimeFloatImpl value = new ChannelAccessTimeFloatImpl();
            value.deserializeAlarms(byteSource);
            // Deserialize time-stamp.
            value.timeStampSeconds = ValueCodecUtils
                    .deserializeTimeSeconds(byteSource);
            value.timeStampNanoseconds = ValueCodecUtils
                    .deserializeTimeNanoseconds(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for float with display information.
     */
    public static abstract class ChannelAccessGraphicsFloatBase<ImplementationType extends ChannelAccessGraphicsFloatBase<ImplementationType>>
            extends ChannelAccessAlarmFloatBase<ImplementationType> implements
            ChannelAccessGraphicsFloat {

        private static final int ENGINEERING_UNITS_STRING_SIZE = 8;

        /**
         * Engineering units.
         */
        protected String engineeringUnits;

        /**
         * Raw representation of engineering units.
         */
        protected byte[] rawEngineeringUnits;

        /**
         * Upper display limit.
         */
        protected float upperDisplayLimit;

        /**
         * Lower display limit.
         */
        protected float lowerDisplayLimit;

        /**
         * Upper alarm limit.
         */
        protected float upperAlarmLimit;

        /**
         * Upper warning limit.
         */
        protected float upperWarningLimit;

        /**
         * Lower warning limit.
         */
        protected float lowerWarningLimit;

        /**
         * Lower alarm limit.
         */
        protected float lowerAlarmLimit;

        /**
         * Display precision.
         */
        protected short precision;

        /**
         * Charset used for encoding/decoding strings.
         */
        protected Charset charset;

        /**
         * Base constructor.
         * 
         * @param value
         *            float value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param precision
         *            display precision.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsFloatBase(float[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, short precision,
                String engineeringUnits, float upperDisplayLimit,
                float lowerDisplayLimit, float upperAlarmLimit,
                float upperWarningLimit, float lowerWarningLimit,
                float lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity);
            this.precision = precision;
            this.engineeringUnits = engineeringUnits;
            assert (charset != null);
            this.charset = charset;
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
            this.upperDisplayLimit = upperDisplayLimit;
            this.lowerDisplayLimit = lowerDisplayLimit;
            this.upperAlarmLimit = upperAlarmLimit;
            this.upperWarningLimit = upperWarningLimit;
            this.lowerWarningLimit = lowerWarningLimit;
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsFloatBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsFloatBase(Charset charset) {
            this.charset = charset;
        }

        @Override
        public String getUnits() {
            if (delegate != null) {
                return delegate.getUnits();
            }
            return engineeringUnits;
        }

        @Override
        public byte[] getRawUnits() {
            if (delegate != null) {
                return delegate.getRawUnits();
            }
            return rawEngineeringUnits.clone();
        }

        @Override
        public void setUnits(String units) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.engineeringUnits = units;
            // Update raw representation.
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
        }

        @Override
        public void setRawUnits(byte[] rawUnits) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            // We use a new array for storing the raw units. This way, we do
            // not have to copy the array in the clone() method and we do not
            // have to fill the array with null bytes explicitly.
            this.rawEngineeringUnits = new byte[ENGINEERING_UNITS_STRING_SIZE];
            // Copy first seven bytes.
            int copyLength = Math.min(rawUnits.length,
                    ENGINEERING_UNITS_STRING_SIZE - 1);
            System.arraycopy(rawUnits, 0, this.rawEngineeringUnits, 0,
                    copyLength);
            // Update string.
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
        }

        @Override
        public Float getGenericUpperDisplayLimit() {
            return getUpperDisplayLimit();
        }

        @Override
        public Float getGenericLowerDisplayLimit() {
            return getLowerDisplayLimit();
        }

        @Override
        public Float getGenericUpperAlarmLimit() {
            return getUpperAlarmLimit();
        }

        @Override
        public Float getGenericUpperWarningLimit() {
            return getUpperWarningLimit();
        }

        @Override
        public Float getGenericLowerWarningLimit() {
            return getLowerWarningLimit();
        }

        @Override
        public Float getGenericLowerAlarmLimit() {
            return getLowerAlarmLimit();
        }

        @Override
        public Charset getCharset() {
            if (delegate != null) {
                return delegate.getCharset();
            }
            return charset;
        }

        @Override
        public float getUpperDisplayLimit() {
            if (delegate != null) {
                return delegate.getUpperDisplayLimit();
            }
            return upperDisplayLimit;
        }

        @Override
        public void setUpperDisplayLimit(float upperDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperDisplayLimit = upperDisplayLimit;
        }

        @Override
        public float getLowerDisplayLimit() {
            if (delegate != null) {
                return delegate.getLowerDisplayLimit();
            }
            return lowerDisplayLimit;
        }

        @Override
        public void setLowerDisplayLimit(float lowerDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerDisplayLimit = lowerDisplayLimit;
        }

        @Override
        public float getUpperAlarmLimit() {
            if (delegate != null) {
                return delegate.getUpperAlarmLimit();
            }
            return upperAlarmLimit;
        }

        @Override
        public void setUpperAlarmLimit(float upperAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperAlarmLimit = upperAlarmLimit;
        }

        @Override
        public float getUpperWarningLimit() {
            if (delegate != null) {
                return delegate.getUpperWarningLimit();
            }
            return upperWarningLimit;
        }

        @Override
        public void setUpperWarningLimit(float upperWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperWarningLimit = upperWarningLimit;
        }

        @Override
        public float getLowerWarningLimit() {
            if (delegate != null) {
                return delegate.getLowerWarningLimit();
            }
            return lowerWarningLimit;
        }

        @Override
        public void setLowerWarningLimit(float lowerWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerWarningLimit = lowerWarningLimit;
        }

        @Override
        public float getLowerAlarmLimit() {
            if (delegate != null) {
                return delegate.getLowerAlarmLimit();
            }
            return lowerAlarmLimit;
        }

        @Override
        public void setLowerAlarmLimit(float lowerAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        @Override
        public short getPrecision() {
            if (delegate != null) {
                return delegate.getPrecision();
            }
            return precision;
        }

        @Override
        public void setPrecision(short precision) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.precision = precision;
        }

        /**
         * Deserializes the display information from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeGraphics(ByteSource byteSource) {
            this.precision = byteSource.getShort();
            // Skip two extra padding bytes.
            byteSource.skip(2);
            this.rawEngineeringUnits = byteSource
                    .getByteArray(ENGINEERING_UNITS_STRING_SIZE);
            // Ensure that engineering units are null terminated.
            this.rawEngineeringUnits[ENGINEERING_UNITS_STRING_SIZE - 1] = 0;
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
            this.upperDisplayLimit = byteSource.getFloat();
            this.lowerDisplayLimit = byteSource.getFloat();
            this.upperAlarmLimit = byteSource.getFloat();
            this.upperWarningLimit = byteSource.getFloat();
            this.lowerWarningLimit = byteSource.getFloat();
            this.lowerAlarmLimit = byteSource.getFloat();
        }

        /**
         * Serializes the display information.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeGraphics(ByteSink byteSink) {
            byteSink.putShort(precision);
            // Insert two extra padding bytes.
            byteSink.putShort((short) 0);
            byteSink.putByteArray(rawEngineeringUnits);
            byteSink.putFloat(upperDisplayLimit);
            byteSink.putFloat(lowerDisplayLimit);
            byteSink.putFloat(upperAlarmLimit);
            byteSink.putFloat(upperWarningLimit);
            byteSink.putFloat(lowerWarningLimit);
            byteSink.putFloat(lowerAlarmLimit);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessGraphicsFloatBase<ImplementationType> other = (ChannelAccessGraphicsFloatBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.precision, other.precision)
                    .append(this.engineeringUnits, other.engineeringUnits)
                    .append(this.rawEngineeringUnits, other.rawEngineeringUnits)
                    .append(this.upperDisplayLimit, other.upperDisplayLimit)
                    .append(this.lowerDisplayLimit, other.lowerDisplayLimit)
                    .append(this.upperAlarmLimit, other.upperAlarmLimit)
                    .append(this.upperWarningLimit, other.upperWarningLimit)
                    .append(this.lowerWarningLimit, other.lowerWarningLimit)
                    .append(this.lowerAlarmLimit, other.lowerAlarmLimit)
                    .append(this.charset, other.charset).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.precision).append(this.engineeringUnits)
                    .append(this.rawEngineeringUnits)
                    .append(this.upperDisplayLimit)
                    .append(this.lowerDisplayLimit)
                    .append(this.upperAlarmLimit)
                    .append(this.upperWarningLimit)
                    .append(this.lowerWarningLimit)
                    .append(this.lowerAlarmLimit).append(this.charset)
                    .toHashCode();
        }

        @Override
        public ChannelAccessGraphicsFloat clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            // We do not have to do anything special because we never modify
            // the rawEngineeringUnits array.
            return (ChannelAccessGraphicsFloat) super.clone();
        }

    }

    /**
     * Implementation of DBR_GR_FLOAT.
     */
    public static final class ChannelAccessGraphicsFloatImpl extends
            ChannelAccessGraphicsFloatBase<ChannelAccessGraphicsFloatImpl>
            implements ChannelAccessGraphicsOnlyFloat {

        /**
         * Creates a DBR_GR_FLOAT.
         * 
         * @param value
         *            float value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param precision
         *            display precision.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessGraphicsFloatImpl(float[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, short precision,
                String engineeringUnits, float upperDisplayLimit,
                float lowerDisplayLimit, float upperAlarmLimit,
                float upperWarningLimit, float lowerWarningLimit,
                float lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, precision,
                    engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                    upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                    lowerAlarmLimit, charset);
        }

        /**
         * Creates a read-only DBR_GR_DOUBLE. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsFloatImpl(
                ChannelAccessGraphicsFloatImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsFloatImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_GR_FLOAT;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessGraphicsOnlyFloat asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessGraphicsFloatImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessGraphicsOnlyFloat clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessGraphicsOnlyFloat) super.clone();
        }

        /**
         * Creates a DBR_GR_FLOAT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_GR_FLOAT instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessGraphicsFloatImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessGraphicsFloatImpl value = new ChannelAccessGraphicsFloatImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_CTRL_FLOAT.
     */
    public static final class ChannelAccessControlsFloatImpl extends
            ChannelAccessGraphicsFloatBase<ChannelAccessControlsFloatImpl>
            implements ChannelAccessControlsFloat {

        /**
         * Upper control limit.
         */
        protected float upperControlLimit;

        /**
         * Lower control limit.
         */
        protected float lowerControlLimit;

        /**
         * Creates a DBR_CTRL_FLOAT.
         * 
         * @param value
         *            float value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param precision
         *            display precision.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param upperControlLimit
         *            upper control limit.
         * @param lowerControlLimit
         *            lower control limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessControlsFloatImpl(float[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, short precision,
                String engineeringUnits, float upperDisplayLimit,
                float lowerDisplayLimit, float upperAlarmLimit,
                float upperWarningLimit, float lowerWarningLimit,
                float lowerAlarmLimit, float upperControlLimit,
                float lowerControlLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, precision,
                    engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                    upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                    lowerAlarmLimit, charset);
            this.upperControlLimit = upperControlLimit;
            this.lowerControlLimit = lowerControlLimit;
        }

        /**
         * Creates a read-only DBR_GR_FLOAT. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessControlsFloatImpl(
                ChannelAccessControlsFloatImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessControlsFloatImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_CTRL_FLOAT;
        }

        @Override
        public Float getGenericUpperControlLimit() {
            return getUpperControlLimit();
        }

        @Override
        public Float getGenericLowerControlLimit() {
            return getLowerControlLimit();
        }

        @Override
        public float getUpperControlLimit() {
            if (delegate != null) {
                return delegate.getUpperControlLimit();
            }
            return upperControlLimit;
        }

        @Override
        public void setUpperControlLimit(float upperControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperControlLimit = upperControlLimit;
        }

        @Override
        public float getLowerControlLimit() {
            if (delegate != null) {
                return delegate.getLowerControlLimit();
            }
            return lowerControlLimit;
        }

        @Override
        public void setLowerControlLimit(float lowerControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerControlLimit = lowerControlLimit;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            byteSink.putFloat(upperControlLimit);
            byteSink.putFloat(lowerControlLimit);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessControlsFloat asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessControlsFloatImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessControlsFloatImpl other = (ChannelAccessControlsFloatImpl) obj;
            return new EqualsBuilder()
                    .append(this.upperControlLimit, other.upperControlLimit)
                    .append(this.lowerControlLimit, other.lowerControlLimit)
                    .isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.upperControlLimit)
                    .append(this.lowerControlLimit).toHashCode();
        }

        @Override
        public ChannelAccessControlsFloat clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessControlsFloat) super.clone();
        }

        /**
         * Creates a DBR_CTRL_FLOAT from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_CTRL_FLOAT instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessControlsFloatImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessControlsFloatImpl value = new ChannelAccessControlsFloatImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.upperControlLimit = byteSource.getFloat();
            value.lowerControlLimit = byteSource.getFloat();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    protected float[] value;

    /**
     * Base constructor.
     * 
     * @param value
     *            float value.
     */
    private ChannelAccessFloatBase(float[] value) {
        assert (value != null);
        this.value = value;
    }

    /**
     * Base constructor for read-only variant.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessFloatBase(ImplementationType delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     */
    private ChannelAccessFloatBase() {
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.length;
    }

    @Override
    public Float getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value[index];
    }

    @Override
    public FloatBuffer getValue() {
        if (delegate != null) {
            return delegate.getValue().asReadOnlyBuffer();
        }
        return FloatBuffer.wrap(value);
    }

    @Override
    public void setValue(float[] value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value == null) {
            throw new NullPointerException();
        }
        this.value = value;
    }

    /**
     * Deserializes the value elements from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @param count
     *            number of elements to deserialize.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    protected void deserializeValue(ByteSource byteSource, int count) {
        assert (count >= 0);
        if (count == 0) {
            value = ArrayUtils.EMPTY_FLOAT_ARRAY;
        } else {
            value = byteSource.getFloatArray(count);
        }
    }

    /**
     * Serializes the value elements to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write to.
     * @param count
     *            number of elements to be serialized.
     */
    protected void serializeValue(ByteSink byteSink, int count) {
        if (count != 0) {
            byteSink.putFloatArray(value, 0, count);
        }
    }

    @Override
    public boolean equals(Object obj) {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been
        // checked.
        @SuppressWarnings("unchecked")
        ChannelAccessFloatBase<ImplementationType> other = (ChannelAccessFloatBase<ImplementationType>) obj;
        return new EqualsBuilder().append(this.value, other.value).isEquals();
    }

    @Override
    public int hashCode() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(value).toHashCode();
    }

    @Override
    public ChannelAccessFloat clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        @SuppressWarnings("unchecked")
        ChannelAccessFloatBase<ImplementationType> copy = (ChannelAccessFloatBase<ImplementationType>) super
                .clone();
        copy.value = this.value.clone();
        return copy;
    }

}
