/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.IntBuffer;
import java.nio.charset.Charset;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeStamp;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Base class for Channel Access char type. This class contains the various
 * specializations of the char type as inlined classes.
 * 
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessLongBase<ImplementationType extends ChannelAccessLongBase<ImplementationType>>
        extends ChannelAccessValueBase<Integer, ImplementationType> implements
        ChannelAccessLong {

    /**
     * Implementation of DBR_LONG.
     */
    public static final class ChannelAccessLongImpl extends
            ChannelAccessLongBase<ChannelAccessLongImpl> implements
            ChannelAccessSimpleOnlyLong {

        /**
         * Creates a DBR_LONG.
         * 
         * @param value
         *            char value.
         */
        public ChannelAccessLongImpl(int[] value) {
            super(value);
        }

        /**
         * Creates a read-only DBR_LONG. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessLongImpl(ChannelAccessLongImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessLongImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_LONG;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessSimpleOnlyLong asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessLongImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessSimpleOnlyLong clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessSimpleOnlyLong) super.clone();
        }

        /**
         * Creates a DBR_LONG from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_LONG instance created from the serialized data read from
         *         <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessLongImpl deserialize(ByteSource byteSource,
                int count) {
            ChannelAccessLongImpl value = new ChannelAccessLongImpl();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for char with alarms.
     */
    public static abstract class ChannelAccessAlarmLongBase<ImplementationType extends ChannelAccessAlarmLongBase<ImplementationType>>
            extends ChannelAccessLongBase<ImplementationType> implements
            ChannelAccessAlarmLong {

        /**
         * Alarm status.
         */
        protected ChannelAccessAlarmStatus alarmStatus;

        /**
         * Alarm severity.
         */
        protected ChannelAccessAlarmSeverity alarmSeverity;

        /**
         * Base constructor.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        private ChannelAccessAlarmLongBase(int[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value);
            this.alarmStatus = alarmStatus;
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmLongBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmLongBase() {
        }

        @Override
        public ChannelAccessAlarmStatus getAlarmStatus() {
            if (delegate != null) {
                return delegate.getAlarmStatus();
            }
            return this.alarmStatus;
        }

        @Override
        public void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmStatus != null);
            this.alarmStatus = alarmStatus;
        }

        @Override
        public ChannelAccessAlarmSeverity getAlarmSeverity() {
            if (delegate != null) {
                return delegate.getAlarmSeverity();
            }
            return this.alarmSeverity;
        }

        @Override
        public void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmSeverity != null);
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Deserializes the alarm status from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeAlarms(ByteSource byteSource) {
            this.alarmStatus = ValueCodecUtils
                    .deserializeAlarmStatus(byteSource);
            this.alarmSeverity = ValueCodecUtils
                    .deserializeAlarmSeverity(byteSource);
        }

        /**
         * Serializes the alarm status to the byte sink.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeAlarms(ByteSink byteSink) {
            ValueCodecUtils.serializeAlarms(byteSink, alarmStatus,
                    alarmSeverity);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessAlarmLongBase<ImplementationType> other = (ChannelAccessAlarmLongBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.alarmStatus, other.alarmStatus)
                    .append(this.alarmSeverity, other.alarmSeverity).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.alarmStatus).append(this.alarmSeverity)
                    .toHashCode();
        }

        @Override
        public ChannelAccessAlarmLong clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return (ChannelAccessAlarmLong) super.clone();
        }

    }

    /**
     * Implementation of DBR_STS_LONG.
     */
    public static final class ChannelAccessAlarmLongImpl extends
            ChannelAccessAlarmLongBase<ChannelAccessAlarmLongImpl> implements
            ChannelAccessAlarmOnlyLong {

        /**
         * Creates a DBR_STS_LONG.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        public ChannelAccessAlarmLongImpl(int[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value, alarmStatus, alarmSeverity);
        }

        /**
         * Creates a read-only DBR_STS_LONG. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmLongImpl(ChannelAccessAlarmLongImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmLongImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_STS_LONG;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessAlarmOnlyLong asReadOnlyValue() {
            if (delegate != null) {
                return this;
            } else {
                return new ChannelAccessAlarmLongImpl(this);
            }
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessAlarmOnlyLong clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessAlarmOnlyLong) super.clone();
        }

        /**
         * Creates a DBR_STS_LONG from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_STS_LONG instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessAlarmLongImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessAlarmLongImpl value = new ChannelAccessAlarmLongImpl();
            value.deserializeAlarms(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_TIME_LONG.
     */
    public static final class ChannelAccessTimeLongImpl extends
            ChannelAccessAlarmLongBase<ChannelAccessTimeLongImpl> implements
            ChannelAccessTimeLong {

        /**
         * Seconds part of time-stamp.
         */
        protected int timeStampSeconds;

        /**
         * Nanoseconds part of time-stamp.
         */
        protected int timeStampNanoseconds;

        /**
         * Creates a DBR_TIME_LONG.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param timeStampSeconds
         *            seconds part of time-stamp.
         * @param timeStampNanoSeconds
         *            nanoseconds part of time-stamp.
         */
        public ChannelAccessTimeLongImpl(int[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
                int timeStampNanoSeconds) {
            super(value, alarmStatus, alarmSeverity);
            this.timeStampSeconds = timeStampSeconds;
            if (timeStampNanoSeconds < 0 || timeStampNanoSeconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of timestamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoSeconds;
        }

        /**
         * Creates a read-only DBR_TIME_LONG. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessTimeLongImpl(ChannelAccessTimeLongImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessTimeLongImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_TIME_LONG;
        }

        @Override
        public ChannelAccessTimeStamp getTimeStamp() {
            if (delegate != null) {
                return delegate.getTimeStamp();
            }
            return ChannelAccessValueFactory.createTimeStamp(timeStampSeconds,
                    timeStampNanoseconds);
        }

        @Override
        public int getTimeSeconds() {
            if (delegate != null) {
                return delegate.getTimeSeconds();
            }
            return timeStampSeconds;
        }

        @Override
        public int getTimeNanoseconds() {
            if (delegate != null) {
                return delegate.getTimeNanoseconds();
            }
            return timeStampNanoseconds;
        }

        @Override
        public void setTimeStamp(ChannelAccessTimeStamp timeStamp) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStamp.getSeconds();
            this.timeStampNanoseconds = timeStamp.getNanoseconds();
        }

        @Override
        public void setTimeSeconds(int timeStampSeconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStampSeconds;
        }

        @Override
        public void setTimeNanoseconds(int timeStampNanoseconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (timeStampNanoseconds < 0 || timeStampNanoseconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of time-stamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoseconds;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            ValueCodecUtils.serializeTime(byteSink, timeStampSeconds,
                    timeStampNanoseconds);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessTimeLong asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessTimeLongImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessTimeLongImpl other = (ChannelAccessTimeLongImpl) obj;
            return new EqualsBuilder()
                    .append(this.timeStampSeconds, other.timeStampSeconds)
                    .append(this.timeStampNanoseconds,
                            other.timeStampNanoseconds).isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.timeStampSeconds)
                    .append(this.timeStampNanoseconds).toHashCode();
        }

        @Override
        public ChannelAccessTimeLong clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessTimeLong) super.clone();
        }

        /**
         * Creates a DBR_TIME_LONG from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_TIME_LONG instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessTimeLongImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessTimeLongImpl value = new ChannelAccessTimeLongImpl();
            value.deserializeAlarms(byteSource);
            // Deserialize time-stamp.
            value.timeStampSeconds = ValueCodecUtils
                    .deserializeTimeSeconds(byteSource);
            value.timeStampNanoseconds = ValueCodecUtils
                    .deserializeTimeNanoseconds(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for char with display information.
     */
    public static abstract class ChannelAccessGraphicsLongBase<ImplementationType extends ChannelAccessGraphicsLongBase<ImplementationType>>
            extends ChannelAccessAlarmLongBase<ImplementationType> implements
            ChannelAccessGraphicsLong {

        private static final int ENGINEERING_UNITS_STRING_SIZE = 8;

        /**
         * Engineering units.
         */
        protected String engineeringUnits;

        /**
         * Raw representation of engineering units.
         */
        protected byte[] rawEngineeringUnits;

        /**
         * Upper display limit.
         */
        protected int upperDisplayLimit;

        /**
         * Lower display limit.
         */
        protected int lowerDisplayLimit;

        /**
         * Upper alarm limit.
         */
        protected int upperAlarmLimit;

        /**
         * Upper warning limit.
         */
        protected int upperWarningLimit;

        /**
         * Lower warning limit.
         */
        protected int lowerWarningLimit;

        /**
         * Lower alarm limit.
         */
        protected int lowerAlarmLimit;

        /**
         * Charset used for encoding/decoding strings.
         */
        protected Charset charset;

        /**
         * Base constructor.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsLongBase(int[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, int upperDisplayLimit,
                int lowerDisplayLimit, int upperAlarmLimit,
                int upperWarningLimit, int lowerWarningLimit,
                int lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity);
            this.engineeringUnits = engineeringUnits;
            assert (charset != null);
            this.charset = charset;
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
            this.upperDisplayLimit = upperDisplayLimit;
            this.lowerDisplayLimit = lowerDisplayLimit;
            this.upperAlarmLimit = upperAlarmLimit;
            this.upperWarningLimit = upperWarningLimit;
            this.lowerWarningLimit = lowerWarningLimit;
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsLongBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsLongBase(Charset charset) {
            this.charset = charset;
        }

        @Override
        public String getUnits() {
            if (delegate != null) {
                return delegate.getUnits();
            }
            return engineeringUnits;
        }

        @Override
        public byte[] getRawUnits() {
            if (delegate != null) {
                return delegate.getRawUnits();
            }
            return rawEngineeringUnits.clone();
        }

        @Override
        public void setUnits(String units) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.engineeringUnits = units;
            // Update raw representation.
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
        }

        @Override
        public void setRawUnits(byte[] rawUnits) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            // We use a new array for storing the raw units. This way, we do
            // not have to copy the array in the clone() method and we do not
            // have to fill the array with null bytes explicitly.
            this.rawEngineeringUnits = new byte[ENGINEERING_UNITS_STRING_SIZE];
            // Copy first seven bytes.
            int copyLength = Math.min(rawUnits.length,
                    ENGINEERING_UNITS_STRING_SIZE - 1);
            System.arraycopy(rawUnits, 0, this.rawEngineeringUnits, 0,
                    copyLength);
            // Update string.
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
        }

        @Override
        public Integer getGenericUpperDisplayLimit() {
            return getUpperDisplayLimit();
        }

        @Override
        public Integer getGenericLowerDisplayLimit() {
            return getLowerDisplayLimit();
        }

        @Override
        public Integer getGenericUpperAlarmLimit() {
            return getUpperAlarmLimit();
        }

        @Override
        public Integer getGenericUpperWarningLimit() {
            return getUpperWarningLimit();
        }

        @Override
        public Integer getGenericLowerWarningLimit() {
            return getLowerWarningLimit();
        }

        @Override
        public Integer getGenericLowerAlarmLimit() {
            return getLowerAlarmLimit();
        }

        @Override
        public Charset getCharset() {
            if (delegate != null) {
                return delegate.getCharset();
            }
            return charset;
        }

        @Override
        public int getUpperDisplayLimit() {
            if (delegate != null) {
                return delegate.getUpperDisplayLimit();
            }
            return upperDisplayLimit;
        }

        @Override
        public void setUpperDisplayLimit(int upperDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperDisplayLimit = upperDisplayLimit;
        }

        @Override
        public int getLowerDisplayLimit() {
            if (delegate != null) {
                return delegate.getLowerDisplayLimit();
            }
            return lowerDisplayLimit;
        }

        @Override
        public void setLowerDisplayLimit(int lowerDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerDisplayLimit = lowerDisplayLimit;
        }

        @Override
        public int getUpperAlarmLimit() {
            if (delegate != null) {
                return delegate.getUpperAlarmLimit();
            }
            return upperAlarmLimit;
        }

        @Override
        public void setUpperAlarmLimit(int upperAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperAlarmLimit = upperAlarmLimit;
        }

        @Override
        public int getUpperWarningLimit() {
            if (delegate != null) {
                return delegate.getUpperWarningLimit();
            }
            return upperWarningLimit;
        }

        @Override
        public void setUpperWarningLimit(int upperWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperWarningLimit = upperWarningLimit;
        }

        @Override
        public int getLowerWarningLimit() {
            if (delegate != null) {
                return delegate.getLowerWarningLimit();
            }
            return lowerWarningLimit;
        }

        @Override
        public void setLowerWarningLimit(int lowerWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerWarningLimit = lowerWarningLimit;
        }

        @Override
        public int getLowerAlarmLimit() {
            if (delegate != null) {
                return delegate.getLowerAlarmLimit();
            }
            return lowerAlarmLimit;
        }

        @Override
        public void setLowerAlarmLimit(int lowerAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        /**
         * Deserializes the display information from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeGraphics(ByteSource byteSource) {
            this.rawEngineeringUnits = byteSource
                    .getByteArray(ENGINEERING_UNITS_STRING_SIZE);
            // Ensure that engineering units are null terminated.
            this.rawEngineeringUnits[ENGINEERING_UNITS_STRING_SIZE - 1] = 0;
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
            this.upperDisplayLimit = byteSource.getInt();
            this.lowerDisplayLimit = byteSource.getInt();
            this.upperAlarmLimit = byteSource.getInt();
            this.upperWarningLimit = byteSource.getInt();
            this.lowerWarningLimit = byteSource.getInt();
            this.lowerAlarmLimit = byteSource.getInt();
        }

        /**
         * Serializes the display information.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeGraphics(ByteSink byteSink) {
            byteSink.putByteArray(rawEngineeringUnits);
            byteSink.putInt(upperDisplayLimit);
            byteSink.putInt(lowerDisplayLimit);
            byteSink.putInt(upperAlarmLimit);
            byteSink.putInt(upperWarningLimit);
            byteSink.putInt(lowerWarningLimit);
            byteSink.putInt(lowerAlarmLimit);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessGraphicsLongBase<ImplementationType> other = (ChannelAccessGraphicsLongBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.engineeringUnits, other.engineeringUnits)
                    .append(this.rawEngineeringUnits, other.rawEngineeringUnits)
                    .append(this.upperDisplayLimit, other.upperDisplayLimit)
                    .append(this.lowerDisplayLimit, other.lowerDisplayLimit)
                    .append(this.upperAlarmLimit, other.upperAlarmLimit)
                    .append(this.upperWarningLimit, other.upperWarningLimit)
                    .append(this.lowerWarningLimit, other.lowerWarningLimit)
                    .append(this.lowerAlarmLimit, other.lowerAlarmLimit)
                    .append(this.charset, other.charset).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.engineeringUnits)
                    .append(this.rawEngineeringUnits)
                    .append(this.upperDisplayLimit)
                    .append(this.lowerDisplayLimit)
                    .append(this.upperAlarmLimit)
                    .append(this.upperWarningLimit)
                    .append(this.lowerWarningLimit)
                    .append(this.lowerAlarmLimit).append(this.charset)
                    .toHashCode();
        }

        @Override
        public ChannelAccessGraphicsLong clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            // We do not have to do anything special because we never modify
            // the rawEngineeringUnits array.
            return (ChannelAccessGraphicsLong) super.clone();
        }

    }

    /**
     * Implementation of DBR_GR_LONG.
     */
    public static final class ChannelAccessGraphicsLongImpl extends
            ChannelAccessGraphicsLongBase<ChannelAccessGraphicsLongImpl>
            implements ChannelAccessGraphicsOnlyLong {

        /**
         * Creates a DBR_GR_LONG.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessGraphicsLongImpl(int[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, int upperDisplayLimit,
                int lowerDisplayLimit, int upperAlarmLimit,
                int upperWarningLimit, int lowerWarningLimit,
                int lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, engineeringUnits,
                    upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                    upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                    charset);
        }

        /**
         * Creates a read-only DBR_GR_LONG. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsLongImpl(
                ChannelAccessGraphicsLongImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsLongImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_GR_LONG;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessGraphicsOnlyLong asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessGraphicsLongImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessGraphicsOnlyLong clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessGraphicsOnlyLong) super.clone();
        }

        /**
         * Creates a DBR_GR_LONG from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_GR_LONG instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessGraphicsLongImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessGraphicsLongImpl value = new ChannelAccessGraphicsLongImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_CTRL_LONG.
     */
    public static final class ChannelAccessControlsLongImpl extends
            ChannelAccessGraphicsLongBase<ChannelAccessControlsLongImpl>
            implements ChannelAccessControlsLong {

        /**
         * Upper control limit.
         */
        protected int upperControlLimit;

        /**
         * Lower control limit.
         */
        protected int lowerControlLimit;

        /**
         * Creates a DBR_CTRL_LONG.
         * 
         * @param value
         *            char value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param upperControlLimit
         *            upper control limit.
         * @param lowerControlLimit
         *            lower control limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessControlsLongImpl(int[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity,
                String engineeringUnits, int upperDisplayLimit,
                int lowerDisplayLimit, int upperAlarmLimit,
                int upperWarningLimit, int lowerWarningLimit,
                int lowerAlarmLimit, int upperControlLimit,
                int lowerControlLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, engineeringUnits,
                    upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                    upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                    charset);
            this.upperControlLimit = upperControlLimit;
            this.lowerControlLimit = lowerControlLimit;
        }

        /**
         * Creates a read-only DBR_GR_LONG. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessControlsLongImpl(
                ChannelAccessControlsLongImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessControlsLongImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_CTRL_LONG;
        }

        @Override
        public Integer getGenericUpperControlLimit() {
            return getUpperControlLimit();
        }

        @Override
        public Integer getGenericLowerControlLimit() {
            return getLowerControlLimit();
        }

        @Override
        public int getUpperControlLimit() {
            if (delegate != null) {
                return delegate.getUpperControlLimit();
            }
            return upperControlLimit;
        }

        @Override
        public void setUpperControlLimit(int upperControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperControlLimit = upperControlLimit;
        }

        @Override
        public int getLowerControlLimit() {
            if (delegate != null) {
                return delegate.getLowerControlLimit();
            }
            return lowerControlLimit;
        }

        @Override
        public void setLowerControlLimit(int lowerControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerControlLimit = lowerControlLimit;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            byteSink.putInt(upperControlLimit);
            byteSink.putInt(lowerControlLimit);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessControlsLong asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessControlsLongImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessControlsLongImpl other = (ChannelAccessControlsLongImpl) obj;
            return new EqualsBuilder()
                    .append(this.upperControlLimit, other.upperControlLimit)
                    .append(this.lowerControlLimit, other.lowerControlLimit)
                    .isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.upperControlLimit)
                    .append(this.lowerControlLimit).toHashCode();
        }

        @Override
        public ChannelAccessControlsLong clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessControlsLong) super.clone();
        }

        /**
         * Creates a DBR_CTRL_LONG from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_CTRL_LONG instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessControlsLongImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessControlsLongImpl value = new ChannelAccessControlsLongImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.upperControlLimit = byteSource.getInt();
            value.lowerControlLimit = byteSource.getInt();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    protected int[] value;

    /**
     * Base constructor.
     * 
     * @param value
     *            char value.
     */
    private ChannelAccessLongBase(int[] value) {
        assert (value != null);
        this.value = value;
    }

    /**
     * Base constructor for read-only variant.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessLongBase(ImplementationType delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     */
    private ChannelAccessLongBase() {
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.length;
    }

    @Override
    public Integer getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value[index];
    }

    @Override
    public IntBuffer getValue() {
        if (delegate != null) {
            return delegate.getValue().asReadOnlyBuffer();
        }
        return IntBuffer.wrap(value);
    }

    @Override
    public void setValue(int[] value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value == null) {
            throw new NullPointerException();
        }
        this.value = value;
    }

    /**
     * Deserializes the value elements from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @param count
     *            number of elements to deserialize.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    protected void deserializeValue(ByteSource byteSource, int count) {
        assert (count >= 0);
        if (count == 0) {
            value = ArrayUtils.EMPTY_INT_ARRAY;
        } else {
            value = byteSource.getIntArray(count);
        }
    }

    /**
     * Serializes the value elements to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write to.
     * @param count
     *            number of elements to be serialized.
     */
    protected void serializeValue(ByteSink byteSink, int count) {
        if (count != 0) {
            byteSink.putIntArray(value, 0, count);
        }
    }

    @Override
    public boolean equals(Object obj) {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been
        // checked.
        @SuppressWarnings("unchecked")
        ChannelAccessLongBase<ImplementationType> other = (ChannelAccessLongBase<ImplementationType>) obj;
        return new EqualsBuilder().append(this.value, other.value).isEquals();
    }

    @Override
    public int hashCode() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(value).toHashCode();
    }

    @Override
    public ChannelAccessLong clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        @SuppressWarnings("unchecked")
        ChannelAccessLongBase<ImplementationType> copy = (ChannelAccessLongBase<ImplementationType>) super
                .clone();
        copy.value = this.value.clone();
        return copy;
    }

}
