/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import static org.junit.Assert.*;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.Collections;

import org.apache.commons.lang3.ArrayUtils;
import org.junit.Test;

import com.aquenos.epics.jackie.common.io.ByteBufferByteSink;
import com.aquenos.epics.jackie.common.io.ByteBufferByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Tests for the {@link ChannelAccessMessageCodec}.
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessMessageCodecTest {

    private static final Charset UTF8 = StandardCharsets.UTF_8;
    private static final Inet4Address LOCALHOST;
    static {
        try {
            LOCALHOST = (Inet4Address) InetAddress.getByName("127.0.0.1");
        } catch (UnknownHostException e) {
            throw new RuntimeException("Could not create IP address 127.0.0.1.",
                    e);
        }
    }

    private final ChannelAccessMessageCodec codec = new ChannelAccessMessageCodec();
    private final ChannelAccessVersion version = ChannelAccessVersion.V4_13;

    private static ByteBufferByteSink createByteSink() {
        ByteBufferByteSink byteSink = new ByteBufferByteSink();
        byteSink.addBuffer(ByteBuffer.allocate(8192));
        return byteSink;
    }

    private static ByteBufferByteSource createByteSource(
            ByteBufferByteSink byteSink) {
        return createByteSource(byteSink, 0);
    }

    private static ByteBufferByteSource createByteSource(
            ByteBufferByteSink byteSink, int cutPayloadBytes) {
        ByteBufferByteSource byteSource = new ByteBufferByteSource();
        if (cutPayloadBytes > 0) {
            ByteBuffer[] buffers = byteSink.getWrittenData();
            // For our tests, we always use a byte source with a single buffer.
            // Therefore, we can change the message quite easily.
            assert buffers.length == 1;
            ByteBuffer buffer = buffers[0];
            // This is a hack to change the payload size: We know the layout of
            // the header of a CA message, so we can directly change the
            // payload field.
            short payloadSizeField = buffer.getShort(2);
            if (payloadSizeField == (short) 0xffff) {
                // The message uses the “large header” extension.
                int largePayloadSizeField = buffer.getInt(16);
                assert largePayloadSizeField >= cutPayloadBytes;
                buffer.putInt(16, largePayloadSizeField - cutPayloadBytes);
            } else {
                assert payloadSizeField >= cutPayloadBytes;
                buffer.putShort(2,
                        (short) (payloadSizeField - cutPayloadBytes));
            }
            buffer.limit(buffer.limit() - cutPayloadBytes);
            byteSource.appendBuffer(buffer);
        } else {
            byteSource.appendBuffers(byteSink.getWrittenData());
        }
        return byteSource;
    }

    private ChannelAccessMessage reserializeMessageFromRepeater(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageFromRepeater(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec
                .decodeMessageFromRepeater(byteSource, version);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToRepeater(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToRepeater(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec
                .decodeMessageToRepeater(byteSource, version, LOCALHOST);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToTCPClient(
            ChannelAccessMessage message) {
        return reserializeMessageToTCPClient(message, 0);
    }

    private ChannelAccessMessage reserializeMessageToTCPClient(
            ChannelAccessMessage message, int cutPayloadBytes) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToTCPClient(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink,
                cutPayloadBytes);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec
                .decodeMessageToTCPClient(byteSource, version, LOCALHOST, 1234);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToTCPServer(
            ChannelAccessMessage message) {
        return reserializeMessageToTCPServer(message, 0);
    }

    private ChannelAccessMessage reserializeMessageToTCPServer(
            ChannelAccessMessage message, int cutPayloadBytes) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToTCPServer(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink,
                cutPayloadBytes);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec
                .decodeMessageToTCPServer(byteSource, version);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToUDPClient(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToUDPClient(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec
                .decodeMessageToUDPClient(byteSource, version, LOCALHOST, 1234);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToUDPServer(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToUDPServer(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec
                .decodeMessageToUDPServer(byteSource, version);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    /**
     * Encode and decode a message that has a large header (using the extended
     * header format). This is particularly designed to detect a bug where a
     * payload size of 0xffff (that signals a large header) would be treated
     * incorrectly.
     */
    @Test
    public void largeHeader() {
        // We use a create channel message (server to client), but any other
        // message that can have a large header shares the same code-path
        // regarding header processing.
        // When the count is greater than 65535, the extended header format is
        // used. We test a few different numbers to ensure that we handle
        // corner cases correctly.
        ChannelAccessConnectChannelServerMessage message;
        message = new ChannelAccessConnectChannelServerMessage(55, 23,
                ChannelAccessValueType.DBR_LONG, 65534);
        assertEquals(message, reserializeMessageToTCPClient(message));
        message = new ChannelAccessConnectChannelServerMessage(55, 23,
                ChannelAccessValueType.DBR_LONG, 65535);
        assertEquals(message, reserializeMessageToTCPClient(message));
        message = new ChannelAccessConnectChannelServerMessage(55, 23,
                ChannelAccessValueType.DBR_LONG, 65536);
        assertEquals(message, reserializeMessageToTCPClient(message));
        message = new ChannelAccessConnectChannelServerMessage(55, 23,
                ChannelAccessValueType.DBR_LONG, 4000000);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode an error message that is in response to a message with
     * a large header (using the extended header format). This is particularly
     * designed to detect a bug where a payload size of 0xffff in the original
     * message header would be treated incorrectly.
     */
    @Test
    public void largeHeaderInOriginalMessage() {
        // We need an original message that uses the extended header format.
        // On the other hand, we have to use a message without a payload here,
        // because otherwise the message embedded in the error message will lose
        // its payload and thus the error message sent will not be equal to the
        // error message received.
        // A read message from the client to the server is the best choice
        // because we can use a large count to trigger the extended header
        // format, but still do not have a payload.
        ChannelAccessReadNotifyClientMessage clientMessage = new ChannelAccessReadNotifyClientMessage(
                ChannelAccessValueType.DBR_CTRL_DOUBLE, 131072, 123, 999);
        // We have to reserialize the client message so that the original
        // headers are available.
        clientMessage = (ChannelAccessReadNotifyClientMessage) reserializeMessageToTCPServer(
                clientMessage);
        ChannelAccessErrorMessage message = new ChannelAccessErrorMessage(99,
                ChannelAccessStatus.StatusMessage.ECA_ALLOCMEM.toStatus(),
                clientMessage, "Cannot allocate memory.");
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a REPEATER_CONFIRM message sent by a repeater.
     */
    @Test
    public void repeaterSendRepeaterConfirm() {
        ChannelAccessRepeaterConfirmMessage message = new ChannelAccessRepeaterConfirmMessage(
                LOCALHOST);
        assertEquals(message, reserializeMessageFromRepeater(message));
    }

    /**
     * Encode and decode a CA_PROTO_RSRV_IS_UP message sent by a repeater.
     */
    @Test
    public void repeaterSendBeacon() {
        ChannelAccessBeaconMessage message = new ChannelAccessBeaconMessage(
                LOCALHOST, 1234, version, 42);
        assertEquals(message, reserializeMessageFromRepeater(message));
    }

    /**
     * Encode and decode a CA_PROTO_RSRV_IS_UP message received by a repeater.
     */
    @Test
    public void repeaterReceiveBeacon() {
        ChannelAccessBeaconMessage message = new ChannelAccessBeaconMessage(
                LOCALHOST, 1234, version, 42);
        assertEquals(message, reserializeMessageToRepeater(message));
    }

    /**
     * Encode and decode a CA_PROTO_VERSION message sent by a TCP client.
     */
    @Test
    public void tcpClientVersion() {
        ChannelAccessVersionTCPClientMessage message = new ChannelAccessVersionTCPClientMessage(
                version, (short) 22);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENT_ADD message sent by a TCP client.
     */
    @Test
    public void tcpClientSubscription() {
        ChannelAccessSubscriptionClientMessage message = new ChannelAccessSubscriptionClientMessage(
                ChannelAccessValueType.DBR_CTRL_DOUBLE, 42, -123, 23,
                ChannelAccessEventMask.DBE_VALUE
                        .or(ChannelAccessEventMask.DBE_ALARM));
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENT_CANCEL message sent by a TCP client.
     */
    @Test
    public void tcpClientCancelSubscription() {
        ChannelAccessCancelSubscriptionClientMessage message = new ChannelAccessCancelSubscriptionClientMessage(
                ChannelAccessValueType.DBR_STSACK_STRING, 55, 123, 42);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ message sent by a TCP client.
     */
    @Test
    public void tcpClientRead() {
        ChannelAccessReadClientMessage message = new ChannelAccessReadClientMessage(
                ChannelAccessValueType.DBR_CLASS_NAME, 1, 5, 123);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_WRITE message sent by a TCP client.
     */
    @Test
    public void tcpClientWrite() {
        ChannelAccessWriteMessage message = new ChannelAccessWriteMessage(
                ChannelAccessValueType.DBR_PUT_ACKS, 1, -1, 0,
                ChannelAccessValueFactory.createAcknowledgeAlarm(
                        ChannelAccessAlarmSeverity.MAJOR_ALARM));
        assertEquals(message, reserializeMessageToTCPServer(message));
        message = new ChannelAccessWriteMessage(
                ChannelAccessValueType.DBR_STRING, 0, -1, 0,
                ChannelAccessValueFactory.createString(Collections.emptyList(),
                        UTF8));
        assertEquals(message, reserializeMessageToTCPServer(message));
        // Test that a “short” string works as expected.
        message = new ChannelAccessWriteMessage(
                ChannelAccessValueType.DBR_STRING, 1, -1, 0,
                ChannelAccessValueFactory.createString(
                        Collections.singleton("abc"), UTF8));
        assertEquals(message, reserializeMessageToTCPServer(message, 32));
        message = new ChannelAccessWriteMessage(
                ChannelAccessValueType.DBR_STRING, 1, -1, 0,
                ChannelAccessValueFactory.createString(
                        Collections.singleton(""), UTF8));
        assertEquals(message, reserializeMessageToTCPServer(message, 40));
    }

    /**
     * Encode and decode a CA_PROTO_EVENTS_OFF message sent by a TCP client.
     */
    @Test
    public void tcpClientEventsOff() {
        ChannelAccessEventsOffMessage message = new ChannelAccessEventsOffMessage();
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENTS_ON message sent by a TCP client.
     */
    @Test
    public void tcpClientEventsOn() {
        ChannelAccessEventsOnMessage message = new ChannelAccessEventsOnMessage();
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ_SYNC message sent by a TCP client.
     */
    @Test
    public void tcpClientReadSync() {
        ChannelAccessReadSyncMessage message = new ChannelAccessReadSyncMessage();
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_CLEAR_CHANNEL message sent by a TCP client.
     */
    @Test
    public void tcpClientDisconnectChannel() {
        ChannelAccessDisconnectChannelMessage message = new ChannelAccessDisconnectChannelMessage(
                128, 821);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ_NOTIFY message sent by a TCP client.
     */
    @Test
    public void tcpClientReadNotify() {
        ChannelAccessReadNotifyClientMessage message = new ChannelAccessReadNotifyClientMessage(
                ChannelAccessValueType.DBR_CLASS_NAME, 1, 5, 123);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_CREATE_CHAN message sent by a TCP client.
     */
    @Test
    public void tcpClientConnectChannel() {
        ChannelAccessConnectChannelClientMessage message = new ChannelAccessConnectChannelClientMessage(
                42, version, "myChannel");
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_WRITE_NOTIFY message sent by a TCP client.
     */
    @Test
    public void tcpClientWriteNotify() {
        ChannelAccessWriteNotifyClientMessage message = new ChannelAccessWriteNotifyClientMessage(
                ChannelAccessValueType.DBR_DOUBLE, 4, -1, 0,
                ChannelAccessValueFactory
                        .createDouble(new double[] { 24, 8, 13, 5 }));
        assertEquals(message, reserializeMessageToTCPServer(message));
        message = new ChannelAccessWriteNotifyClientMessage(
                ChannelAccessValueType.DBR_STRING, 0, -1, 0,
                ChannelAccessValueFactory.createString(Collections.emptyList(),
                        UTF8));
        // Test that a “short” string works as expected.
        message = new ChannelAccessWriteNotifyClientMessage(
                ChannelAccessValueType.DBR_STRING, 1, -1, 0,
                ChannelAccessValueFactory.createString(
                        Collections.singleton("Hello World!"), UTF8));
        assertEquals(message, reserializeMessageToTCPServer(message, 24));
        message = new ChannelAccessWriteNotifyClientMessage(
                ChannelAccessValueType.DBR_STRING, 1, -1, 0,
                ChannelAccessValueFactory.createString(
                        Collections.singleton(""), UTF8));
        assertEquals(message, reserializeMessageToTCPServer(message, 40));
    }

    /**
     * Encode and decode a CA_PROTO_CLIENT_NAME message sent by a TCP client.
     */
    @Test
    public void tcpClientUserName() {
        ChannelAccessUserNameMessage message = new ChannelAccessUserNameMessage(
                "user42");
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_HOST_NAME message sent by a TCP client.
     */
    @Test
    public void tcpClientHostName() {
        ChannelAccessHostNameMessage message = new ChannelAccessHostNameMessage(
                "www.example.com");
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_ECHO message sent by a TCP client.
     */
    @Test
    public void tcpClientEcho() {
        ChannelAccessEchoMessage message = new ChannelAccessEchoMessage();
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_ECHO message sent by a TCP client.
     */
    @Test
    public void tcpClientSearch() {
        ChannelAccessSearchClientMessage message = new ChannelAccessSearchClientMessage(
                555, version, false, "myChannel");
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_VERSION message sent by a TCP server.
     */
    @Test
    public void tcpServerVersion() {
        ChannelAccessVersionTCPServerMessage message = new ChannelAccessVersionTCPServerMessage(
                version);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENT_ADD message sent by a TCP server.
     */
    @Test
    public void tcpServerSubscription() {
        ChannelAccessSubscriptionServerMessage message = new ChannelAccessSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 123,
                ChannelAccessValueFactory.createString(
                        Collections.singletonList("abc"), UTF8));
        assertEquals(message, reserializeMessageToTCPClient(message));
        message = new ChannelAccessSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1,
                ChannelAccessStatus.StatusMessage.ECA_ALLOCMEM.toStatus(), 123,
                null);
        assertEquals(message, reserializeMessageToTCPClient(message));
        // Test that a “short” string works as expected.
        message = new ChannelAccessSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 456,
                ChannelAccessValueFactory.createString(
                        Collections.singleton("Abc123"), UTF8));
        assertEquals(message, reserializeMessageToTCPClient(message, 32));
        // In this specific case, cutting the payload down to zero bytes
        // results in a different message because a CA_PROTO_EVENT_ADD message
        // with an empty payload acts as a cancellation confirmation for a
        // subscription. The exact behavior depends on how the codec is
        // configured.
        message = new ChannelAccessSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 456,
                ChannelAccessValueFactory.createString(
                        Collections.singleton(""), UTF8));
        ChannelAccessCancelSubscriptionServerMessage cancelMessage = new ChannelAccessCancelSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1, 1, 456);
        assertEquals(cancelMessage, reserializeMessageToTCPClient(message, 40));
        message = new ChannelAccessSubscriptionServerMessage(
                ChannelAccessValueType.DBR_DOUBLE, 0,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 123,
                ChannelAccessValueFactory.createDouble(
                        ArrayUtils.EMPTY_DOUBLE_ARRAY));
        cancelMessage = new ChannelAccessCancelSubscriptionServerMessage(
                ChannelAccessValueType.DBR_DOUBLE, 0, 1, 123);
        assertEquals(cancelMessage, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENT_ADD message sent by a TCP server.
     */
    @Test
    public void tcpServerCancelSubscription() {
        ChannelAccessCancelSubscriptionServerMessage message = new ChannelAccessCancelSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1, 453, 123);
        assertEquals(message, reserializeMessageToTCPClient(message));
        message = new ChannelAccessCancelSubscriptionServerMessage(
                ChannelAccessValueType.DBR_DOUBLE, 0, 453, 123);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ message sent by a TCP server.
     */
    @Test
    public void tcpServerRead() {
        ChannelAccessReadServerMessage message = new ChannelAccessReadServerMessage(
                ChannelAccessValueType.DBR_STS_FLOAT, 1, 0, 99,
                ChannelAccessValueFactory.createAlarmFloat(new float[] { 3.3f },
                        ChannelAccessAlarmSeverity.NO_ALARM,
                        ChannelAccessAlarmStatus.NO_ALARM));
        assertEquals(message, reserializeMessageToTCPClient(message));
        message = new ChannelAccessReadServerMessage(
                ChannelAccessValueType.DBR_STS_CHAR, 0, 0, 99,
                ChannelAccessValueFactory.createAlarmChar(
                        ArrayUtils.EMPTY_BYTE_ARRAY,
                        ChannelAccessAlarmSeverity.NO_ALARM,
                        ChannelAccessAlarmStatus.NO_ALARM));
        assertEquals(message, reserializeMessageToTCPClient(message));
        // Test that a “short” string works as expected.
        message = new ChannelAccessReadServerMessage(
                ChannelAccessValueType.DBR_STRING, 1, 456, 123,
                ChannelAccessValueFactory.createString(
                        Collections.singleton("Abc123"), UTF8));
        assertEquals(message, reserializeMessageToTCPClient(message, 8));
        message = new ChannelAccessReadServerMessage(
                ChannelAccessValueType.DBR_STRING, 1, 456, 123,
                ChannelAccessValueFactory.createString(
                        Collections.singleton(""), UTF8));
        assertEquals(message, reserializeMessageToTCPClient(message, 40));
    }

    /**
     * Encode and decode a CA_PROTO_READ_SYNC message sent by a TCP server.
     */
    @Test
    public void tcpServerReadSync() {
        ChannelAccessReadSyncMessage clientMessage = new ChannelAccessReadSyncMessage();
        ChannelAccessReadSyncMessage message = new ChannelAccessReadSyncMessage(
                clientMessage);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ERROR message sent by a TCP server.
     */
    @Test
    public void tcpServerError() {
        // We have to use a message without a payload here, because otherwise
        // the message embedded in the error message will lose its payload and
        // thus the error message sent will not be equal to the error message
        // received.
        ChannelAccessReadNotifyClientMessage clientMessage = new ChannelAccessReadNotifyClientMessage(
                ChannelAccessValueType.DBR_CTRL_DOUBLE, 5, 123, 999);
        // We have to reserialize the client message so that the original
        // headers are available.
        clientMessage = (ChannelAccessReadNotifyClientMessage) reserializeMessageToTCPServer(
                clientMessage);
        ChannelAccessErrorMessage message = new ChannelAccessErrorMessage(99,
                ChannelAccessStatus.StatusMessage.ECA_ALLOCMEM.toStatus(),
                clientMessage, "Cannot allocate memory.");
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ERROR message sent by a TCP server in
     * response to a CA_PROTO_WRITE_NOTIFY message. This test is similar to
     * {@link #tcpServerError()}, but specifically tests for a bug where
     * deserializing the headers of a CA_PROTO_WRITE_NOTIFY message attached to
     * an error message would fail.
     */
    @Test
    public void tcpServerErrorForWriteNotify() {
        ChannelAccessWriteNotifyClientMessage clientMessage = new ChannelAccessWriteNotifyClientMessage(
                ChannelAccessValueType.DBR_DOUBLE, 1, 41231, -7232,
                ChannelAccessValueFactory.createDouble(new double[] { 42.42 }));
        // We have to reserialize the client message so that the original
        // headers are available.
        clientMessage = (ChannelAccessWriteNotifyClientMessage) reserializeMessageToTCPServer(
                clientMessage);
        ChannelAccessErrorMessage message = new ChannelAccessErrorMessage(99,
                ChannelAccessStatus.StatusMessage.ECA_ALLOCMEM.toStatus(),
                clientMessage, "Cannot allocate memory.");
        // We cannot simply compare the two error messages because the original
        // message loses its payload and thus the whole message is different.
        ChannelAccessErrorMessage reserializedMessage = (ChannelAccessErrorMessage) reserializeMessageToTCPClient(
                message);
        assertEquals(message.getChannelCID(),
                reserializedMessage.getChannelCID());
        assertEquals(message.getCommand(), reserializedMessage.getCommand());
        assertEquals(message.getMessageString(),
                reserializedMessage.getMessageString());
        assertEquals(message.getStatus(), reserializedMessage.getStatus());
        ChannelAccessWriteNotifyClientMessage clientMessageHeadersOnly = (ChannelAccessWriteNotifyClientMessage) reserializedMessage
                .getOriginalMessage();
        assertEquals(clientMessage.getChannelSID(),
                clientMessageHeadersOnly.getChannelSID());
        assertEquals(clientMessage.getCommand(),
                clientMessageHeadersOnly.getCommand());
        assertEquals(clientMessage.getCount(),
                clientMessageHeadersOnly.getCount());
        assertEquals(clientMessage.getDataType(),
                clientMessageHeadersOnly.getDataType());
        assertEquals(clientMessage.getOperationId(),
                clientMessageHeadersOnly.getOperationId());
        assertNull(clientMessageHeadersOnly.getValue());
    }

    /**
     * Encode and decode a CA_PROTO_CLEAR_CHANNEL message sent by a TCP server.
     */
    @Test
    public void tcpServerDisconnectChannel() {
        ChannelAccessDisconnectChannelMessage message = new ChannelAccessDisconnectChannelMessage(
                99, 42);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ_NOTIFY message sent by a TCP server.
     */
    @Test
    public void tcpServerReadNotify() {
        ChannelAccessReadNotifyServerMessage message = new ChannelAccessReadNotifyServerMessage(
                ChannelAccessValueType.DBR_STS_LONG, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 99,
                ChannelAccessValueFactory.createAlarmLong(new int[] { 33 },
                        ChannelAccessAlarmSeverity.NO_ALARM,
                        ChannelAccessAlarmStatus.NO_ALARM));
        assertEquals(message, reserializeMessageToTCPClient(message));
        message = new ChannelAccessReadNotifyServerMessage(
                ChannelAccessValueType.DBR_FLOAT, 0,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 99,
                ChannelAccessValueFactory.createFloat(
                        ArrayUtils.EMPTY_FLOAT_ARRAY));
        assertEquals(message, reserializeMessageToTCPClient(message));
        // Test that a “short” string works as expected.
        message = new ChannelAccessReadNotifyServerMessage(
                ChannelAccessValueType.DBR_STS_STRING, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 123,
                ChannelAccessValueFactory.createAlarmString(
                        Collections.singleton("Abc123"), UTF8,
                        ChannelAccessAlarmSeverity.NO_ALARM,
                        ChannelAccessAlarmStatus.NO_ALARM));
        assertEquals(message, reserializeMessageToTCPClient(message, 16));
        message = new ChannelAccessReadNotifyServerMessage(
                ChannelAccessValueType.DBR_TIME_STRING, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 123,
                ChannelAccessValueFactory.createTimeString(
                        Collections.singleton(""), UTF8,
                        ChannelAccessAlarmSeverity.NO_ALARM,
                        ChannelAccessAlarmStatus.NO_ALARM, 123, 456));
        assertEquals(message, reserializeMessageToTCPClient(message, 40));
    }

    /**
     * Encode and decode a CA_PROTO_READ_NOTIFY message sent by a TCP server.
     * This test is similar to {@link #tcpServerReadNotify()}, but tests a large
     * message (with a count and payload greater than 65535).
     */
    @Test
    public void tcpServerReadNotifyLongMessage() {
        int[] numbers = new int[100000];
        for (int i = 0; i < numbers.length; ++i) {
            numbers[i] = i;
        }
        ChannelAccessReadNotifyServerMessage message = new ChannelAccessReadNotifyServerMessage(
                ChannelAccessValueType.DBR_STS_LONG, numbers.length,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 99,
                ChannelAccessValueFactory.createAlarmLong(numbers,
                        ChannelAccessAlarmSeverity.NO_ALARM,
                        ChannelAccessAlarmStatus.NO_ALARM));
        // The code that we use for most tests limits the payload size to 16384
        // and the byte sinks and sources that we use have a small buffer
        // (8192 bytes). For these reasons, we have to create a seprate codec
        // and use large buffers.
        ChannelAccessMessageCodec codec = new ChannelAccessMessageCodec();
        codec.setMaxPayloadReceiveSize(4194304);
        codec.setMaxPayloadSendSize(4194304);
        ByteBufferByteSink byteSink = new ByteBufferByteSink();
        byteSink.addBuffer(ByteBuffer.allocate(4194304));
        codec.encodeMessageToTCPClient(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec
                .decodeMessageToTCPClient(byteSource, version, LOCALHOST, 1234);
        assertEquals(0, byteSource.remaining());
        assertEquals(message, newMessage);
    }

    /**
     * Encode and decode a CA_PROTO_CREATE_CHAN message sent by a TCP server.
     */
    @Test
    public void tcpServerConnectChannel() {
        ChannelAccessConnectChannelServerMessage message = new ChannelAccessConnectChannelServerMessage(
                55, 23, ChannelAccessValueType.DBR_ENUM, 20);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_WRITE_NOTIFY message sent by a TCP server.
     */
    @Test
    public void tcpServerWriteNotify() {
        ChannelAccessWriteNotifyServerMessage message = new ChannelAccessWriteNotifyServerMessage(
                ChannelAccessValueType.DBR_SHORT, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 22);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ACCESS_RIGHTS message sent by a TCP server.
     */
    @Test
    public void tcpServerAccessRights() {
        ChannelAccessAccessRightsMessage message = new ChannelAccessAccessRightsMessage(
                321, true, false);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ECHO message sent by a TCP server.
     */
    @Test
    public void tcpServerEcho() {
        ChannelAccessEchoMessage message = new ChannelAccessEchoMessage();
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_CREATE_CH_FAIL message sent by a TCP server.
     */
    @Test
    public void tcpServerConnectChannelFailed() {
        ChannelAccessConnectChannelFailedMessage message = new ChannelAccessConnectChannelFailedMessage(
                234);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_SERVER_DISCONN message sent by a TCP server.
     */
    @Test
    public void tcpServerChannelDisconnectedByServer() {
        ChannelAccessChannelDisconnectedByServerMessage message = new ChannelAccessChannelDisconnectedByServerMessage(
                -789);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_SEARCH message sent by a TCP server.
     */
    @Test
    public void tcpServerSearchTCPServer() {
        ChannelAccessSearchTCPServerMessage message = new ChannelAccessSearchTCPServerMessage(
                987, LOCALHOST, 1234);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_NOT_FOUND message sent by a TCP server.
     */
    @Test
    public void tcpServerNotFound() {
        ChannelAccessSearchClientMessage clientMessage = new ChannelAccessSearchClientMessage(
                999, version, true, "myChannel");
        ChannelAccessNotFoundMessage message = new ChannelAccessNotFoundMessage(
                clientMessage);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_VERSION message sent by a UDP client.
     */
    @Test
    public void udpClientVersion() {
        ChannelAccessVersionUDPMessage message = new ChannelAccessVersionUDPMessage(
                version, true, 25);
        assertEquals(message, reserializeMessageToUDPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_SEARCH message sent by a UDP client.
     */
    @Test
    public void udpClientSearch() {
        ChannelAccessSearchClientMessage message = new ChannelAccessSearchClientMessage(
                456, version, false, "someChannel");
        assertEquals(message, reserializeMessageToUDPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_VERSION message sent by a UDP server.
     */
    @Test
    public void udpServerVersion() {
        ChannelAccessVersionUDPMessage message = new ChannelAccessVersionUDPMessage(
                version, false, 37);
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_SEARCH message sent by a UDP server.
     */
    @Test
    public void udpServerSearch() {
        ChannelAccessSearchUDPServerMessage message = new ChannelAccessSearchUDPServerMessage(
                version, 321, LOCALHOST, 1234);
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ERROR message sent by a UDP server.
     */
    @Test
    public void udpServerError() {
        ChannelAccessErrorMessage message;
        message = new ChannelAccessErrorMessage(4321,
                ChannelAccessStatus.StatusMessage.ECA_BADSTR.toStatus(), null,
                null);
        assertEquals(message, reserializeMessageToUDPClient(message));
        // We also test an error message that is sent in response to a version
        // message. This scenario is very unlikely to ever appear in practice,
        // but it is designed to specifically test for a bug, where an original
        // message attached to an error message sent over UDP would be treated
        // like it had been received over TCP.
        ChannelAccessVersionUDPMessage clientMessage = new ChannelAccessVersionUDPMessage(
                version, true, 87);
        // We have to reserialize the client message so that the original
        // headers are available.
        clientMessage = (ChannelAccessVersionUDPMessage) reserializeMessageToUDPServer(
                clientMessage);
        message = new ChannelAccessErrorMessage(
                ChannelAccessErrorMessage.UNKNOWN_CHANNEL_CID,
                ChannelAccessStatus.StatusMessage.ECA_ALLOCMEM.toStatus(),
                clientMessage, "Cannot allocate memory.");
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_NOT_FOUND message sent by a UDP server.
     */
    @Test
    public void udpServerNotFound() {
        ChannelAccessSearchClientMessage clientMessage = new ChannelAccessSearchClientMessage(
                4321, version, true, "abcdef");
        ChannelAccessNotFoundMessage message = new ChannelAccessNotFoundMessage(
                clientMessage);
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ECHO message sent by a UDP server.
     */
    @Test
    public void udpServerEcho() {
        ChannelAccessEchoMessage message = new ChannelAccessEchoMessage();
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

}
