/*
 * Copyright 2014-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import static org.junit.Assert.*;

import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;

import org.apache.commons.lang3.tuple.Pair;
import org.junit.Test;

/**
 * Tests for the {@link NullTerminatedStringUtil}.
 * 
 * @author Sebastian Marsching
 */
public class NullTerminatedStringUtilTest {

    private final static Charset UTF8 = StandardCharsets.UTF_8;

    @Test
    public void nullTerminatedBytesToString() {
        String testString = "Hello World with non-ASCII characters: ®€äöü!";
        byte[] testBytes = testString.getBytes(UTF8);
        // Add some extra null-bytes...
        testBytes = Arrays.copyOf(testBytes, testBytes.length + 5);
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(testBytes, UTF8);
        assertEquals(testString, resultString);
        assertEquals(testString.length(), resultString.length());
    }

    @Test
    public void nullTerminatedBytesToStringStrict() {
        String testString = "Hello World with non-ASCII characters: ®€äöü!";
        byte[] testBytes = testString.getBytes(UTF8);
        // Add some extra null-bytes...
        testBytes = Arrays.copyOf(testBytes, testBytes.length + 5);
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(testBytes, UTF8, true);
        assertEquals(testString, resultString);
        assertEquals(testString.length(), resultString.length());
    }

    @Test
    public void nullTerminatedBytesToStringWithoutNullByte() {
        String testString = "Hello World with non-ASCII characters: ®€äöü!";
        byte[] testBytes = testString.getBytes(UTF8);
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(testBytes, UTF8);
        assertEquals(testString, resultString);
        assertEquals(testString.length(), resultString.length());
    }

    @Test(expected = IllegalArgumentException.class)
    public void nullTerminatedBytesToStringWithoutNullByteStrict() {
        String testString = "Hello World with non-ASCII characters: ®€äöü!";
        byte[] testBytes = testString.getBytes(UTF8);
        NullTerminatedStringUtil.nullTerminatedBytesToString(testBytes, UTF8,
                true);
    }

    @Test
    public void nullTerminatedBytesToStringPartial() {
        byte[] testBytes = new byte[] { 97, 97, 97, 98, 98, 98, 0 };
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(testBytes, 2, 5, UTF8, false);
        assertEquals("abbb", resultString);
        assertEquals(4, resultString.length());
    }

    @Test
    public void nullTerminatedBytesToStringPartialWithoutNullByte() {
        byte[] testBytes = new byte[] { 97, 97, 97, 98, 98, 98, 0 };
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(testBytes, 2, 3, UTF8, false);
        assertEquals("abb", resultString);
        assertEquals(3, resultString.length());
    }

    @Test
    public void nullTerminatedBytesToStringPartialStrict() {
        byte[] testBytes = new byte[] { 97, 97, 97, 98, 98, 98, 0 };
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(testBytes, 2, 5, UTF8, true);
        assertEquals("abbb", resultString);
        assertEquals(4, resultString.length());
    }

    @Test(expected = IllegalArgumentException.class)
    public void nullTerminatedBytesToStringPartialWithoutNullByteStrict() {
        byte[] testBytes = new byte[] { 97, 97, 97, 98, 98, 98, 0 };
        NullTerminatedStringUtil.nullTerminatedBytesToString(testBytes, 2, 3,
                UTF8, true);
    }

    @Test
    public void truncatedNonAlignedLongString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes(
                testString, 14, UTF8);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // Sanity check: The returned array should not be longer than the
        // specified maximum.
        assertTrue(bytes.length <= 14);
        // The returned length should be 13 (12 bytes for the first 11
        // characters plus one null-byte).
        assertEquals(13, length);
        // All characters after length - 1 should be null-bytes.
        for (int i = length - 1; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
    }

    @Test
    public void truncatedNonAlignedShortString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes(
                testString, 1024, UTF8);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // Sanity check: The returned array should not be longer than the
        // specified maximum.
        assertTrue(bytes.length <= 1024);
        // The returned length should be 17 (16 bytes for the 13 characters plus
        // one null-byte).
        assertEquals(17, length);
        // All characters after length - 1 should be null-bytes.
        for (int i = length - 1; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
        // The string should completely fit into the byte array.
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(bytes, UTF8);
        assertEquals(testString, resultString);
        assertEquals(testString.length(), resultString.length());
    }

    @Test
    public void truncatedNonAlignedExactlyMatchingString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes(
                testString, 17, UTF8);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // Sanity check: The returned array should not be longer than the
        // specified maximum.
        assertTrue(bytes.length <= 17);
        // The returned length should be 17 (16 bytes for the 13 characters plus
        // one null-byte).
        assertEquals(17, length);
        // All characters after length - 1 should be null-bytes.
        for (int i = length - 1; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
        // The string should completely fit into the byte array.
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(bytes, UTF8);
        assertEquals(testString, resultString);
        assertEquals(testString.length(), resultString.length());
    }

    @Test
    public void truncatedNonAlignedEmptyString() {
        String testString = "";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes(
                testString, Integer.MAX_VALUE, UTF8);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // The returned length should be 1 (0 bytes for the characters but one
        // null byte).
        assertEquals(1, length);
        // All characters after length - 1 should be null-bytes.
        for (int i = length - 1; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
    }

    @Test(expected = IllegalArgumentException.class)
    public void stringToTruncatedNullTerminatedBytesNegativeLength1() {
        NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes("foo",
                -5, UTF8);
    }

    @Test(expected = IllegalArgumentException.class)
    public void stringToTruncatedNullTerminatedBytesNegativeLength2() {
        NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes("foo",
                -5, 10, UTF8);
    }

    @Test(expected = IllegalArgumentException.class)
    public void stringToTruncatedNullTerminatedBytesNegativeAlignment() {
        NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes("foo",
                10, -5, UTF8);
    }

    @Test
    public void nonAlignedLongString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(testString, 14,
                        UTF8);
        // The string is too long, thus the result should be null
        assertNull(result);
    }

    @Test
    public void nonAlignedShortString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(testString, 1024,
                        UTF8);
        assertNotNull(result);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // Sanity check: The returned array should not be longer than the
        // specified maximum.
        assertTrue(bytes.length <= 1024);
        // The returned length should be 17 (16 bytes for the 13 characters plus
        // one null-byte).
        assertEquals(17, length);
        // All characters after length - 1 should be null-bytes.
        for (int i = length - 1; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
        // The string should completely fit into the byte array.
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(bytes, UTF8);
        assertEquals(testString, resultString);
        assertEquals(testString.length(), resultString.length());
    }

    @Test
    public void nonAlignedExactlyMatchingString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(testString, 17,
                        UTF8);
        assertNotNull(result);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // Sanity check: The returned array should not be longer than the
        // specified maximum.
        assertTrue(bytes.length <= 17);
        // The returned length should be 17 (16 bytes for the 13 characters plus
        // one null-byte).
        assertEquals(17, length);
        // All characters after length - 1 should be null-bytes.
        for (int i = length - 1; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
        // The string should completely fit into the byte array.
        String resultString = NullTerminatedStringUtil
                .nullTerminatedBytesToString(bytes, UTF8);
        assertEquals(testString, resultString);
        assertEquals(testString.length(), resultString.length());
    }

    @Test(expected = IllegalArgumentException.class)
    public void stringToSizeLimitedNullTerminatedBytesOrNullNegativeLength1() {
        NullTerminatedStringUtil.stringToSizeLimitedNullTerminatedBytesOrNull(
                "foo", -5, UTF8);
    }

    @Test(expected = IllegalArgumentException.class)
    public void stringToSizeLimitedNullTerminatedBytesOrNullNegativeLength2() {
        NullTerminatedStringUtil.stringToSizeLimitedNullTerminatedBytesOrNull(
                "foo", -5, 10, UTF8);
    }

    @Test(expected = IllegalArgumentException.class)
    public void stringToSizeLimitedNullTerminatedBytesOrNullNegativeAlignment() {
        NullTerminatedStringUtil.stringToSizeLimitedNullTerminatedBytesOrNull(
                "foo", 10, -5, UTF8);
    }

    @Test
    public void truncatedAlignedLongString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes(
                testString, 14, 8, UTF8);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // Sanity check: The returned array should not be longer than the
        // specified maximum.
        assertTrue(bytes.length <= 14);
        // The returned length should be 8 (7 character bytes and one
        // null-byte).
        assertEquals(8, length);
        // All characters after length - 1 should be null-bytes.
        for (int i = length - 1; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
    }

    @Test
    public void truncatedAlignedShortString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes(
                testString, 1024, 8, UTF8);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // Sanity check: The returned array should not be longer than the
        // specified maximum.
        assertTrue(bytes.length <= 1024);
        // The returned length should be 24 (16 character bytes and 8
        // null-bytes).
        assertEquals(24, length);
        // All characters after position 16 should be null-bytes.
        for (int i = 16; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
    }

    @Test
    public void truncatedAlignedEmptyString() {
        String testString = "";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil.stringToTruncatedNullTerminatedBytes(
                testString, Integer.MAX_VALUE, 8, UTF8);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // The returned length should be 8 (0 bytes for the characters but 8
        // null bytes).
        assertEquals(8, length);
        // All characters should be null-bytes.
        for (int i = 0; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
    }

    @Test
    public void alignedShortString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(testString, 1024,
                        8, UTF8);
        byte[] bytes = result.getLeft();
        int length = result.getRight();
        // Sanity check: The returned array must have space for all the bytes.
        assertTrue(bytes.length >= length);
        // Sanity check: The returned array should not be longer than the
        // specified maximum.
        assertTrue(bytes.length <= 1024);
        // The returned length should be 24 (16 character bytes and 8
        // null-bytes).
        assertEquals(24, length);
        // All characters after position 16 should be null-bytes.
        for (int i = 16; i < bytes.length; i++) {
            assertEquals(0, bytes[i]);
        }
    }

    @Test
    public void alignedLongString() {
        String testString = "1234567890äöü";
        Pair<byte[], Integer> result;
        result = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(testString, 16,
                        8, UTF8);
        // The string would fit into the array and even be aligned, however
        // there is no space for the terminating null-byte.
        assertNull(result);
    }

    @Test
    public void fixSizedString() {
        String testString = "abc";
        byte[] result = NullTerminatedStringUtil
                .stringToTruncatedNullTerminatedFixedBytes(testString, 12, UTF8);
        assertEquals(12, result.length);
    }

}
