/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import static org.junit.Assert.*;

import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

import org.apache.commons.lang3.ArrayUtils;
import org.junit.Test;

import com.aquenos.epics.jackie.common.io.ByteBufferByteSink;
import com.aquenos.epics.jackie.common.io.ByteBufferByteSource;

/**
 * Tests for the {@link ChannelAccessValueCodec}.
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessValueCodecTest {

    private static final Charset UTF8 = StandardCharsets.UTF_8;

    private static ByteBufferByteSink createByteSink() {
        ByteBufferByteSink byteSink = new ByteBufferByteSink();
        byteSink.addBuffer(ByteBuffer.allocate(8192));
        return byteSink;
    }

    private static ByteBufferByteSource createByteSource(
            ByteBufferByteSink byteSink) {
        return createByteSource(byteSink, 0);
    }

    private static ByteBufferByteSource createByteSource(
            ByteBufferByteSink byteSink, int cutBytes) {
        ByteBufferByteSource byteSource = new ByteBufferByteSource();
        if (cutBytes > 0) {
            long size = byteSink.getWrittenDataSizeInBytes();
            // The test code should never request to cut more bytes than were
            // written.
            assert size >= cutBytes;
            long remainingSize = size - cutBytes;
            ByteBuffer[] originalBuffers = byteSink.getWrittenData();
            ArrayList<ByteBuffer> buffers = new ArrayList<>(
                    originalBuffers.length);
            for (ByteBuffer buffer : originalBuffers) {
                if (buffer.remaining() <= remainingSize) {
                    buffers.add(buffer);
                    remainingSize -= buffer.remaining();
                } else if (remainingSize == 0) {
                    break;
                } else {
                    buffer.limit(
                            buffer.position() + (int) remainingSize);
                    buffers.add(buffer);
                    break;
                }
            }
            byteSource.appendBuffers(buffers);
        } else {
            byteSource.appendBuffers(byteSink.getWrittenData());
        }
        return byteSource;
    }

    private static ChannelAccessValue<?> reserializeValue(
            ChannelAccessValue<?> value) {
        return reserializeAndCutValue(value, 0);
    }

    private static ChannelAccessValue<?> reserializeAndCutValue(
            ChannelAccessValue<?> value, int cutByBytes) {
        ChannelAccessValueType type = value.getType();
        int count = value.getValueSize();
        int dataSize = ChannelAccessValueCodec.calculatePayloadSize(type,
                count);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodeValue(byteSink, value, count);
        ByteBufferByteSource byteSource = createByteSource(byteSink,
                cutByBytes);
        assertEquals(dataSize - cutByBytes, byteSource.remaining());
        ChannelAccessValue<?> newValue = ChannelAccessValueCodec
                .decodeValue(byteSource, type, dataSize - cutByBytes, count,
                        UTF8);
        assertEquals(0, byteSource.remaining());
        return newValue;
    }

    /**
     * Test deserialization of “short” string values.
     */
    @Test
    public void shortStrings() {
        ChannelAccessString value;
        value = ChannelAccessValueFactory.createString(
                Arrays.asList(new String[] { "Hello World!" }), UTF8);
        assertEquals(value, reserializeAndCutValue(value, 24));
        value = ChannelAccessValueFactory.createString(
                Arrays.asList(new String[] { "Abc123" }), UTF8);
        assertEquals(value, reserializeAndCutValue(value, 32));
        value = ChannelAccessValueFactory.createString(
                Arrays.asList(new String[] { "" }), UTF8);
        assertEquals(value, reserializeAndCutValue(value, 40));
        value = ChannelAccessValueFactory.createAlarmString(
                Arrays.asList(new String[] { "Hello World!" }),
                UTF8,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM);
        assertEquals(value, reserializeAndCutValue(value, 16));
        value = ChannelAccessValueFactory.createTimeString(
                Arrays.asList(new String[] { "Hello World!" }),
                UTF8,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM,
                123,
                456);
        assertEquals(value, reserializeAndCutValue(value, 8));
    }

    /**
     * Serialize and deserialize a DBR_STRING value.
     */
    @Test
    public void simpleString() {
        ChannelAccessString value;
        value = ChannelAccessValueFactory.createString(Collections.emptyList(),
                UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createString(
                Arrays.asList(new String[] { "Hello World!" }), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createString(Arrays.asList(new String[] {}), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createString(
                        Arrays.asList(new String[] { "Hello World!", "",
                                "123456789012345678901234567890123456789" }),
                        UTF8);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_SHORT value.
     */
    @Test
    public void simpleShort() {
        ChannelAccessShort value;
        value = ChannelAccessValueFactory.createShort(
                ArrayUtils.EMPTY_SHORT_ARRAY);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createShort(new short[] { 5 });
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createShort(new short[] {});
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createShort(new short[] { 18, 9353, -2352, 124, 0 });
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_FLOAT value.
     */
    @Test
    public void simpleFloat() {
        ChannelAccessFloat value;
        value = ChannelAccessValueFactory.createFloat(
                ArrayUtils.EMPTY_FLOAT_ARRAY);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createFloat(new float[] { 5.0f });
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createFloat(new float[] {});
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createFloat(new float[] { 5.0f });
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_ENUM value.
     */
    @Test
    public void simpleEnum() {
        ChannelAccessEnum value;
        value = ChannelAccessValueFactory.createEnum(
                ArrayUtils.EMPTY_SHORT_ARRAY);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createEnum(new short[] { 5 });
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createEnum(new short[] {});
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createEnum(new short[] { 17, -88, 2 });
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CHAR value.
     */
    @Test
    public void simpleChar() {
        ChannelAccessChar value;
        value = ChannelAccessValueFactory.createChar(
                ArrayUtils.EMPTY_BYTE_ARRAY);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createChar(new byte[] { 5 });
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createChar(new byte[] {});
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createChar(new byte[] { 5, 37, -18 });
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_LONG value.
     */
    @Test
    public void simpleLong() {
        ChannelAccessLong value;
        value = ChannelAccessValueFactory.createLong(
                ArrayUtils.EMPTY_INT_ARRAY);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createLong(new int[] { 5 });
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createLong(new int[] {});
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createLong(new int[] { 18, 9353, -2352, 124, 0 });
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_DOUBLE value.
     */
    @Test
    public void simpleDouble() {
        ChannelAccessDouble value;
        value = ChannelAccessValueFactory.createDouble(
                ArrayUtils.EMPTY_DOUBLE_ARRAY);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createDouble(new double[] { 5.0 });
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createDouble(new double[] {});
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createDouble(new double[] { 5.0, -3.0, 999.8 });
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_STS_STRING value.
     */
    @Test
    public void alarmString() {
        ChannelAccessAlarmString value;
        value = ChannelAccessValueFactory.createAlarmString(
                Collections.emptyList(), UTF8,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmString(
                Arrays.asList(new String[] { "Hello World!" }), UTF8,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmString(
                Arrays.asList(new String[] {}), UTF8,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmString(
                Arrays.asList(new String[] { "Hello World!", "",
                        "123456789012345678901234567890123456789" }),
                UTF8, ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_STS_SHORT value.
     */
    @Test
    public void alarmShort() {
        ChannelAccessAlarmShort value;
        value = ChannelAccessValueFactory.createAlarmShort(
                ArrayUtils.EMPTY_SHORT_ARRAY,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmShort(new short[] { 5 },
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmShort(new short[] {},
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmShort(
                new short[] { 18, 9353, -2352, 124, 0 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_STS_FLOAT value.
     */
    @Test
    public void alarmFloat() {
        ChannelAccessAlarmFloat value;
        value = ChannelAccessValueFactory.createAlarmFloat(
                ArrayUtils.EMPTY_FLOAT_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmFloat(new float[] { 5.0f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmFloat(new float[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmFloat(
                new float[] { 5.0f, -3.0f, 999.8f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_STS_ENUM value.
     */
    @Test
    public void alarmEnum() {
        ChannelAccessAlarmEnum value;
        value = ChannelAccessValueFactory.createAlarmEnum(
                ArrayUtils.EMPTY_SHORT_ARRAY,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmEnum(new short[] { 5 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmEnum(new short[] {},
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmEnum(
                new short[] { 17, -88, 2 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_STS_CHAR value.
     */
    @Test
    public void alarmChar() {
        ChannelAccessAlarmChar value;
        value = ChannelAccessValueFactory.createAlarmChar(
                ArrayUtils.EMPTY_BYTE_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmChar(new byte[] { 5 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmChar(new byte[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmChar(
                new byte[] { 5, 37, -18 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_STS_LONG value.
     */
    @Test
    public void alarmLong() {
        ChannelAccessAlarmLong value;
        value = ChannelAccessValueFactory.createAlarmLong(
                ArrayUtils.EMPTY_INT_ARRAY,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmLong(new int[] { 5 },
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmLong(new int[] {},
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmLong(
                new int[] { 18, 9353, -2352, 124, 0 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_STS_DOUBLE value.
     */
    @Test
    public void alarmDouble() {
        ChannelAccessAlarmDouble value;
        value = ChannelAccessValueFactory.createAlarmDouble(
                ArrayUtils.EMPTY_DOUBLE_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmDouble(
                new double[] { 5.0 }, ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmDouble(new double[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmDouble(
                new double[] { 5.0, -3.0, 999.8 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_TIME_STRING value.
     */
    @Test
    public void timeString() {
        ChannelAccessTimeString value;
        value = ChannelAccessValueFactory.createTimeString(
                Collections.emptyList(), UTF8,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT, 500000000, 1237);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeString(
                Arrays.asList(new String[] { "Hello World!" }), UTF8,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT, 500000000, 1237);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeString(
                Arrays.asList(new String[] {}), UTF8,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT, -50000, 1231);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeString(
                Arrays.asList(new String[] { "Hello World!", "",
                        "123456789012345678901234567890123456789" }),
                UTF8, ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, 9171241, 0);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_TIME_SHORT value.
     */
    @Test
    public void timeShort() {
        ChannelAccessTimeShort value;
        value = ChannelAccessValueFactory.createTimeShort(
                ArrayUtils.EMPTY_SHORT_ARRAY,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, 7123131, 15);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeShort(new short[] { 5 },
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, 7123131, 15);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeShort(new short[] {},
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, 124121, 7151);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeShort(
                new short[] { 18, 9353, -2352, 124, 0 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, -34, 72372342);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_TIME_FLOAT value.
     */
    @Test
    public void timeFloat() {
        ChannelAccessTimeFloat value;
        value = ChannelAccessValueFactory.createTimeFloat(
                ArrayUtils.EMPTY_FLOAT_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 72354251, 142);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeFloat(new float[] { 5.0f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 72354251, 142);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeFloat(new float[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 535251, 86);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeFloat(
                new float[] { 5.0f, -3.0f, 999.8f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, -67342, 87266);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_TIME_ENUM value.
     */
    @Test
    public void timeEnum() {
        ChannelAccessTimeEnum value;
        value = ChannelAccessValueFactory.createTimeEnum(
                ArrayUtils.EMPTY_SHORT_ARRAY,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT, 1235, 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeEnum(new short[] { 5 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT, 1235, 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeEnum(new short[] {},
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT, 0, 81);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeEnum(
                new short[] { 17, -88, 2 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT, 646, 27);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_TIME_CHAR value.
     */
    @Test
    public void timeChar() {
        ChannelAccessTimeChar value;
        value = ChannelAccessValueFactory.createTimeChar(
                ArrayUtils.EMPTY_BYTE_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, 7141, 3);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeChar(new byte[] { 5 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, 7141, 3);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeChar(new byte[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, -835712312, 3);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeChar(
                new byte[] { 5, 37, -18 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, 0, 0);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_TIME_LONG value.
     */
    @Test
    public void timeLong() {
        ChannelAccessTimeLong value;
        value = ChannelAccessValueFactory.createTimeLong(
                ArrayUtils.EMPTY_INT_ARRAY,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, 87, 51);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeLong(new int[] { 5 },
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, 87, 51);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeLong(new int[] {},
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, 17, 3);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeLong(
                new int[] { 18, 9353, -2352, 124, 0 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, -5, 88343);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_TIME_DOUBLE value.
     */
    @Test
    public void timeDouble() {
        ChannelAccessTimeDouble value;
        value = ChannelAccessValueFactory.createTimeDouble(
                ArrayUtils.EMPTY_DOUBLE_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, -17, 0);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeDouble(new double[] { 5.0 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, -17, 0);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeDouble(new double[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 83, 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createTimeDouble(
                new double[] { 5.0, -3.0, 999.8 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 1241, 1);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_GR_STRING value.
     */
    @Test
    public void graphicsString() {
        ChannelAccessAlarmString value;
        value = ChannelAccessValueFactory.createGraphicsString(
                Collections.emptyList(), UTF8,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsString(
                Arrays.asList(new String[] { "Hello World!" }), UTF8,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsString(
                Arrays.asList(new String[] {}), UTF8,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsString(
                Arrays.asList(new String[] { "Hello World!", "",
                        "123456789012345678901234567890123456789" }),
                UTF8, ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_GR_SHORT value.
     */
    @Test
    public void graphicsShort() {
        ChannelAccessGraphicsShort value;
        value = ChannelAccessValueFactory.createGraphicsShort(
                ArrayUtils.EMPTY_SHORT_ARRAY,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, (short) 0, (short) 150,
                (short) 10, (short) 120, (short) -20, (short) 190, "mA", UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsShort(new short[] { 5 },
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, (short) 0, (short) 150,
                (short) 10, (short) 120, (short) -20, (short) 190, "mA", UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsShort(new short[] {},
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, (short) 0, (short) 150,
                (short) 10, (short) 120, (short) -20, (short) 190, "mA", UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsShort(
                new short[] { 18, 9353, -2352, 124, 0 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, (short) 0, (short) 150,
                (short) 10, (short) 120, (short) -20, (short) 190, "mA", UTF8);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_GR_FLOAT value.
     */
    @Test
    public void graphicsFloat() {
        ChannelAccessGraphicsFloat value;
        value = ChannelAccessValueFactory.createGraphicsFloat(
                ArrayUtils.EMPTY_FLOAT_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0f, 150.0f, 10.0f, 120.0f,
                -20.0f, 190.0f, "mA", UTF8, (short) 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsFloat(
                new float[] { 5.0f }, ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0f, 150.0f, 10.0f, 120.0f,
                -20.0f, 190.0f, "mA", UTF8, (short) 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsFloat(new float[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0f, 150.0f, 10.0f, 120.0f,
                -20.0f, 190.0f, "mA", UTF8, (short) 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsFloat(
                new float[] { 5.0f, -3.0f, 999.8f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0f, 150.0f, 10.0f, 120.0f,
                -20.0f, 190.0f, "mA", UTF8, (short) 2);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_GR_ENUM value.
     */
    @Test
    public void graphicsEnum() {
        ChannelAccessGraphicsEnum value;
        value = ChannelAccessValueFactory.createGraphicsEnum(
                ArrayUtils.EMPTY_SHORT_ARRAY,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT,
                Arrays.asList(new String[] { "off" }), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsEnum(new short[] { 5 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT,
                Arrays.asList(new String[] { "off" }), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsEnum(new short[] {},
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT,
                Arrays.asList(new String[] {}), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsEnum(
                new short[] { 17, -88, 2 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT,
                Arrays.asList(new String[] { "off", "on", "unknown" }), UTF8);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_GR_CHAR value.
     */
    @Test
    public void graphicsChar() {
        ChannelAccessGraphicsChar value;
        value = ChannelAccessValueFactory.createGraphicsChar(
                ArrayUtils.EMPTY_BYTE_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, (byte) 0, (byte) (byte) 130,
                (byte) 10, (byte) 100, (byte) -20, (byte) 100, "dB", UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsChar(new byte[] { 5 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, (byte) 0, (byte) (byte) 130,
                (byte) 10, (byte) 100, (byte) -20, (byte) 100, "dB", UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsChar(new byte[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, (byte) 0, (byte) (byte) 130,
                (byte) 10, (byte) 100, (byte) -20, (byte) 100, "dB", UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsChar(
                new byte[] { 5, 37, -18 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, (byte) 0, (byte) (byte) 130,
                (byte) 10, (byte) 100, (byte) -20, (byte) 100, "dB", UTF8);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_GR_LONG value.
     */
    @Test
    public void graphicsLong() {
        ChannelAccessGraphicsLong value;
        value = ChannelAccessValueFactory.createGraphicsLong(
                ArrayUtils.EMPTY_INT_ARRAY,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, 0, 150, 10, 120, -20, 190, "mA",
                UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsLong(new int[] { 5 },
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, 0, 150, 10, 120, -20, 190, "mA",
                UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsLong(new int[] {},
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, 0, 150, 10, 120, -20, 190,
                "mA", UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsLong(
                new int[] { 18, 9353, -2352, 124, 0 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, 0, 150, 10, 120, -20, 190,
                "mA", UTF8);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_GR_DOUBLE value.
     */
    @Test
    public void graphicsDouble() {
        ChannelAccessGraphicsDouble value;
        value = ChannelAccessValueFactory.createGraphicsDouble(
                ArrayUtils.EMPTY_DOUBLE_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0, 150.0, 10.0, 120.0,
                -20.0, 190.0, "mA", UTF8, (short) 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsDouble(
                new double[] { 5.0 }, ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0, 150.0, 10.0, 120.0,
                -20.0, 190.0, "mA", UTF8, (short) 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsDouble(new double[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0, 150.0, 10.0, 120.0,
                -20.0, 190.0, "mA", UTF8, (short) 2);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createGraphicsDouble(
                new double[] { 5.0, -3.0, 999.8 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0, 150.0, 10.0, 120.0,
                -20.0, 190.0, "mA", UTF8, (short) 2);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CTRL_STRING value.
     */
    @Test
    public void controlsString() {
        ChannelAccessAlarmString value;
        value = ChannelAccessValueFactory.createControlsString(
                Collections.emptyList(), UTF8,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsString(
                Arrays.asList(new String[] { "Hello World!" }), UTF8,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsString(
                Arrays.asList(new String[] {}), UTF8,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsString(
                Arrays.asList(new String[] { "Hello World!", "",
                        "123456789012345678901234567890123456789" }),
                UTF8, ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CTRL_SHORT value.
     */
    @Test
    public void controlsShort() {
        ChannelAccessControlsShort value;
        value = ChannelAccessValueFactory.createControlsShort(
                ArrayUtils.EMPTY_SHORT_ARRAY,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, (short) 0, (short) 150,
                (short) 10, (short) 120, (short) -20, (short) 190, "mA", UTF8,
                (short) -15, (short) 185);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsShort(new short[] { 5 },
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, (short) 0, (short) 150,
                (short) 10, (short) 120, (short) -20, (short) 190, "mA", UTF8,
                (short) -15, (short) 185);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsShort(new short[] {},
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, (short) 0, (short) 150,
                (short) 10, (short) 120, (short) -20, (short) 190, "mA", UTF8,
                (short) -15, (short) 185);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsShort(
                new short[] { 18, 9353, -2352, 124, 0 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, (short) 0, (short) 150,
                (short) 10, (short) 120, (short) -20, (short) 190, "mA", UTF8,
                (short) -15, (short) 185);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CTRL_FLOAT value.
     */
    @Test
    public void controlsFloat() {
        ChannelAccessControlsFloat value;
        value = ChannelAccessValueFactory.createControlsFloat(
                ArrayUtils.EMPTY_FLOAT_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0f, 150.0f, 10.0f, 120.0f,
                -20.0f, 190.0f, "mA", UTF8, (short) 2, -15.0f, 185.0f);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsFloat(
                new float[] { 5.0f }, ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0f, 150.0f, 10.0f, 120.0f,
                -20.0f, 190.0f, "mA", UTF8, (short) 2, -15.0f, 185.0f);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsFloat(new float[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0f, 150.0f, 10.0f, 120.0f,
                -20.0f, 190.0f, "mA", UTF8, (short) 2, -15.0f, 185.0f);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsFloat(
                new float[] { 5.0f, -3.0f, 999.8f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0f, 150.0f, 10.0f, 120.0f,
                -20.0f, 190.0f, "mA", UTF8, (short) 2, -15.0f, 185.0f);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CTRL_ENUM value.
     */
    @Test
    public void controlsEnum() {
        ChannelAccessGraphicsEnum value;
        value = ChannelAccessValueFactory.createControlsEnum(
                ArrayUtils.EMPTY_SHORT_ARRAY,
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT,
                Arrays.asList(new String[] { "off" }), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsEnum(new short[] { 5 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT,
                Arrays.asList(new String[] { "off" }), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsEnum(new short[] {},
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT,
                Arrays.asList(new String[] {}), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsEnum(
                new short[] { 17, -88, 2 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.HWLIMIT,
                Arrays.asList(new String[] { "off", "on", "unknown" }), UTF8);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CTRL_CHAR value.
     */
    @Test
    public void controlsChar() {
        ChannelAccessControlsChar value;
        value = ChannelAccessValueFactory.createControlsChar(
                ArrayUtils.EMPTY_BYTE_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, (byte) 0, (byte) (byte) 130,
                (byte) 10, (byte) 100, (byte) -20, (byte) 100, "dB", UTF8,
                (byte) -15, (byte) 66);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsChar(new byte[] { 5 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, (byte) 0, (byte) (byte) 130,
                (byte) 10, (byte) 100, (byte) -20, (byte) 100, "dB", UTF8,
                (byte) -15, (byte) 66);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsChar(new byte[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, (byte) 0, (byte) (byte) 130,
                (byte) 10, (byte) 100, (byte) -20, (byte) 100, "dB", UTF8,
                (byte) -15, (byte) 66);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsChar(
                new byte[] { 5, 37, -18 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.LOW, (byte) 0, (byte) (byte) 130,
                (byte) 10, (byte) 100, (byte) -20, (byte) 100, "dB", UTF8,
                (byte) -15, (byte) 66);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CTRL_LONG value.
     */
    @Test
    public void controlsLong() {
        ChannelAccessControlsLong value;
        value = ChannelAccessValueFactory.createControlsLong(
                ArrayUtils.EMPTY_INT_ARRAY,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, 0, 150, 10, 120, -20, 190, "mA",
                UTF8, -15, 185);
        value = ChannelAccessValueFactory.createControlsLong(new int[] { 5 },
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.COMM, 0, 150, 10, 120, -20, 190, "mA",
                UTF8, -15, 185);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsLong(new int[] {},
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, 0, 150, 10, 120, -20, 190,
                "mA", UTF8, -15, 185);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsLong(
                new int[] { 18, 9353, -2352, 124, 0 },
                ChannelAccessAlarmSeverity.INVALID_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, 0, 150, 10, 120, -20, 190,
                "mA", UTF8, -15, 185);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CTRL_DOUBLE value.
     */
    @Test
    public void controlsDouble() {
        ChannelAccessControlsDouble value;
        value = ChannelAccessValueFactory.createControlsDouble(
                ArrayUtils.EMPTY_DOUBLE_ARRAY,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0, 150.0, 10.0, 120.0,
                -20.0, 190.0, "mA", UTF8, (short) 2, -15.0, 185.0);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsDouble(
                new double[] { 5.0 }, ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0, 150.0, 10.0, 120.0,
                -20.0, 190.0, "mA", UTF8, (short) 2, -15.0, 185.0);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsDouble(new double[] {},
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0, 150.0, 10.0, 120.0,
                -20.0, 190.0, "mA", UTF8, (short) 2, -15.0, 185.0);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createControlsDouble(
                new double[] { 5.0, -3.0, 999.8 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, 0.0, 150.0, 10.0, 120.0,
                -20.0, 190.0, "mA", UTF8, (short) 2, -15.0, 185.0);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_PUT_ACKT value.
     */
    @Test
    public void configureAcknowledgeTransientAlarms() {
        ChannelAccessConfigureAcknowledgeTransientAlarms value;
        value = ChannelAccessValueFactory
                .createConfigureAcknowledgeTransientAlarms(true);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createConfigureAcknowledgeTransientAlarms(false);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createConfigureAcknowledgeTransientAlarms(false);
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
        // We will never use such a value, but we might receive it over the
        // network, so we have to test it.
        value.setValue(new short[] { 5, 26, -3 });
        assertEquals(value, reserializeValue(value));
        // The same applies to an empty value.
        value.setValue(ArrayUtils.EMPTY_SHORT_ARRAY);
    }

    /**
     * Serialize and deserialize a DBR_PUT_ACKS value.
     */
    @Test
    public void acknowledgeAlarm() {
        ChannelAccessAcknowledgeAlarm value;
        value = ChannelAccessValueFactory
                .createAcknowledgeAlarm(ChannelAccessAlarmSeverity.NO_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createAcknowledgeAlarm(ChannelAccessAlarmSeverity.MINOR_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createAcknowledgeAlarm(ChannelAccessAlarmSeverity.MAJOR_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAcknowledgeAlarm(
                ChannelAccessAlarmSeverity.INVALID_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createAcknowledgeAlarm(ChannelAccessAlarmSeverity.NO_ALARM);
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
        // We will never use such a value, but we might receive it over the
        // network, so we have to test it.
        value.setValue(new short[] { 5, 26, -3 });
        ChannelAccessAcknowledgeAlarm reserializedValue = (ChannelAccessAcknowledgeAlarm) reserializeValue(
                value);
        assertEquals(value, reserializedValue);
        assertEquals(ChannelAccessAlarmSeverity.INVALID_ALARM,
                reserializedValue.getAcknowledgeAlarmSeverity());
        // The same applies to an empty value.
        value.setValue(ArrayUtils.EMPTY_SHORT_ARRAY);
        assertEquals(value, reserializeValue(value));
    }

    /**
     * Serialize and deserialize a DBR_STSACK_STRING value.
     */
    @Test
    public void alarmAcknowledgementStatus() {
        ChannelAccessAlarmAcknowledgementStatus value;
        value = ChannelAccessValueFactory.createAlarmAcknowledgementStatus(
                Collections.emptyList(), UTF8,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, true,
                ChannelAccessAlarmSeverity.MINOR_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmAcknowledgementStatus(
                Arrays.asList(new String[] { "foo" }), UTF8,
                ChannelAccessAlarmSeverity.NO_ALARM,
                ChannelAccessAlarmStatus.NO_ALARM, true,
                ChannelAccessAlarmSeverity.MINOR_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmAcknowledgementStatus(
                Arrays.asList(new String[] {}), UTF8,
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.HIGH, false,
                ChannelAccessAlarmSeverity.MAJOR_ALARM);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createAlarmAcknowledgementStatus(
                Arrays.asList(new String[] { "abc", "def" }), UTF8,
                ChannelAccessAlarmSeverity.MAJOR_ALARM,
                ChannelAccessAlarmStatus.LOLO, false,
                ChannelAccessAlarmSeverity.INVALID_ALARM);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Serialize and deserialize a DBR_CLASS_NAME value.
     */
    @Test
    public void className() {
        ChannelAccessClassName value;
        value = ChannelAccessValueFactory.createClassName(
                Collections.emptyList(), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory.createClassName(
                Arrays.asList(new String[] { "Hello World!" }), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createClassName(Arrays.asList(new String[] {}), UTF8);
        assertEquals(value, reserializeValue(value));
        value = ChannelAccessValueFactory
                .createClassName(
                        Arrays.asList(new String[] { "Hello World!", "",
                                "123456789012345678901234567890123456789" }),
                        UTF8);
        assertEquals(value, reserializeValue(value));
        // Test serialization of a read-only value.
        assertEquals(value, reserializeValue(value.asReadOnlyValue()));
    }

    /**
     * Test the <code>encodeGettableValue</code> and
     * <code>decodeGettableValue</code> methods. They are just wrappers around
     * <code>encodeValue</code> and <code>decodeValue</code>, so we keep the
     * test simple.
     */
    @Test
    public void gettableValue() {
        ChannelAccessTimeDouble value;
        value = ChannelAccessValueFactory.createTimeDouble(new double[] { 5.0 },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB, -17, 0);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodeGettableValue(byteSink, value,
                value.getValueSize());
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        ChannelAccessTimeDouble reserializedValue = (ChannelAccessTimeDouble) ChannelAccessValueCodec
                .decodeGettableValue(byteSource, value.getType(),
                        byteSource.remaining(), value.getValueSize(), UTF8);
        assertEquals(value, reserializedValue);
    }

    /**
     * Test the <code>encodePuttableValue</code> and
     * <code>decodePuttableValue</code> methods. They are just wrappers around
     * <code>encodeValue</code> and <code>decodeValue</code>, so we keep the
     * test simple.
     */
    @Test
    public void puttableValue() {
        ChannelAccessSimpleOnlyString value;
        value = ChannelAccessValueFactory.createString(
                Arrays.asList(new String[] { "Hello World!" }), UTF8);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodePuttableValue(byteSink, value,
                value.getValueSize());
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        ChannelAccessSimpleOnlyString reserializedValue = (ChannelAccessSimpleOnlyString) ChannelAccessValueCodec
                .decodeGettableValue(byteSource, value.getType(),
                        byteSource.remaining(), value.getValueSize(), UTF8);
        assertEquals(value, reserializedValue);
    }

    /**
     * Test the <code>decodeGettableValue</code> method with a type that is not
     * gettable.
     */
    @Test(expected = IllegalArgumentException.class)
    public void decodeGettableValueInvalidType() {
        ChannelAccessAcknowledgeAlarm value;
        value = ChannelAccessValueFactory
                .createAcknowledgeAlarm(ChannelAccessAlarmSeverity.MAJOR_ALARM);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodePuttableValue(byteSink, value,
                value.getValueSize());
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        ChannelAccessValueCodec.decodeGettableValue(byteSource, value.getType(),
                byteSource.remaining(), value.getValueSize(), UTF8);
    }

    /**
     * Test the <code>decodePuttableValue</code> method with a type that is not
     * puttable.
     */
    @Test(expected = IllegalArgumentException.class)
    public void decodePuttableValueInvalidType() {
        ChannelAccessAlarmFloat value;
        value = ChannelAccessValueFactory.createAlarmFloat(new float[] { 5.0f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodeGettableValue(byteSink, value,
                value.getValueSize());
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        ChannelAccessValueCodec.decodePuttableValue(byteSource, value.getType(),
                byteSource.remaining(), value.getValueSize(), UTF8);
    }

    /**
     * Test the <code>decodeValue</code> method with a negative count.
     */
    @Test(expected = IllegalArgumentException.class)
    public void decodeValueNegativeCount() {
        ChannelAccessAlarmFloat value;
        value = ChannelAccessValueFactory.createAlarmFloat(new float[] { 5.0f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodeValue(byteSink, value,
                value.getValueSize());
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        ChannelAccessValueCodec.decodeValue(byteSource, value.getType(),
                byteSource.remaining(), -1, UTF8);
    }

    /**
     * Test the <code>decodeValue</code> method with a too small data size.
     */
    @Test(expected = IllegalArgumentException.class)
    public void decodeValueTooSmallDataSize() {
        ChannelAccessAlarmFloat value;
        value = ChannelAccessValueFactory.createAlarmFloat(new float[] { 5.0f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodeValue(byteSink, value,
                value.getValueSize());
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        ChannelAccessValueCodec.decodeValue(byteSource, value.getType(),
                byteSource.remaining() - 1, value.getValueSize(), UTF8);
    }

    /**
     * Test the <code>encodeValue</code> method with a too large count.
     */
    @Test(expected = IllegalArgumentException.class)
    public void encodeValueTooLargeCount() {
        ChannelAccessAlarmFloat value;
        value = ChannelAccessValueFactory.createAlarmFloat(new float[] { 5.0f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodeValue(byteSink, value,
                value.getValueSize() + 1);
    }

    /**
     * Test the <code>encodeValue</code> method with a negative count.
     */
    @Test(expected = IllegalArgumentException.class)
    public void encodeValueNegativeCount() {
        ChannelAccessAlarmFloat value;
        value = ChannelAccessValueFactory.createAlarmFloat(new float[] { 5.0f },
                ChannelAccessAlarmSeverity.MINOR_ALARM,
                ChannelAccessAlarmStatus.BAD_SUB);
        ByteBufferByteSink byteSink = createByteSink();
        ChannelAccessValueCodec.encodeValue(byteSink, value, -1);
    }

    /**
     * Test the <code>calculateMaxCount</code> method with a few corner cases.
     */
    @Test
    public void calculateMaxCount() {
        assertEquals(1, ChannelAccessValueCodec
                .calculateMaxCount(ChannelAccessValueType.DBR_CHAR, 1));
        assertEquals(0, ChannelAccessValueCodec
                .calculateMaxCount(ChannelAccessValueType.DBR_CHAR, 0));
        assertEquals(Integer.MAX_VALUE,
                ChannelAccessValueCodec.calculateMaxCount(
                        ChannelAccessValueType.DBR_CHAR, Integer.MAX_VALUE));
        assertEquals(0, ChannelAccessValueCodec.calculateMaxCount(
                ChannelAccessValueType.DBR_GR_ENUM, Integer.MIN_VALUE));
        assertEquals(1, ChannelAccessValueCodec
                .calculateMaxCount(ChannelAccessValueType.DBR_STS_STRING, 44));
        assertEquals(0, ChannelAccessValueCodec
                .calculateMaxCount(ChannelAccessValueType.DBR_STS_STRING, 43));
        assertEquals(1, ChannelAccessValueCodec
                .calculateMaxCount(ChannelAccessValueType.DBR_STS_STRING, 83));
        assertEquals(2, ChannelAccessValueCodec
                .calculateMaxCount(ChannelAccessValueType.DBR_STS_STRING, 84));
    }

    /**
     * Test the <code>calculatePayloadSize</code> with a negative count.
     */
    @Test(expected = IllegalArgumentException.class)
    public void calculatePayloadSizeNegativeCount() {
        ChannelAccessValueCodec
                .calculatePayloadSize(ChannelAccessValueType.DBR_CTRL_CHAR, -1);
    }

    /**
     * Test the <code>calculatePayloadSize</code> with a too large count.
     */
    @Test(expected = IllegalArgumentException.class)
    public void calculatePayloadSizeTooLargeCountCount() {
        assertEquals(2147483644, ChannelAccessValueCodec.calculatePayloadSize(
                ChannelAccessValueType.DBR_DOUBLE, 536870911));
        assertEquals(2147483644, ChannelAccessValueCodec.calculatePayloadSize(
                ChannelAccessValueType.DBR_GR_DOUBLE, 536870903));
        // Should cause an IllegalArgumentException.
        ChannelAccessValueCodec.calculatePayloadSize(
                ChannelAccessValueType.DBR_GR_DOUBLE, 536870904);
    }

}
