/*
 * Copyright 2015-2024 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.net.InetSocketAddress;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;

import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.exception.ConcurrentNotificationException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.util.ListenerLockPolicy;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmAcknowledgementStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyString;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessClassName;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyString;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeString;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Representation of a Channel Access channel. A channel represents a process
 * variable that can be accessed via the Channel Access protocol.
 * </p>
 * 
 * <p>
 * Channels can be retrieved from a {@link ChannelAccessClient}. A channel
 * supports three different kinds of operations:
 * </p>
 * 
 * <ul>
 * <li>get (read),</li>
 * <li>put (write), and</li>
 * <li>monitor (subscription that receives updates on value changes).</li>
 * </ul>
 * 
 * <p>
 * Once created, a channel can be used until it is {@linkplain #destroy()
 * destroyed}. A channel keeps strong references to its
 * {@link ChannelAccessConnectionListener ChannelAccessConnectionListeners} and
 * to the {@link ChannelAccessClient}. However, depending on the implementation,
 * it might keep strong or weak references to its {@link ChannelAccessMonitor
 * ChannelAccessMonitors}. {@link ChannelAccessMonitor ChannelAccessMonitors}
 * keep a strong reference to their channel.
 * </p>
 * 
 * <p>
 * The {@link ChannelAccessClient} may keep strong or weak references to its
 * channels. Therefore, code using a channel (e.g. by registering a
 * {@link ChannelAccessConnectionListener} with it) should always keep a strong
 * reference to this channel in order to keep it alive. This strong reference
 * cannot be kept through the listener (unless an external strong reference to
 * the listener is kept) because this will not avoid garbage collection of the
 * channel. However, you still should call the channel's {@link #destroy()}
 * method if you do not need the channel any longer, because some
 * implementations might choose to keep a strong reference to the channel and in
 * this case the channel will not be destroyed unless you destroy it explicitly.
 * </p>
 * 
 * <p>
 * It is safe to create a separate channel instance for every place where one is
 * needed. Channel instances that refer to the same channel name (and belong to
 * the same {@link ChannelAccessClient}) will actually share one connection, so
 * that there is no significant overhead in having many channel objects for the
 * same channel. Having a separate channel instance for each purpose makes it
 * easier to decide when a channel is not needed any longer and can be
 * destroyed.
 * </p>
 * 
 * <p>
 * Please note that destroying a channel will also destroy all its monitors and
 * destroying a {@link ChannelAccessClient} will destroy all its channels (and
 * thus all their monitors).
 * </p>
 * 
 * <p>
 * Due to their asynchronous nature, the <code>get</code> and <code>put</code>
 * operations supplied by this interface do not return direct results. Instead,
 * {@link ListenableFuture}s that allow to wait for a result or to be notified
 * when a result is available are supplied.
 * </p>
 * 
 * <p>
 * These futures' {@link ListenableFuture#get() get} methods might throw an
 * {@link ExecutionException} that wraps a {@link ChannelAccessException}. This
 * happens if the error cannot be detected before the operation is started.
 * Otherwise, the method that starts the operation will throw the exception
 * directly.
 * </p>
 * 
 * <p>
 * Monitors, on the other hand, can always be created and signal error
 * conditions by calling the
 * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
 * monitorError(...)} method of their listeners.
 * </p>
 *
 * <h3 id="zero-count">Automatically determining the element count</h3>
 *
 * <p>
 * The versions of the <code>get</code> and <code>monitor</code> methods that do
 * not accept an explicit <code>count</code> determine the element count to be
 * requested automatically. The element count is also determined automatically
 * when a count of zero is requested. The exact behavior depends on the
 * configuration:
 * </p>
 *
 * <p>
 * If
 * {@link ChannelAccessClientConfiguration#isUseServerDeterminedElementCount()
 * useServerDeterminedElementCount} is <code>true</code> and the server
 * providing the channel supports version 4.13 of the Channel Access protocol,
 * the number of elements to be sent is determined by the server.
 * </p>
 *
 * <p>
 * If
 * {@link ChannelAccessClientConfiguration#isUseServerDeterminedElementCount()
 * useServerDeterminedElementCount} is <code>false</code> or the server
 * providing the channel does not support version 4.13 of the Channel Access
 * protocol, the client requests the channel’s native element count. Please note
 * that a channel’s native element count may still vary over time as the channel
 * gets disconnected and reconnected, so the number of elements that are
 * received for monitor subscriptions may vary as well, even when using this
 * mode.
 * </p>
 *
 * @author Sebastian Marsching
 */
public interface ChannelAccessChannel {

    /**
     * <p>
     * Destroys this channel. The channel cannot be used any longer. This will
     * also destroy all monitors that belong to this channel. This method should
     * be called when the channel is not needed any longer.
     * </p>
     * 
     * <p>
     * Calling this method will not destroy other channel instances that have
     * been created by separate calls to
     * {@link ChannelAccessClient#getChannel(String)}, even if they refer to the
     * same channel.
     * </p>
     * 
     * <p>
     * If the {@link ListenerLockPolicy#BLOCK BLOCK} policy is in effect, the
     * connection-state listeners are not going to receive another notification
     * after this method returns. If the {@link ListenerLockPolicy#IGNORE
     * IGNORE} policy is in effect, the listeners might be notified once more
     * after this method returns. If the {@link ListenerLockPolicy#REPORT
     * REPORT} policy is in effect, this method throws a
     * {@link ConcurrentNotificationException} if the listeners might be
     * notified once more. The channel is still destroyed, despite the exception
     * being thrown.
     * </p>
     * 
     * @throws ConcurrentNotificationException
     *             if the {@link ListenerLockPolicy} is set to {@code REPORT}
     *             and the connection-state listeners might receive one more
     *             notification after this method has returned. The channel is
     *             still destroyed, even if this exception is thrown.
     * @see #getListenerLockPolicy()
     */
    void destroy();

    /**
     * Returns the listener lock-policy used by this channel. The listener
     * lock-policy defines the behavior of the {@link #destroy()} and
     * {@link #removeConnectionListener(ChannelAccessConnectionListener)}
     * methods.
     * 
     * @return listener lock-policy used by this channel.
     * @since EPICS Jackie 2.0.0
     */
    ListenerLockPolicy getListenerLockPolicy();

    /**
     * Returns the name of this channel. The name identifies the channel in the
     * network.
     * 
     * @return channel name.
     */
    String getName();

    /**
     * Returns the Channel Access client this channel belongs to. This is the
     * client that was used to create this channel.
     * 
     * @return Channel Access client this channel belongs to.
     */
    ChannelAccessClient getClient();

    /**
     * <p>
     * Returns the native element count of this channel. The native element
     * count is the number of element that the server sends when the channel is
     * connected.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute might change throughout the life-cycle of the channel. For
     * a disconnected channel, this method throws an
     * {@link IllegalStateException}.
     * </p>
     * 
     * @return native element count reported by the server.
     * @throws IllegalStateException
     *             if this channel is not connected ( {@link #isConnected()}
     *             returns <code>false</code>).
     */
    int getNativeCount();

    /**
     * <p>
     * Returns the native data type of this channel. The native data type is the
     * data type that the server sends when the channel is connected.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute change throughout the life-cycle of the channel. For a
     * disconnected channel, this method throws an {@link IllegalStateException}
     * .
     * </p>
     * 
     * @return native data type reported by the server.
     * @throws IllegalStateException
     *             if this channel is not connected ( {@link #isConnected()}
     *             returns <code>false</code>).
     */
    ChannelAccessValueType getNativeDataType();

    /**
     * <p>
     * Returns the IP address and port number of the server that provides this
     * channel. In theory, there can be more than one server providing the same
     * channel. In this case, the returned information pertains to the server
     * that is actually used by this channel instance.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute might change throughout the life-cycle of the channel. For
     * a disconnected channel, this method throws an
     * {@link IllegalStateException}.
     * </p>
     * 
     * @return IP address and port number of the server that provides this
     *         channel.
     * @throws IllegalStateException
     *             if this channel is not connected ({@link #isConnected()}
     *             returns <code>false</code>).
     * @since EPICS Jackie 3.0.0
     */
    InetSocketAddress getServerAddress();

    /**
     * Tells whether this channel is currently connected.
     * 
     * @return <code>true</code> if the channel is currently connected,
     *         <code>false</code> if it is currently disconnected.
     * @see #isConnectedOnce()
     */
    boolean isConnected();

    /**
     * Tells whether this channel has been connected at least once.
     * 
     * @return <code>true</code> if this channel is currently connected or has
     *         previously been connected, <code>false</code> otherwise.
     * @see #isConnected()
     */
    boolean isConnectedOnce();

    /**
     * <p>
     * Tells whether this channel has been destroyed. A channel is destroyed
     * when its {@link #destroy()} method is called or when its
     * {@linkplain #getClient() client} is destroyed. It might also be destroyed
     * if it is not used any longer (there are no strong references to it and it
     * is garbage collected).
     * </p>
     * 
     * <p>
     * After a channel has been destroyed, it cannot be used any longer and most
     * of its methods will throw an exception.
     * </p>
     * 
     * @return <code>true</code> if this channel has been destroyed,
     *         <code>false</code> if it is still alive.
     */
    boolean isDestroyed();

    /**
     * <p>
     * Tells whether this client is allowed to read from this channel. If the
     * client is allowed to read, the get and monitor operations can be used.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute might change throughout the life-cycle of the channel. For
     * a disconnected channel, this method throws an
     * {@link IllegalStateException}.
     * </p>
     * 
     * @return <code>true</code> if this client is allowed to read from this
     *         channel and <code>false</code> if it not allowed to use read
     *         operations on this channel.
     * @throws IllegalStateException
     *             if this channel is not connected ({@link #isConnected()}
     *             returns <code>false</code>).
     */
    boolean isMayRead();

    /**
     * <p>
     * Tells whether this client is allowed to write this channel. If the client
     * is allowed to write, the put operations can be used.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute might change throughout the life-cycle of the channel. For
     * a disconnected channel, this method throws an
     * {@link IllegalStateException}.
     * </p>
     * 
     * @return <code>true</code> if this client is allowed to write to this
     *         channel and <code>false</code> if it not allowed to use write
     *         operations on this channel.
     * @throws IllegalStateException
     *             if this channel is not connected ({@link #isConnected()}
     *             returns <code>false</code>).
     */
    boolean isMayWrite();

    /**
     * <p>
     * Registers a connection state listener with this channel. The listener
     * will be notified when this channel's connection state changes. When the
     * listener is added, it is called once to notify it of the current
     * connection state.
     * </p>
     * 
     * <p>
     * If the specified listener is already registered with this channel, no
     * action is taken and this method returns <code>false</code>.
     * </p>
     * 
     * <p>
     * <strong>Important note</strong>: Typically, the listener is executed by
     * the thread that performs network I/O. For this reason, it is important
     * that the listener does not block or perform any long-running operations.
     * Instead, the listener should delegate any such operation to a different
     * thread (e.g. using an {@link Executor}). Any violation of this rule might
     * result in unexpected or unpredictable behavior. In particular, a blocking
     * listener might cause a dead-lock, which results in the whole Channel
     * Access client being stalled indefinitely.
     * </p>
     * 
     * <p>
     * If the calling code is not interested in receiving permanent status
     * updates and only wants to wait for a certain connection state to be
     * established, {@link #waitForConnectionState(boolean)} might be a simpler
     * alternative to this method.
     * </p>
     * 
     * @param listener
     *            listener to be registered with this channel. Must not be
     *            <code>null</code>.
     * @return <code>true</code> if the listener has been added,
     *         <code>false</code> if the listener has already been added
     *         earlier.
     * @see #removeConnectionListener(ChannelAccessConnectionListener)
     * @see #waitForConnectionState(boolean)
     * @throws NullPointerException
     *             if <code>listener</code> is null.
     */
    boolean addConnectionListener(ChannelAccessConnectionListener listener);

    /**
     * <p>
     * Unregisters a connection state listener from this channel. The exact
     * behavior of this method depends on the
     * {@linkplain #getListenerLockPolicy() listener lock-policy} used by this
     * channel.
     * </p>
     * 
     * <p>
     * If the {@link ListenerLockPolicy#BLOCK BLOCK} policy is in effect, the
     * listener is not going to receive another notification after this method
     * returns. If the {@link ListenerLockPolicy#IGNORE IGNORE} policy is in
     * effect, the listener might be notified once more after this method
     * returns. If the {@link ListenerLockPolicy#REPORT REPORT} policy is in
     * effect, this method throws a {@link ConcurrentNotificationException} if
     * the listener might be notified once more. The listener is still removed,
     * despite the exception being thrown.
     * </p>
     * 
     * <p>
     * If the specified listener has already been removed from this channel or
     * has never been added, no action is taken and this method returns
     * <code>false</code>.
     * </p>
     * 
     * @param listener
     *            listener to be unregistered from this channel.
     * @return <code>true</code> if the listener has been removed,
     *         <code>false</code> if the listener has already been removed
     *         earlier or if it has never been added.
     * @see #addConnectionListener(ChannelAccessConnectionListener)
     * @throws ConcurrentNotificationException
     *             if the {@link ListenerLockPolicy} is set to {@code REPORT}
     *             and the removed listener might receive one more notification
     *             after this method has returned. The listener is still
     *             removed, even if this exception is thrown.
     * @see #getListenerLockPolicy()
     */
    boolean removeConnectionListener(ChannelAccessConnectionListener listener);

    /**
     * <p>
     * Returns a listenable future that completes when the specified connection
     * state is established. This method can be used if certain actions are to
     * be delayed until the channel is connected or disconnected.
     * </p>
     * 
     * <p>
     * If the channel is already in the specified connection state, the future
     * returned by this method completes immediately.
     * </p>
     * 
     * <p>
     * Due to the asynchronous nature of connection events, there is no
     * guarantee that the channel is still in the specified connection state
     * when the future has completed. It is possible that the channel again
     * changed its state after first switching to the desired state. If the code
     * waiting for the future detects such a situation, it should simply make
     * another call to this method to wait for the connection state to change
     * again.
     * </p>
     * 
     * @param connectionState
     *            desired connection state. If <code>true</code>, the future
     *            returned by this method completes when the channel is
     *            connected. If <code>false</code>, the future returned by this
     *            method completes when the channel is disconnected.
     * @return listenable future that completes when the channel changes to the
     *         specified <code>connectionState</code>. If the channel already is
     *         in the specified <code>connectionState</code>, the future
     *         completes immediately. The completed future's value will be the
     *         connection state at the time when it completed (always the
     *         specified <code>connectionState</code>).
     */
    ListenableFuture<Boolean> waitForConnectionState(boolean connectionState);

    /**
     * <p>
     * Gets the current value from the channel using the specified type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be returned by the server. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes. The
     *         type of the value returned by the future can be expected to match
     *         the specified <code>valueType</code>.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<? extends ChannelAccessGettableValue<?>> get(
            ChannelAccessValueType valueType) {
        return get(valueType, 0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the specified type and
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be returned by the server. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes. The
     *         type of the value returned by the future can be expected to match
     *         the specified <code>valueType</code>.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation or if the specified <code>count</code> is
     *             negative.
     * @throws IllegalStateException
     *             if this channel has already been {@link #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<? extends ChannelAccessGettableValue<?>> get(
            ChannelAccessValueType valueType, int count);

    /**
     * <p>
     * Gets the current value from the channel using the channel's
     * {@linkplain #getNativeDataType() native data type} and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<? extends ChannelAccessGettableValue<?>> getNative() {
        return get(null, 0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the channel's
     * {@linkplain #getNativeDataType() native data type} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<? extends ChannelAccessGettableValue<?>> getNative(
            int count) {
        return get(null, count);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessSimpleOnlyString> getString() {
        return getString(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyString> getString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyString> getAlarmString() {
        return getAlarmString(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getAlarmString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessTimeString> getTimeString() {
        return getTimeString(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeString> getTimeString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type and
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyString> getGraphicsString() {
        return getGraphicsString(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getGraphicsString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyString> getControlsString() {
        return getControlsString(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getControlsString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessSimpleOnlyShort> getShort() {
        return getShort(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyShort> getShort(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyShort> getAlarmShort() {
        return getAlarmShort(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyShort> getAlarmShort(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessTimeShort> getTimeShort() {
        return getTimeShort(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeShort> getTimeShort(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessGraphicsOnlyShort> getGraphicsShort() {
        return getGraphicsShort(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyShort> getGraphicsShort(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessControlsShort> getControlsShort() {
        return getControlsShort(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsShort> getControlsShort(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessSimpleOnlyFloat> getFloat() {
        return getFloat(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyFloat> getFloat(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyFloat> getAlarmFloat() {
        return getAlarmFloat(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyFloat> getAlarmFloat(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessTimeFloat> getTimeFloat() {
        return getTimeFloat(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeFloat> getTimeFloat(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessGraphicsOnlyFloat> getGraphicsFloat() {
        return getGraphicsFloat(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyFloat> getGraphicsFloat(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessControlsFloat> getControlsFloat() {
        return getControlsFloat(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsFloat> getControlsFloat(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessSimpleOnlyEnum> getEnum() {
        return getEnum(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyEnum> getEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyEnum> getAlarmEnum() {
        return getAlarmEnum(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyEnum> getAlarmEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessTimeEnum> getTimeEnum() {
        return getTimeEnum(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeEnum> getTimeEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessGraphicsEnum> getGraphicsEnum() {
        return getGraphicsEnum(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsEnum> getGraphicsEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessGraphicsEnum> getControlsEnum() {
        return getControlsEnum(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsEnum> getControlsEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessSimpleOnlyChar> getChar() {
        return getChar(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyChar> getChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyChar> getAlarmChar() {
        return getAlarmChar(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyChar> getAlarmChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessTimeChar> getTimeChar() {
        return getTimeChar(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeChar> getTimeChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessGraphicsOnlyChar> getGraphicsChar() {
        return getGraphicsChar(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyChar> getGraphicsChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessControlsChar> getControlsChar() {
        return getControlsChar(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsChar> getControlsChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessSimpleOnlyLong> getLong() {
        return getLong(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyLong> getLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyLong> getAlarmLong() {
        return getAlarmLong(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyLong> getAlarmLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessTimeLong> getTimeLong() {
        return getTimeLong(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeLong> getTimeLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessGraphicsOnlyLong> getGraphicsLong() {
        return getGraphicsLong(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyLong> getGraphicsLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessControlsLong> getControlsLong() {
        return getControlsLong(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsLong> getControlsLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type and an
     * <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessSimpleOnlyDouble> getDouble() {
        return getDouble(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyDouble> getDouble(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmOnlyDouble> getAlarmDouble() {
        return getAlarmDouble(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyDouble> getAlarmDouble(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessTimeDouble> getTimeDouble() {
        return getTimeDouble(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeDouble> getTimeDouble(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type and
     * an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessGraphicsOnlyDouble> getGraphicsDouble() {
        return getGraphicsDouble(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyDouble> getGraphicsDouble(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessControlsDouble> getControlsDouble() {
        return getControlsDouble(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsDouble> getControlsDouble(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type and an <a href="#zero-count">automatically determined element
     * count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessAlarmAcknowledgementStatus> getAlarmAcknowledgementStatus() {
        return getAlarmAcknowledgementStatus(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} and
     * the specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmAcknowledgementStatus> getAlarmAcknowledgementStatus(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type
     * and an <a href="#zero-count">automatically determined element count</a>.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<ChannelAccessClassName> getClassName() {
        return getClassName(0);
    }

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the get operation only returns
     *            the number of elements that is equal to the channel's native
     *            element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessClassName> getClassName(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the specified type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be sent for monitor events. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The type of the values supplied with monitor
     *         events can be expected to match the specified
     *         <code>valueType</code>.
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType) {
        return monitor(valueType, 0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the specified type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be sent for monitor events. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The type
     *         of the values supplied with monitor events can be expected to
     *         match the specified <code>valueType</code>.
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation or if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType,
            ChannelAccessEventMask eventMask) {
        return monitor(valueType, 0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the specified type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be sent for monitor events. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The type of the values supplied with monitor
     *         events can be expected to match the specified
     *         <code>valueType</code> and the values can be expected to not
     *         exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation or if the specified <code>count</code> is
     *             negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType, int count) {
        return monitor(valueType, count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the specified type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be sent for monitor events. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The type
     *         of the values supplied with monitor events can be expected to
     *         match the specified <code>valueType</code> and the values can be
     *         expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation, if the specified <code>count</code> is
     *             negative, or if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType, int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the channel's
     * {@linkplain #getNativeDataType() native data type}, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native data type might vary over
     * time, the type of the values might vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative() {
        return monitor(null, 0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the channel's
     * {@linkplain #getNativeDataType() native data type}, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask. As a channel's native data type might vary over
     * time, the type of the values might vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            ChannelAccessEventMask eventMask) {
        return monitor(null, 0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the channel's
     * {@linkplain #getNativeDataType() native data type}, the specified element
     * count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native data type might vary over
     * time, the type of the values of monitor events might vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            int count) {
        return monitor(null, count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the channel's
     * {@linkplain #getNativeDataType() native data type}, the specified element
     * count, and the specified event mask. As a channel's native data type
     * might vary over time, the type of the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            int count, ChannelAccessEventMask eventMask) {
        return monitor(null, count, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString() {
        return monitorString(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(
            ChannelAccessEventMask eventMask) {
        return monitorString(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(
            int count) {
        return monitorString(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString() {
        return monitorAlarmString(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmString(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            int count) {
        return monitorAlarmString(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString() {
        return monitorTimeString(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(
            ChannelAccessEventMask eventMask) {
        return monitorTimeString(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(
            int count) {
        return monitorTimeString(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString() {
        return monitorGraphicsString(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsString(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            int count) {
        return monitorGraphicsString(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString() {
        return monitorControlsString(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            ChannelAccessEventMask eventMask) {
        return monitorControlsString(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            int count) {
        return monitorControlsString(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort() {
        return monitorShort(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(
            ChannelAccessEventMask eventMask) {
        return monitorShort(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(
            int count) {
        return monitorShort(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort() {
        return monitorAlarmShort(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmShort(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            int count) {
        return monitorAlarmShort(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort() {
        return monitorTimeShort(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(
            ChannelAccessEventMask eventMask) {
        return monitorTimeShort(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(
            int count) {
        return monitorTimeShort(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort() {
        return monitorGraphicsShort(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsShort(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            int count) {
        return monitorGraphicsShort(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort() {
        return monitorControlsShort(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            ChannelAccessEventMask eventMask) {
        return monitorControlsShort(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            int count) {
        return monitorControlsShort(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat() {
        return monitorFloat(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(
            ChannelAccessEventMask eventMask) {
        return monitorFloat(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(
            int count) {
        return monitorFloat(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat() {
        return monitorAlarmFloat(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmFloat(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            int count) {
        return monitorAlarmFloat(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat() {
        return monitorTimeFloat(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(
            ChannelAccessEventMask eventMask) {
        return monitorTimeFloat(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(
            int count) {
        return monitorTimeFloat(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat() {
        return monitorGraphicsFloat(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsFloat(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            int count) {
        return monitorGraphicsFloat(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat() {
        return monitorControlsFloat(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            ChannelAccessEventMask eventMask) {
        return monitorControlsFloat(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            int count) {
        return monitorControlsFloat(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum() {
        return monitorEnum(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(
            ChannelAccessEventMask eventMask) {
        return monitorEnum(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type, the specified
     * element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(
            int count) {
        return monitorEnum(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type, the specified
     * element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum() {
        return monitorAlarmEnum(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmEnum(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(
            int count) {
        return monitorAlarmEnum(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum() {
        return monitorTimeEnum(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(
            ChannelAccessEventMask eventMask) {
        return monitorTimeEnum(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(
            int count) {
        return monitorTimeEnum(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum() {
        return monitorGraphicsEnum(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsEnum(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            int count) {
        return monitorGraphicsEnum(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum() {
        return monitorControlsEnum(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            ChannelAccessEventMask eventMask) {
        return monitorControlsEnum(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            int count) {
        return monitorControlsEnum(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar() {
        return monitorChar(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(
            ChannelAccessEventMask eventMask) {
        return monitorChar(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type, the specified
     * element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(
            int count) {
        return monitorChar(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type, the specified
     * element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar() {
        return monitorAlarmChar(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmChar(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(
            int count) {
        return monitorAlarmChar(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar() {
        return monitorTimeChar(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(
            ChannelAccessEventMask eventMask) {
        return monitorTimeChar(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(
            int count) {
        return monitorTimeChar(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar() {
        return monitorGraphicsChar(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsChar(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            int count) {
        return monitorGraphicsChar(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar() {
        return monitorControlsChar(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            ChannelAccessEventMask eventMask) {
        return monitorControlsChar(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            int count) {
        return monitorControlsChar(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong() {
        return monitorLong(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(
            ChannelAccessEventMask eventMask) {
        return monitorLong(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type, the specified
     * element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(
            int count) {
        return monitorLong(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type, the specified
     * element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong() {
        return monitorAlarmLong(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmLong(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(
            int count) {
        return monitorAlarmLong(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong() {
        return monitorTimeLong(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(
            ChannelAccessEventMask eventMask) {
        return monitorTimeLong(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(
            int count) {
        return monitorTimeLong(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong() {
        return monitorGraphicsLong(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsLong(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            int count) {
        return monitorGraphicsLong(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong() {
        return monitorControlsLong(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTLR_LONG} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            ChannelAccessEventMask eventMask) {
        return monitorControlsLong(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            int count) {
        return monitorControlsLong(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble() {
        return monitorDouble(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(
            ChannelAccessEventMask eventMask) {
        return monitorDouble(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(
            int count) {
        return monitorDouble(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble() {
        return monitorAlarmDouble(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmDouble(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            int count) {
        return monitorAlarmDouble(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble() {
        return monitorTimeDouble(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(
            ChannelAccessEventMask eventMask) {
        return monitorTimeDouble(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(
            int count) {
        return monitorTimeDouble(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble() {
        return monitorGraphicsDouble(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type, an
     * <a href="#zero-count">automatically determined element count</a>, and the
     * specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsDouble(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            int count) {
        return monitorGraphicsDouble(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble() {
        return monitorControlsDouble(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            ChannelAccessEventMask eventMask) {
        return monitorControlsDouble(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            int count) {
        return monitorControlsDouble(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type, an <a href="#zero-count">automatically determined element
     * count</a>, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus() {
        return monitorAlarmAcknowledgementStatus(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type, an <a href="#zero-count">automatically determined element
     * count</a>, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmAcknowledgementStatus(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type, the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            int count) {
        return monitorAlarmAcknowledgementStatus(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type, the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessClassName> monitorClassName() {
        return monitorClassName(0, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type,
     * an <a href="#zero-count">automatically determined element count</a>, and
     * the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(
            ChannelAccessEventMask eventMask) {
        return monitorClassName(0, eventMask);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(
            int count) {
        return monitorClassName(count, null);
    }

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     *
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero means that the number of
     *            elements shall be <a href="#zero-count">determined
     *            automatically</a>. If the specified number exceeds the
     *            channel's native element count, the monitor event will only
     *            provide the number of elements that is equal to the channel's
     *            native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Puts the specified value to the channel. The value can be any puttable
     * value. However, specifying an element count that is larger than the
     * channel's {@linkplain #getNativeCount() native element count} might
     * result in an error. However, most servers will simply ignore the extra
     * element's and only use the beginning of the array.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> put(ChannelAccessPuttableValue<?> value);

    /**
     * <p>
     * Puts the specified single element
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    default ListenableFuture<Void> putString(String value) {
        return putString(Collections.singleton(value));
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is copied, so subsequent changes to the
     * collection will not have any effect on the put operation.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     */
    default ListenableFuture<Void> putString(String[] value) {
        return putString(Arrays.asList(value));
    }

    /**
     * <p>
     * Puts the specified collection to the channel as a
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING}. Specifying a
     * collection that is larger than the channel's
     * {@linkplain #getNativeCount() native element count} might result in an
     * error. However, most servers will simply ignore the extra element's and
     * only use the beginning of the array. The specified collection is copied,
     * so subsequent changes to the collection will not have any effect on the
     * put operation.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     */
    default ListenableFuture<Void> putString(Collection<String> value) {
        return put(ChannelAccessValueFactory.createString(value,
                getClient().getConfiguration().getCharset()));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_SHORT
     * DBR_SHORT} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<Void> putShort(short value) {
        return putShort(new short[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    default ListenableFuture<Void> putShort(short[] value) {
        return put(ChannelAccessValueFactory.createShort(value));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_FLOAT
     * DBR_FLOAT} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<Void> putFloat(float value) {
        return putFloat(new float[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    default ListenableFuture<Void> putFloat(float[] value) {
        return put(ChannelAccessValueFactory.createFloat(value));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_ENUM
     * DBR_ENUM} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<Void> putEnum(short value) {
        return putEnum(new short[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    default ListenableFuture<Void> putEnum(short[] value) {
        return put(ChannelAccessValueFactory.createEnum(value));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_CHAR
     * DBR_CHAR} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<Void> putChar(byte value) {
        return putChar(new byte[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    default ListenableFuture<Void> putChar(byte[] value) {
        return put(ChannelAccessValueFactory.createChar(value));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_LONG
     * DBR_LONG} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<Void> putLong(int value) {
        return putLong(new int[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    default ListenableFuture<Void> putLong(int[] value) {
        return put(ChannelAccessValueFactory.createLong(value));
    }

    /**
     * <p>
     * Puts the specified single element
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<Void> putDouble(double value) {
        return putDouble(new double[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    default ListenableFuture<Void> putDouble(double[] value) {
        return put(ChannelAccessValueFactory.createDouble(value));
    }

    /**
     * <p>
     * Puts the specified {@link ChannelAccessValueType#DBR_PUT_ACKT
     * DBR_PUT_ACKT} to the channel. This can be used to configure the transient
     * alarm setting of the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param explicitlyAcknowledgeTransientAlarms
     *            <code>true</code> if transient alarms should have to be
     *            acknowledged explicitly, <code>false</code> if they should be
     *            automatically acknowledged when they disappear.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    default ListenableFuture<Void> putConfigureAcknowledgeTransientAlarms(
            boolean explicitlyAcknowledgeTransientAlarms) {
        return put(ChannelAccessValueFactory
                .createConfigureAcknowledgeTransientAlarms(
                        explicitlyAcknowledgeTransientAlarms));
    }

    /**
     * <p>
     * Puts the specified {@link ChannelAccessValueType#DBR_PUT_ACKS
     * DBR_PUT_ACKS} to the channel. This can be used to acknowledge alarms for
     * the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param acknowledgeAlarmSeverity
     *            highest alarm severity that should be acknowledged (never
     *            <code>null</code>). This means that the alarm will be
     *            acknowledged if the severity of the unacknowledged alarm is
     *            less than or equal to the specified severity.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if <code>acknowledgeAlarmSeverity</code> is
     *             <code>null</code>.
     */
    default ListenableFuture<Void> putAcknowledgeAlarm(
            ChannelAccessAlarmSeverity acknowledgeAlarmSeverity) {
        return put(ChannelAccessValueFactory
                .createAcknowledgeAlarm(acknowledgeAlarmSeverity));
    }

    /**
     * <p>
     * Puts the specified value to the channel. The value can be any puttable
     * value. However, specifying an element count that is larger than the
     * channel's {@linkplain #getNativeCount() native element count} might
     * result in an error. However, most servers will simply ignore the extra
     * element's and only use the beginning of the array.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    void putNoCallback(ChannelAccessPuttableValue<?> value);

    /**
     * <p>
     * Puts the specified single element
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} to the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    default void putStringNoCallback(String value) {
        putStringNoCallback(Collections.singleton(value));
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is copied, so subsequent changes to the
     * collection will not have any effect on the put operation.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     * @since EPICS Jackie 3.0.0
     */
    default void putStringNoCallback(String[] value) {
        putStringNoCallback(Arrays.asList(value));
    }

    /**
     * <p>
     * Puts the specified collection to the channel as a
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING}. Specifying a
     * collection that is larger than the channel's
     * {@linkplain #getNativeCount() native element count} might result in an
     * error. However, most servers will simply ignore the extra element's and
     * only use the beginning of the array. The specified collection is copied,
     * so subsequent changes to the collection will not have any effect on the
     * put operation.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     * @since EPICS Jackie 3.0.0
     */
    default void putStringNoCallback(Collection<String> value) {
        putNoCallback(ChannelAccessValueFactory.createString(value,
                getClient().getConfiguration().getCharset()));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_SHORT
     * DBR_SHORT} to the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @since EPICS Jackie 3.0.0
     */
    default void putShortNoCallback(short value) {
        putShortNoCallback(new short[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    default void putShortNoCallback(short[] value) {
        putNoCallback(ChannelAccessValueFactory.createShort(value));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_FLOAT
     * DBR_FLOAT} to the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @since EPICS Jackie 3.0.0
     */
    default void putFloatNoCallback(float value) {
        putFloatNoCallback(new float[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    default void putFloatNoCallback(float[] value) {
        putNoCallback(ChannelAccessValueFactory.createFloat(value));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_ENUM
     * DBR_ENUM} to the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @since EPICS Jackie 3.0.0
     */
    default void putEnumNoCallback(short value) {
        putEnumNoCallback(new short[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    default void putEnumNoCallback(short[] value) {
        putNoCallback(ChannelAccessValueFactory.createEnum(value));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_CHAR
     * DBR_CHAR} to the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @since EPICS Jackie 3.0.0
     */
    default void putCharNoCallback(byte value) {
        putCharNoCallback(new byte[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    default void putCharNoCallback(byte[] value) {
        putNoCallback(ChannelAccessValueFactory.createChar(value));
    }

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_LONG
     * DBR_LONG} to the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @since EPICS Jackie 3.0.0
     */
    default void putLongNoCallback(int value) {
        putLongNoCallback(new int[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    default void putLongNoCallback(int[] value) {
        putNoCallback(ChannelAccessValueFactory.createLong(value));
    }

    /**
     * <p>
     * Puts the specified single element
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} to the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @since EPICS Jackie 3.0.0
     */
    default void putDoubleNoCallback(double value) {
        putDoubleNoCallback(new double[] { value });
    }

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    default void putDoubleNoCallback(double[] value) {
        putNoCallback(ChannelAccessValueFactory.createDouble(value));
    }

    /**
     * <p>
     * Puts the specified {@link ChannelAccessValueType#DBR_PUT_ACKT
     * DBR_PUT_ACKT} to the channel. This can be used to configure the transient
     * alarm setting of the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param explicitlyAcknowledgeTransientAlarms
     *            <code>true</code> if transient alarms should have to be
     *            acknowledged explicitly, <code>false</code> if they should be
     *            automatically acknowledged when they disappear.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @since EPICS Jackie 3.0.0
     */
    default void putConfigureAcknowledgeTransientAlarmsNoCallback(
            boolean explicitlyAcknowledgeTransientAlarms) {
        putNoCallback(ChannelAccessValueFactory
                .createConfigureAcknowledgeTransientAlarms(
                        explicitlyAcknowledgeTransientAlarms));
    }

    /**
     * <p>
     * Puts the specified {@link ChannelAccessValueType#DBR_PUT_ACKS
     * DBR_PUT_ACKS} to the channel. This can be used to acknowledge alarms for
     * the channel.
     * </p>
     * 
     * <p>
     * This version of put uses <code>CA_PROTO_WRITE</code> so that the server
     * does not respond to the operation. This means that there is no indication
     * of whether the operation succeeded or failed. This method will return
     * immediately after queuing the message that is supposed to be sent to the
     * server.
     * </p>
     * 
     * <p>
     * Only errors in the parameters to this method or a channel that is
     * disconnected, destroyed, or not writable will result in an immediate
     * exception thrown by this method.
     * </p>
     * 
     * @param acknowledgeAlarmSeverity
     *            highest alarm severity that should be acknowledged (never
     *            <code>null</code>). This means that the alarm will be
     *            acknowledged if the severity of the unacknowledged alarm is
     *            less than or equal to the specified severity.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or not writable or there is a problem with the
     *             specified parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if <code>acknowledgeAlarmSeverity</code> is
     *             <code>null</code>.
     * @since EPICS Jackie 3.0.0
     */
    default void putAcknowledgeAlarmNoCallback(
            ChannelAccessAlarmSeverity acknowledgeAlarmSeverity) {
        putNoCallback(ChannelAccessValueFactory
                .createAcknowledgeAlarm(acknowledgeAlarmSeverity));
    }

}
